/* Copyright 1996 - 1997 by Abacus Research and
 * Development, Inc.  All rights reserved.
 */

#if !defined (OMIT_RCSID_STRINGS)
char ROMlib_rcsid_trapname[] = "$Id: trapname.c 63 2004-12-24 18:19:43Z ctm $";
#endif

#include "rsys/common.h"
#include "rsys/trapname.h"

#if !defined (NDEBUG)

/* Returns the trap name for the given trap number.  For convenience, the
 * "0xA" in the trap number is optional.
 */
const char *
trap_name (int trapno)
{
  const char *retval;

  if ((trapno & ~0xFFF) == 0xA000 || ((trapno & ~0xFFF) == 0))
    retval = trap_name_array[trapno & 0xFFF];
  else
    retval = "??? unknown trap number ???";

  return retval;
}


/* FIXME - this should be machine-generated from a central source.
 * This can wait until Executor 2 ships and we have a central source
 * of all trap information.
 */
const char *trap_name_array[0x1000] = {
  "PBOpen",                       /* 0xA000 */
  "PBClose",                      /* 0xA001 */
  "PBRead",                       /* 0xA002 */
  "PBWrite",                      /* 0xA003 */
  "PBControl",                    /* 0xA004 */
  "PBStatus",                     /* 0xA005 */
  "PBKillIO",                     /* 0xA006 */
  "PBGetVInfo",                   /* 0xA007 */
  "PBCreate",                     /* 0xA008 */
  "PBDelete",                     /* 0xA009 */
  "PBOpenRF",                     /* 0xA00A */
  "PBRename",                     /* 0xA00B */
  "PBGetFInfo",                   /* 0xA00C */
  "PBSetFInfo",                   /* 0xA00D */
  "PBUnmountVol",                 /* 0xA00E */
  "PBMountVol",                   /* 0xA00F */
  "PBAllocate",                   /* 0xA010 */
  "PBGetEOF",                     /* 0xA011 */
  "PBSetEOF",                     /* 0xA012 */
  "PBFlushVol",                   /* 0xA013 */
  "PBGetVol",                     /* 0xA014 */
  "PBSetVol",                     /* 0xA015 */
  "FInitQueue",                   /* 0xA016 */
  "PBEject",                      /* 0xA017 */
  "PBGetFPos",                    /* 0xA018 */
  "InitZone",                     /* 0xA019 */
  "GetZone",                      /* 0xA01A */
  "SetZone",                      /* 0xA01B */
  "FreeMem",                      /* 0xA01C */
  "MaxMem",                       /* 0xA01D */
  "NewPtr",                       /* 0xA01E */
  "DisposPtr",                    /* 0xA01F */
  "SetPtrSize",                   /* 0xA020 */
  "GetPtrSize",                   /* 0xA021 */
  "NewHandle",                    /* 0xA022 */
  "DisposHandle",                 /* 0xA023 */
  "SetHandleSize",                /* 0xA024 */
  "GetHandleSize",                /* 0xA025 */
  "HandleZone",                   /* 0xA026 */
  "ReallocHandle",                /* 0xA027 */
  "RecoverHandle",                /* 0xA028 */
  "HLock",                        /* 0xA029 */
  "HUnlock",                      /* 0xA02A */
  "EmptyHandle",                  /* 0xA02B */
  "InitApplZone",                 /* 0xA02C */
  "SetApplLimit",                 /* 0xA02D */
  "BlockMove",                    /* 0xA02E */
  "PostEvent",                    /* 0xA02F */
  "OSEventAvail",                 /* 0xA030 */
  "GetOSEvent",                   /* 0xA031 */
  "FlushEvents",                  /* 0xA032 */
  "VInstall",                     /* 0xA033 */
  "VRemove",                      /* 0xA034 */
  "PBOffLine",                    /* 0xA035 */
  "MoreMasters",                  /* 0xA036 */
  "A037",                          /* 0xA037 */
  "WriteParam",                   /* 0xA038 */
  "ReadDateTime",                 /* 0xA039 */
  "SetDateTime",                  /* 0xA03A */
  "Delay",                        /* 0xA03B */
  "EqualString",                  /* 0xA03C */
  "DrvrInstall",                  /* 0xA03D */
  "DrvrRemove",                   /* 0xA03E */
  "InitUtil",                     /* 0xA03F */
  "ResrvMem",                     /* 0xA040 */
  "PBSetFLock",                   /* 0xA041 */
  "PBRstFLock",                   /* 0xA042 */
  "PBSetFVers",                   /* 0xA043 */
  "PBSetFPos",                    /* 0xA044 */
  "PBFlushFile",                  /* 0xA045 */
  "GetTrapAddress",               /* 0xA046 */
  "SetTrapAddress",               /* 0xA047 */
  "PtrZone",                      /* 0xA048 */
  "HPurge",                       /* 0xA049 */
  "HNoPurge",                     /* 0xA04A */
  "SetGrowZone",                  /* 0xA04B */
  "CompactMem",                   /* 0xA04C */
  "PurgeMem",                     /* 0xA04D */
  "AddDrive",                     /* 0xA04E */
  "RDrvrInstall",                 /* 0xA04F */
  "RelString",                    /* 0xA050 */
  "IMVI_ReadXPRam",               /* 0xA051 */
  "IMVI_WriteXPRam",              /* 0xA052 */
  "A053",                          /* 0xA053 */
  "UprString",                    /* 0xA054 */
  "StripAddress",                 /* 0xA055 */
  "IMVI_LowerText",               /* 0xA056 */
  "SetApplBase",                  /* 0xA057 */
  "InsTime",                      /* 0xA058 */
  "RmvTime",                      /* 0xA059 */
  "PrimeTime",                    /* 0xA05A */
  "PowerOff",                     /* 0xA05B */
  "IMVI_MemoryDispatch",          /* 0xA05C */
  "SwapMMUMode",                  /* 0xA05D */
  "NMInstall",                    /* 0xA05E */
  "NMRemove",                     /* 0xA05F */
  "HFSDispatch",                  /* 0xA060 */
  "MaxBlock",                     /* 0xA061 */
  "PurgeSpace",                   /* 0xA062 */
  "MaxApplZone",                  /* 0xA063 */
  "MoveHHi",                      /* 0xA064 */
  "StackSpace",                   /* 0xA065 */
  "NewEmptyHandle",               /* 0xA066 */
  "HSetRBit",                     /* 0xA067 */
  "HClrRBit",                     /* 0xA068 */
  "HGetState",                    /* 0xA069 */
  "HSetState",                    /* 0xA06A */
  "A06B",                          /* 0xA06B */
  "InitFS",                       /* 0xA06C */
  "InitEvents",                   /* 0xA06D */
  "SlotManager",                  /* 0xA06E */
  "SlotVInstall",                 /* 0xA06F */
  "SlotVRemove",                  /* 0xA070 */
  "AttachVBL",                    /* 0xA071 */
  "DoVBLTask",                    /* 0xA072 */
  "A073",                          /* 0xA073 */
  "A074",                          /* 0xA074 */
  "DTInstall",                    /* 0xA075 */
  "SIntRemove",                   /* 0xA076 */
  "CountADBs",                    /* 0xA077 */
  "GetIndADB",                    /* 0xA078 */
  "GetADBInfo",                   /* 0xA079 */
  "SetADBInfo",                   /* 0xA07A */
  "ADBReInit",                    /* 0xA07B */
  "ADBOp",                        /* 0xA07C */
  "GetDefaultStartup",            /* 0xA07D */
  "SetDefaultStartup",            /* 0xA07E */
  "InternalWait",                 /* 0xA07F */
  "GetVideoDefault",              /* 0xA080 */
  "SetVideoDefault",              /* 0xA081 */
  "SIntInstall",                  /* 0xA082 */
  "SetOSDefault",                 /* 0xA083 */
  "GetOSDefault",                 /* 0xA084 */
  "IMVI_IdleUpdate",              /* 0xA085 */
  "IOPInfoAccess",                /* 0xA086 */
  "IOPMsgRequest",                /* 0xA087 */
  "IOPMoveData",                  /* 0xA088 */
  "SCSIAtomic",                   /* 0xA089 */
  "IMVI_SlpQInstall",             /* 0xA08A */
  "CommToolboxDispatch",          /* 0xA08B */
  "A08C",                          /* 0xA08C */
  "IMVI_DebugUtil",               /* 0xA08D */
  "A08E",                          /* 0xA08E */
  "IMVI_DeferUserFn",             /* 0xA08F */
  "SysEnvirons",                  /* 0xA090 */
  "IMVI_Translate24To32",         /* 0xA091 */
  "EgretDispatch",                /* 0xA092 */
  "Microseconds",                 /* 0xA093 */
  "A094",                          /* 0xA094 */
  "A095",                          /* 0xA095 */
  "A096",                          /* 0xA096 */
  "A097",                          /* 0xA097 */
  "HWPriv",                       /* 0xA098 */
  "A099",                          /* 0xA099 */
  "A09A",                          /* 0xA09A */
  "A09B",                          /* 0xA09B */
  "A09C",                          /* 0xA09C */
  "A09D",                          /* 0xA09D */
  "A09E",                          /* 0xA09E */
  "PowerDispatch",                /* 0xA09F */
  "A0A0",                          /* 0xA0A0 */
  "A0A1",                          /* 0xA0A1 */
  "A0A2",                          /* 0xA0A2 */
  "A0A3",                          /* 0xA0A3 */
  "HeapDispatch",                 /* 0xA0A4 */
  "A0A5",                          /* 0xA0A5 */
  "A0A6",                          /* 0xA0A6 */
  "A0A7",                          /* 0xA0A7 */
  "A0A8",                          /* 0xA0A8 */
  "A0A9",                          /* 0xA0A9 */
  "A0AA",                          /* 0xA0AA */
  "A0AB",                          /* 0xA0AB */
  "A0AC",                          /* 0xA0AC */
  "Gestalt",                      /* 0xA0AD */
  "VADBProc",                     /* 0xA0AE */
  "A0AF",                          /* 0xA0AF */
  "A0B0",                          /* 0xA0B0 */
  "A0B1",                          /* 0xA0B1 */
  "A0B2",                          /* 0xA0B2 */
  "A0B3",                          /* 0xA0B3 */
  "A0B4",                          /* 0xA0B4 */
  "A0B5",                          /* 0xA0B5 */
  "A0B6",                          /* 0xA0B6 */
  "A0B7",                          /* 0xA0B7 */
  "A0B8",                          /* 0xA0B8 */
  "A0B9",                          /* 0xA0B9 */
  "A0BA",                          /* 0xA0BA */
  "A0BB",                          /* 0xA0BB */
  "A0BC",                          /* 0xA0BC */
  "flushcache",                   /* 0xA0BD */
  "A0BE",                          /* 0xA0BE */
  "A0BF",                          /* 0xA0BF */
  "A0C0",                          /* 0xA0C0 */
  "A0C1",                          /* 0xA0C1 */
  "A0C2",                          /* 0xA0C2 */
  "A0C3",                          /* 0xA0C3 */
  "A0C4",                          /* 0xA0C4 */
  "A0C5",                          /* 0xA0C5 */
  "A0C6",                          /* 0xA0C6 */
  "A0C7",                          /* 0xA0C7 */
  "A0C8",                          /* 0xA0C8 */
  "A0C9",                          /* 0xA0C9 */
  "A0CA",                          /* 0xA0CA */
  "A0CB",                          /* 0xA0CB */
  "A0CC",                          /* 0xA0CC */
  "A0CD",                          /* 0xA0CD */
  "A0CE",                          /* 0xA0CE */
  "A0CF",                          /* 0xA0CF */
  "A0D0",                          /* 0xA0D0 */
  "A0D1",                          /* 0xA0D1 */
  "A0D2",                          /* 0xA0D2 */
  "A0D3",                          /* 0xA0D3 */
  "A0D4",                          /* 0xA0D4 */
  "A0D5",                          /* 0xA0D5 */
  "A0D6",                          /* 0xA0D6 */
  "A0D7",                          /* 0xA0D7 */
  "A0D8",                          /* 0xA0D8 */
  "A0D9",                          /* 0xA0D9 */
  "A0DA",                          /* 0xA0DA */
  "A0DB",                          /* 0xA0DB */
  "A0DC",                          /* 0xA0DC */
  "IMVI_PPC",                     /* 0xA0DD */
  "A0DE",                          /* 0xA0DE */
  "A0DF",                          /* 0xA0DF */
  "A0E0",                          /* 0xA0E0 */
  "A0E1",                          /* 0xA0E1 */
  "A0E2",                          /* 0xA0E2 */
  "A0E3",                          /* 0xA0E3 */
  "A0E4",                          /* 0xA0E4 */
  "A0E5",                          /* 0xA0E5 */
  "A0E6",                          /* 0xA0E6 */
  "A0E7",                          /* 0xA0E7 */
  "A0E8",                          /* 0xA0E8 */
  "A0E9",                          /* 0xA0E9 */
  "A0EA",                          /* 0xA0EA */
  "A0EB",                          /* 0xA0EB */
  "A0EC",                          /* 0xA0EC */
  "A0ED",                          /* 0xA0ED */
  "A0EE",                          /* 0xA0EE */
  "A0EF",                          /* 0xA0EF */
  "A0F0",                          /* 0xA0F0 */
  "A0F1",                          /* 0xA0F1 */
  "A0F2",                          /* 0xA0F2 */
  "A0F3",                          /* 0xA0F3 */
  "A0F4",                          /* 0xA0F4 */
  "A0F5",                          /* 0xA0F5 */
  "A0F6",                          /* 0xA0F6 */
  "A0F7",                          /* 0xA0F7 */
  "A0F8",                          /* 0xA0F8 */
  "A0F9",                          /* 0xA0F9 */
  "A0FA",                          /* 0xA0FA */
  "A0FB",                          /* 0xA0FB */
  "ResourceStub",                 /* 0xA0FC */
  "A0FD",                          /* 0xA0FD */
  "TEFindWord",                   /* 0xA0FE */
  "TEFindLine",                   /* 0xA0FF */
  "PBOpen",                       /* 0xA100 */
  "PBClose",                      /* 0xA101 */
  "PBRead",                       /* 0xA102 */
  "PBWrite",                      /* 0xA103 */
  "PBControl",                    /* 0xA104 */
  "PBStatus",                     /* 0xA105 */
  "PBKillIO",                     /* 0xA106 */
  "PBGetVInfo",                   /* 0xA107 */
  "PBCreate",                     /* 0xA108 */
  "PBDelete",                     /* 0xA109 */
  "PBOpenRF",                     /* 0xA10A */
  "PBRename",                     /* 0xA10B */
  "PBGetFInfo",                   /* 0xA10C */
  "PBSetFInfo",                   /* 0xA10D */
  "PBUnmountVol",                 /* 0xA10E */
  "PBMountVol",                   /* 0xA10F */
  "PBAllocate",                   /* 0xA110 */
  "PBGetEOF",                     /* 0xA111 */
  "PBSetEOF",                     /* 0xA112 */
  "PBFlushVol",                   /* 0xA113 */
  "PBGetVol",                     /* 0xA114 */
  "PBSetVol",                     /* 0xA115 */
  "FInitQueue",                   /* 0xA116 */
  "PBEject",                      /* 0xA117 */
  "PBGetFPos",                    /* 0xA118 */
  "InitZone",                     /* 0xA119 */
  "GetZone",                      /* 0xA11A */
  "SetZone",                      /* 0xA11B */
  "FreeMem",                      /* 0xA11C */
  "MaxMem",                       /* 0xA11D */
  "NewPtr",                       /* 0xA11E */
  "DisposPtr",                    /* 0xA11F */
  "SetPtrSize",                   /* 0xA120 */
  "GetPtrSize",                   /* 0xA121 */
  "NewHandle",                    /* 0xA122 */
  "DisposHandle",                 /* 0xA123 */
  "SetHandleSize",                /* 0xA124 */
  "GetHandleSize",                /* 0xA125 */
  "HandleZone",                   /* 0xA126 */
  "ReallocHandle",                /* 0xA127 */
  "RecoverHandle",                /* 0xA128 */
  "HLock",                        /* 0xA129 */
  "HUnlock",                      /* 0xA12A */
  "EmptyHandle",                  /* 0xA12B */
  "InitApplZone",                 /* 0xA12C */
  "SetApplLimit",                 /* 0xA12D */
  "BlockMove",                    /* 0xA12E */
  "PostEvent",                    /* 0xA12F */
  "OSEventAvail",                 /* 0xA130 */
  "GetOSEvent",                   /* 0xA131 */
  "FlushEvents",                  /* 0xA132 */
  "VInstall",                     /* 0xA133 */
  "VRemove",                      /* 0xA134 */
  "PBOffLine",                    /* 0xA135 */
  "MoreMasters",                  /* 0xA136 */
  "A137",                          /* 0xA137 */
  "WriteParam",                   /* 0xA138 */
  "ReadDateTime",                 /* 0xA139 */
  "SetDateTime",                  /* 0xA13A */
  "Delay",                        /* 0xA13B */
  "EqualString",                  /* 0xA13C */
  "DrvrInstall",                  /* 0xA13D */
  "DrvrRemove",                   /* 0xA13E */
  "InitUtil",                     /* 0xA13F */
  "ResrvMem",                     /* 0xA140 */
  "PBSetFLock",                   /* 0xA141 */
  "PBRstFLock",                   /* 0xA142 */
  "PBSetFVers",                   /* 0xA143 */
  "PBSetFPos",                    /* 0xA144 */
  "PBFlushFile",                  /* 0xA145 */
  "GetTrapAddress",               /* 0xA146 */
  "SetTrapAddress",               /* 0xA147 */
  "PtrZone",                      /* 0xA148 */
  "HPurge",                       /* 0xA149 */
  "HNoPurge",                     /* 0xA14A */
  "SetGrowZone",                  /* 0xA14B */
  "CompactMem",                   /* 0xA14C */
  "PurgeMem",                     /* 0xA14D */
  "AddDrive",                     /* 0xA14E */
  "RDrvrInstall",                 /* 0xA14F */
  "RelString",                    /* 0xA150 */
  "IMVI_ReadXPRam",               /* 0xA151 */
  "IMVI_WriteXPRam",              /* 0xA152 */
  "A153",                          /* 0xA153 */
  "UprString",                    /* 0xA154 */
  "StripAddress",                 /* 0xA155 */
  "IMVI_LowerText",               /* 0xA156 */
  "SetApplBase",                  /* 0xA157 */
  "InsTime",                      /* 0xA158 */
  "RmvTime",                      /* 0xA159 */
  "PrimeTime",                    /* 0xA15A */
  "PowerOff",                     /* 0xA15B */
  "IMVI_MemoryDispatch",          /* 0xA15C */
  "SwapMMUMode",                  /* 0xA15D */
  "NMInstall",                    /* 0xA15E */
  "NMRemove",                     /* 0xA15F */
  "HFSDispatch",                  /* 0xA160 */
  "MaxBlock",                     /* 0xA161 */
  "PurgeSpace",                   /* 0xA162 */
  "MaxApplZone",                  /* 0xA163 */
  "MoveHHi",                      /* 0xA164 */
  "StackSpace",                   /* 0xA165 */
  "NewEmptyHandle",               /* 0xA166 */
  "HSetRBit",                     /* 0xA167 */
  "HClrRBit",                     /* 0xA168 */
  "HGetState",                    /* 0xA169 */
  "HSetState",                    /* 0xA16A */
  "A16B",                          /* 0xA16B */
  "InitFS",                       /* 0xA16C */
  "InitEvents",                   /* 0xA16D */
  "SlotManager",                  /* 0xA16E */
  "SlotVInstall",                 /* 0xA16F */
  "SlotVRemove",                  /* 0xA170 */
  "AttachVBL",                    /* 0xA171 */
  "DoVBLTask",                    /* 0xA172 */
  "A173",                          /* 0xA173 */
  "A174",                          /* 0xA174 */
  "DTInstall",                    /* 0xA175 */
  "SIntRemove",                   /* 0xA176 */
  "CountADBs",                    /* 0xA177 */
  "GetIndADB",                    /* 0xA178 */
  "GetADBInfo",                   /* 0xA179 */
  "SetADBInfo",                   /* 0xA17A */
  "ADBReInit",                    /* 0xA17B */
  "ADBOp",                        /* 0xA17C */
  "GetDefaultStartup",            /* 0xA17D */
  "SetDefaultStartup",            /* 0xA17E */
  "InternalWait",                 /* 0xA17F */
  "GetVideoDefault",              /* 0xA180 */
  "SetVideoDefault",              /* 0xA181 */
  "SIntInstall",                  /* 0xA182 */
  "SetOSDefault",                 /* 0xA183 */
  "GetOSDefault",                 /* 0xA184 */
  "IMVI_IdleUpdate",              /* 0xA185 */
  "IOPInfoAccess",                /* 0xA186 */
  "IOPMsgRequest",                /* 0xA187 */
  "IOPMoveData",                  /* 0xA188 */
  "SCSIAtomic",                   /* 0xA189 */
  "IMVI_SlpQInstall",             /* 0xA18A */
  "CommToolboxDispatch",          /* 0xA18B */
  "A18C",                          /* 0xA18C */
  "IMVI_DebugUtil",               /* 0xA18D */
  "A18E",                          /* 0xA18E */
  "IMVI_DeferUserFn",             /* 0xA18F */
  "SysEnvirons",                  /* 0xA190 */
  "IMVI_Translate24To32",         /* 0xA191 */
  "EgretDispatch",                /* 0xA192 */
  "Microseconds",                 /* 0xA193 */
  "A194",                          /* 0xA194 */
  "A195",                          /* 0xA195 */
  "A196",                          /* 0xA196 */
  "A197",                          /* 0xA197 */
  "HWPriv",                       /* 0xA198 */
  "A199",                          /* 0xA199 */
  "A19A",                          /* 0xA19A */
  "A19B",                          /* 0xA19B */
  "A19C",                          /* 0xA19C */
  "A19D",                          /* 0xA19D */
  "A19E",                          /* 0xA19E */
  "PowerDispatch",                /* 0xA19F */
  "A1A0",                          /* 0xA1A0 */
  "A1A1",                          /* 0xA1A1 */
  "A1A2",                          /* 0xA1A2 */
  "A1A3",                          /* 0xA1A3 */
  "HeapDispatch",                 /* 0xA1A4 */
  "A1A5",                          /* 0xA1A5 */
  "A1A6",                          /* 0xA1A6 */
  "A1A7",                          /* 0xA1A7 */
  "A1A8",                          /* 0xA1A8 */
  "A1A9",                          /* 0xA1A9 */
  "A1AA",                          /* 0xA1AA */
  "A1AB",                          /* 0xA1AB */
  "A1AC",                          /* 0xA1AC */
  "Gestalt",                      /* 0xA1AD */
  "VADBProc",                     /* 0xA1AE */
  "A1AF",                          /* 0xA1AF */
  "A1B0",                          /* 0xA1B0 */
  "A1B1",                          /* 0xA1B1 */
  "A1B2",                          /* 0xA1B2 */
  "A1B3",                          /* 0xA1B3 */
  "A1B4",                          /* 0xA1B4 */
  "A1B5",                          /* 0xA1B5 */
  "A1B6",                          /* 0xA1B6 */
  "A1B7",                          /* 0xA1B7 */
  "A1B8",                          /* 0xA1B8 */
  "A1B9",                          /* 0xA1B9 */
  "A1BA",                          /* 0xA1BA */
  "A1BB",                          /* 0xA1BB */
  "A1BC",                          /* 0xA1BC */
  "flushcache",                   /* 0xA1BD */
  "A1BE",                          /* 0xA1BE */
  "A1BF",                          /* 0xA1BF */
  "A1C0",                          /* 0xA1C0 */
  "A1C1",                          /* 0xA1C1 */
  "A1C2",                          /* 0xA1C2 */
  "A1C3",                          /* 0xA1C3 */
  "A1C4",                          /* 0xA1C4 */
  "A1C5",                          /* 0xA1C5 */
  "A1C6",                          /* 0xA1C6 */
  "A1C7",                          /* 0xA1C7 */
  "A1C8",                          /* 0xA1C8 */
  "A1C9",                          /* 0xA1C9 */
  "A1CA",                          /* 0xA1CA */
  "A1CB",                          /* 0xA1CB */
  "A1CC",                          /* 0xA1CC */
  "A1CD",                          /* 0xA1CD */
  "A1CE",                          /* 0xA1CE */
  "A1CF",                          /* 0xA1CF */
  "A1D0",                          /* 0xA1D0 */
  "A1D1",                          /* 0xA1D1 */
  "A1D2",                          /* 0xA1D2 */
  "A1D3",                          /* 0xA1D3 */
  "A1D4",                          /* 0xA1D4 */
  "A1D5",                          /* 0xA1D5 */
  "A1D6",                          /* 0xA1D6 */
  "A1D7",                          /* 0xA1D7 */
  "A1D8",                          /* 0xA1D8 */
  "A1D9",                          /* 0xA1D9 */
  "A1DA",                          /* 0xA1DA */
  "A1DB",                          /* 0xA1DB */
  "A1DC",                          /* 0xA1DC */
  "IMVI_PPC",                     /* 0xA1DD */
  "A1DE",                          /* 0xA1DE */
  "A1DF",                          /* 0xA1DF */
  "A1E0",                          /* 0xA1E0 */
  "A1E1",                          /* 0xA1E1 */
  "A1E2",                          /* 0xA1E2 */
  "A1E3",                          /* 0xA1E3 */
  "A1E4",                          /* 0xA1E4 */
  "A1E5",                          /* 0xA1E5 */
  "A1E6",                          /* 0xA1E6 */
  "A1E7",                          /* 0xA1E7 */
  "A1E8",                          /* 0xA1E8 */
  "A1E9",                          /* 0xA1E9 */
  "A1EA",                          /* 0xA1EA */
  "A1EB",                          /* 0xA1EB */
  "A1EC",                          /* 0xA1EC */
  "A1ED",                          /* 0xA1ED */
  "A1EE",                          /* 0xA1EE */
  "A1EF",                          /* 0xA1EF */
  "A1F0",                          /* 0xA1F0 */
  "A1F1",                          /* 0xA1F1 */
  "A1F2",                          /* 0xA1F2 */
  "A1F3",                          /* 0xA1F3 */
  "A1F4",                          /* 0xA1F4 */
  "A1F5",                          /* 0xA1F5 */
  "A1F6",                          /* 0xA1F6 */
  "A1F7",                          /* 0xA1F7 */
  "A1F8",                          /* 0xA1F8 */
  "A1F9",                          /* 0xA1F9 */
  "A1FA",                          /* 0xA1FA */
  "A1FB",                          /* 0xA1FB */
  "ResourceStub",                 /* 0xA1FC */
  "A1FD",                          /* 0xA1FD */
  "TEFindWord",                   /* 0xA1FE */
  "TEFindLine",                   /* 0xA1FF */
  "PBOpen",                       /* 0xA200 */
  "PBClose",                      /* 0xA201 */
  "PBRead",                       /* 0xA202 */
  "PBWrite",                      /* 0xA203 */
  "PBControl",                    /* 0xA204 */
  "PBStatus",                     /* 0xA205 */
  "PBKillIO",                     /* 0xA206 */
  "PBGetVInfo",                   /* 0xA207 */
  "PBCreate",                     /* 0xA208 */
  "PBDelete",                     /* 0xA209 */
  "PBOpenRF",                     /* 0xA20A */
  "PBRename",                     /* 0xA20B */
  "PBGetFInfo",                   /* 0xA20C */
  "PBSetFInfo",                   /* 0xA20D */
  "PBUnmountVol",                 /* 0xA20E */
  "PBMountVol",                   /* 0xA20F */
  "PBAllocate",                   /* 0xA210 */
  "PBGetEOF",                     /* 0xA211 */
  "PBSetEOF",                     /* 0xA212 */
  "PBFlushVol",                   /* 0xA213 */
  "PBGetVol",                     /* 0xA214 */
  "PBSetVol",                     /* 0xA215 */
  "FInitQueue",                   /* 0xA216 */
  "PBEject",                      /* 0xA217 */
  "PBGetFPos",                    /* 0xA218 */
  "InitZone",                     /* 0xA219 */
  "GetZone",                      /* 0xA21A */
  "SetZone",                      /* 0xA21B */
  "FreeMem",                      /* 0xA21C */
  "MaxMem",                       /* 0xA21D */
  "NewPtr",                       /* 0xA21E */
  "DisposPtr",                    /* 0xA21F */
  "SetPtrSize",                   /* 0xA220 */
  "GetPtrSize",                   /* 0xA221 */
  "NewHandle",                    /* 0xA222 */
  "DisposHandle",                 /* 0xA223 */
  "SetHandleSize",                /* 0xA224 */
  "GetHandleSize",                /* 0xA225 */
  "HandleZone",                   /* 0xA226 */
  "ReallocHandle",                /* 0xA227 */
  "RecoverHandle",                /* 0xA228 */
  "HLock",                        /* 0xA229 */
  "HUnlock",                      /* 0xA22A */
  "EmptyHandle",                  /* 0xA22B */
  "InitApplZone",                 /* 0xA22C */
  "SetApplLimit",                 /* 0xA22D */
  "BlockMove",                    /* 0xA22E */
  "PostEvent",                    /* 0xA22F */
  "OSEventAvail",                 /* 0xA230 */
  "GetOSEvent",                   /* 0xA231 */
  "FlushEvents",                  /* 0xA232 */
  "VInstall",                     /* 0xA233 */
  "VRemove",                      /* 0xA234 */
  "PBOffLine",                    /* 0xA235 */
  "MoreMasters",                  /* 0xA236 */
  "A237",                          /* 0xA237 */
  "WriteParam",                   /* 0xA238 */
  "ReadDateTime",                 /* 0xA239 */
  "SetDateTime",                  /* 0xA23A */
  "Delay",                        /* 0xA23B */
  "EqualString",                  /* 0xA23C */
  "DrvrInstall",                  /* 0xA23D */
  "DrvrRemove",                   /* 0xA23E */
  "InitUtil",                     /* 0xA23F */
  "ResrvMem",                     /* 0xA240 */
  "PBSetFLock",                   /* 0xA241 */
  "PBRstFLock",                   /* 0xA242 */
  "PBSetFVers",                   /* 0xA243 */
  "PBSetFPos",                    /* 0xA244 */
  "PBFlushFile",                  /* 0xA245 */
  "GetTrapAddress",               /* 0xA246 */
  "SetTrapAddress",               /* 0xA247 */
  "PtrZone",                      /* 0xA248 */
  "HPurge",                       /* 0xA249 */
  "HNoPurge",                     /* 0xA24A */
  "SetGrowZone",                  /* 0xA24B */
  "CompactMem",                   /* 0xA24C */
  "PurgeMem",                     /* 0xA24D */
  "AddDrive",                     /* 0xA24E */
  "RDrvrInstall",                 /* 0xA24F */
  "RelString",                    /* 0xA250 */
  "IMVI_ReadXPRam",               /* 0xA251 */
  "IMVI_WriteXPRam",              /* 0xA252 */
  "A253",                          /* 0xA253 */
  "UprString",                    /* 0xA254 */
  "StripAddress",                 /* 0xA255 */
  "IMVI_LowerText",               /* 0xA256 */
  "SetApplBase",                  /* 0xA257 */
  "InsTime",                      /* 0xA258 */
  "RmvTime",                      /* 0xA259 */
  "PrimeTime",                    /* 0xA25A */
  "PowerOff",                     /* 0xA25B */
  "IMVI_MemoryDispatch",          /* 0xA25C */
  "SwapMMUMode",                  /* 0xA25D */
  "NMInstall",                    /* 0xA25E */
  "NMRemove",                     /* 0xA25F */
  "HFSDispatch",                  /* 0xA260 */
  "MaxBlock",                     /* 0xA261 */
  "PurgeSpace",                   /* 0xA262 */
  "MaxApplZone",                  /* 0xA263 */
  "MoveHHi",                      /* 0xA264 */
  "StackSpace",                   /* 0xA265 */
  "NewEmptyHandle",               /* 0xA266 */
  "HSetRBit",                     /* 0xA267 */
  "HClrRBit",                     /* 0xA268 */
  "HGetState",                    /* 0xA269 */
  "HSetState",                    /* 0xA26A */
  "A26B",                          /* 0xA26B */
  "InitFS",                       /* 0xA26C */
  "InitEvents",                   /* 0xA26D */
  "SlotManager",                  /* 0xA26E */
  "SlotVInstall",                 /* 0xA26F */
  "SlotVRemove",                  /* 0xA270 */
  "AttachVBL",                    /* 0xA271 */
  "DoVBLTask",                    /* 0xA272 */
  "A273",                          /* 0xA273 */
  "A274",                          /* 0xA274 */
  "DTInstall",                    /* 0xA275 */
  "SIntRemove",                   /* 0xA276 */
  "CountADBs",                    /* 0xA277 */
  "GetIndADB",                    /* 0xA278 */
  "GetADBInfo",                   /* 0xA279 */
  "SetADBInfo",                   /* 0xA27A */
  "ADBReInit",                    /* 0xA27B */
  "ADBOp",                        /* 0xA27C */
  "GetDefaultStartup",            /* 0xA27D */
  "SetDefaultStartup",            /* 0xA27E */
  "InternalWait",                 /* 0xA27F */
  "GetVideoDefault",              /* 0xA280 */
  "SetVideoDefault",              /* 0xA281 */
  "SIntInstall",                  /* 0xA282 */
  "SetOSDefault",                 /* 0xA283 */
  "GetOSDefault",                 /* 0xA284 */
  "IMVI_IdleUpdate",              /* 0xA285 */
  "IOPInfoAccess",                /* 0xA286 */
  "IOPMsgRequest",                /* 0xA287 */
  "IOPMoveData",                  /* 0xA288 */
  "SCSIAtomic",                   /* 0xA289 */
  "IMVI_SlpQInstall",             /* 0xA28A */
  "CommToolboxDispatch",          /* 0xA28B */
  "A28C",                          /* 0xA28C */
  "IMVI_DebugUtil",               /* 0xA28D */
  "A28E",                          /* 0xA28E */
  "IMVI_DeferUserFn",             /* 0xA28F */
  "SysEnvirons",                  /* 0xA290 */
  "IMVI_Translate24To32",         /* 0xA291 */
  "EgretDispatch",                /* 0xA292 */
  "Microseconds",                 /* 0xA293 */
  "A294",                          /* 0xA294 */
  "A295",                          /* 0xA295 */
  "A296",                          /* 0xA296 */
  "A297",                          /* 0xA297 */
  "HWPriv",                       /* 0xA298 */
  "A299",                          /* 0xA299 */
  "A29A",                          /* 0xA29A */
  "A29B",                          /* 0xA29B */
  "A29C",                          /* 0xA29C */
  "A29D",                          /* 0xA29D */
  "A29E",                          /* 0xA29E */
  "PowerDispatch",                /* 0xA29F */
  "A2A0",                          /* 0xA2A0 */
  "A2A1",                          /* 0xA2A1 */
  "A2A2",                          /* 0xA2A2 */
  "A2A3",                          /* 0xA2A3 */
  "HeapDispatch",                 /* 0xA2A4 */
  "A2A5",                          /* 0xA2A5 */
  "A2A6",                          /* 0xA2A6 */
  "A2A7",                          /* 0xA2A7 */
  "A2A8",                          /* 0xA2A8 */
  "A2A9",                          /* 0xA2A9 */
  "A2AA",                          /* 0xA2AA */
  "A2AB",                          /* 0xA2AB */
  "A2AC",                          /* 0xA2AC */
  "Gestalt",                      /* 0xA2AD */
  "VADBProc",                     /* 0xA2AE */
  "A2AF",                          /* 0xA2AF */
  "A2B0",                          /* 0xA2B0 */
  "A2B1",                          /* 0xA2B1 */
  "A2B2",                          /* 0xA2B2 */
  "A2B3",                          /* 0xA2B3 */
  "A2B4",                          /* 0xA2B4 */
  "A2B5",                          /* 0xA2B5 */
  "A2B6",                          /* 0xA2B6 */
  "A2B7",                          /* 0xA2B7 */
  "A2B8",                          /* 0xA2B8 */
  "A2B9",                          /* 0xA2B9 */
  "A2BA",                          /* 0xA2BA */
  "A2BB",                          /* 0xA2BB */
  "A2BC",                          /* 0xA2BC */
  "flushcache",                   /* 0xA2BD */
  "A2BE",                          /* 0xA2BE */
  "A2BF",                          /* 0xA2BF */
  "A2C0",                          /* 0xA2C0 */
  "A2C1",                          /* 0xA2C1 */
  "A2C2",                          /* 0xA2C2 */
  "A2C3",                          /* 0xA2C3 */
  "A2C4",                          /* 0xA2C4 */
  "A2C5",                          /* 0xA2C5 */
  "A2C6",                          /* 0xA2C6 */
  "A2C7",                          /* 0xA2C7 */
  "A2C8",                          /* 0xA2C8 */
  "A2C9",                          /* 0xA2C9 */
  "A2CA",                          /* 0xA2CA */
  "A2CB",                          /* 0xA2CB */
  "A2CC",                          /* 0xA2CC */
  "A2CD",                          /* 0xA2CD */
  "A2CE",                          /* 0xA2CE */
  "A2CF",                          /* 0xA2CF */
  "A2D0",                          /* 0xA2D0 */
  "A2D1",                          /* 0xA2D1 */
  "A2D2",                          /* 0xA2D2 */
  "A2D3",                          /* 0xA2D3 */
  "A2D4",                          /* 0xA2D4 */
  "A2D5",                          /* 0xA2D5 */
  "A2D6",                          /* 0xA2D6 */
  "A2D7",                          /* 0xA2D7 */
  "A2D8",                          /* 0xA2D8 */
  "A2D9",                          /* 0xA2D9 */
  "A2DA",                          /* 0xA2DA */
  "A2DB",                          /* 0xA2DB */
  "A2DC",                          /* 0xA2DC */
  "IMVI_PPC",                     /* 0xA2DD */
  "A2DE",                          /* 0xA2DE */
  "A2DF",                          /* 0xA2DF */
  "A2E0",                          /* 0xA2E0 */
  "A2E1",                          /* 0xA2E1 */
  "A2E2",                          /* 0xA2E2 */
  "A2E3",                          /* 0xA2E3 */
  "A2E4",                          /* 0xA2E4 */
  "A2E5",                          /* 0xA2E5 */
  "A2E6",                          /* 0xA2E6 */
  "A2E7",                          /* 0xA2E7 */
  "A2E8",                          /* 0xA2E8 */
  "A2E9",                          /* 0xA2E9 */
  "A2EA",                          /* 0xA2EA */
  "A2EB",                          /* 0xA2EB */
  "A2EC",                          /* 0xA2EC */
  "A2ED",                          /* 0xA2ED */
  "A2EE",                          /* 0xA2EE */
  "A2EF",                          /* 0xA2EF */
  "A2F0",                          /* 0xA2F0 */
  "A2F1",                          /* 0xA2F1 */
  "A2F2",                          /* 0xA2F2 */
  "A2F3",                          /* 0xA2F3 */
  "A2F4",                          /* 0xA2F4 */
  "A2F5",                          /* 0xA2F5 */
  "A2F6",                          /* 0xA2F6 */
  "A2F7",                          /* 0xA2F7 */
  "A2F8",                          /* 0xA2F8 */
  "A2F9",                          /* 0xA2F9 */
  "A2FA",                          /* 0xA2FA */
  "A2FB",                          /* 0xA2FB */
  "ResourceStub",                 /* 0xA2FC */
  "A2FD",                          /* 0xA2FD */
  "TEFindWord",                   /* 0xA2FE */
  "TEFindLine",                   /* 0xA2FF */
  "PBOpen",                       /* 0xA300 */
  "PBClose",                      /* 0xA301 */
  "PBRead",                       /* 0xA302 */
  "PBWrite",                      /* 0xA303 */
  "PBControl",                    /* 0xA304 */
  "PBStatus",                     /* 0xA305 */
  "PBKillIO",                     /* 0xA306 */
  "PBGetVInfo",                   /* 0xA307 */
  "PBCreate",                     /* 0xA308 */
  "PBDelete",                     /* 0xA309 */
  "PBOpenRF",                     /* 0xA30A */
  "PBRename",                     /* 0xA30B */
  "PBGetFInfo",                   /* 0xA30C */
  "PBSetFInfo",                   /* 0xA30D */
  "PBUnmountVol",                 /* 0xA30E */
  "PBMountVol",                   /* 0xA30F */
  "PBAllocate",                   /* 0xA310 */
  "PBGetEOF",                     /* 0xA311 */
  "PBSetEOF",                     /* 0xA312 */
  "PBFlushVol",                   /* 0xA313 */
  "PBGetVol",                     /* 0xA314 */
  "PBSetVol",                     /* 0xA315 */
  "FInitQueue",                   /* 0xA316 */
  "PBEject",                      /* 0xA317 */
  "PBGetFPos",                    /* 0xA318 */
  "InitZone",                     /* 0xA319 */
  "GetZone",                      /* 0xA31A */
  "SetZone",                      /* 0xA31B */
  "FreeMem",                      /* 0xA31C */
  "MaxMem",                       /* 0xA31D */
  "NewPtr",                       /* 0xA31E */
  "DisposPtr",                    /* 0xA31F */
  "SetPtrSize",                   /* 0xA320 */
  "GetPtrSize",                   /* 0xA321 */
  "NewHandle",                    /* 0xA322 */
  "DisposHandle",                 /* 0xA323 */
  "SetHandleSize",                /* 0xA324 */
  "GetHandleSize",                /* 0xA325 */
  "HandleZone",                   /* 0xA326 */
  "ReallocHandle",                /* 0xA327 */
  "RecoverHandle",                /* 0xA328 */
  "HLock",                        /* 0xA329 */
  "HUnlock",                      /* 0xA32A */
  "EmptyHandle",                  /* 0xA32B */
  "InitApplZone",                 /* 0xA32C */
  "SetApplLimit",                 /* 0xA32D */
  "BlockMove",                    /* 0xA32E */
  "PostEvent",                    /* 0xA32F */
  "OSEventAvail",                 /* 0xA330 */
  "GetOSEvent",                   /* 0xA331 */
  "FlushEvents",                  /* 0xA332 */
  "VInstall",                     /* 0xA333 */
  "VRemove",                      /* 0xA334 */
  "PBOffLine",                    /* 0xA335 */
  "MoreMasters",                  /* 0xA336 */
  "A337",                          /* 0xA337 */
  "WriteParam",                   /* 0xA338 */
  "ReadDateTime",                 /* 0xA339 */
  "SetDateTime",                  /* 0xA33A */
  "Delay",                        /* 0xA33B */
  "EqualString",                  /* 0xA33C */
  "DrvrInstall",                  /* 0xA33D */
  "DrvrRemove",                   /* 0xA33E */
  "InitUtil",                     /* 0xA33F */
  "ResrvMem",                     /* 0xA340 */
  "PBSetFLock",                   /* 0xA341 */
  "PBRstFLock",                   /* 0xA342 */
  "PBSetFVers",                   /* 0xA343 */
  "PBSetFPos",                    /* 0xA344 */
  "PBFlushFile",                  /* 0xA345 */
  "GetTrapAddress",               /* 0xA346 */
  "SetTrapAddress",               /* 0xA347 */
  "PtrZone",                      /* 0xA348 */
  "HPurge",                       /* 0xA349 */
  "HNoPurge",                     /* 0xA34A */
  "SetGrowZone",                  /* 0xA34B */
  "CompactMem",                   /* 0xA34C */
  "PurgeMem",                     /* 0xA34D */
  "AddDrive",                     /* 0xA34E */
  "RDrvrInstall",                 /* 0xA34F */
  "RelString",                    /* 0xA350 */
  "IMVI_ReadXPRam",               /* 0xA351 */
  "IMVI_WriteXPRam",              /* 0xA352 */
  "A353",                          /* 0xA353 */
  "UprString",                    /* 0xA354 */
  "StripAddress",                 /* 0xA355 */
  "IMVI_LowerText",               /* 0xA356 */
  "SetApplBase",                  /* 0xA357 */
  "InsTime",                      /* 0xA358 */
  "RmvTime",                      /* 0xA359 */
  "PrimeTime",                    /* 0xA35A */
  "PowerOff",                     /* 0xA35B */
  "IMVI_MemoryDispatch",          /* 0xA35C */
  "SwapMMUMode",                  /* 0xA35D */
  "NMInstall",                    /* 0xA35E */
  "NMRemove",                     /* 0xA35F */
  "HFSDispatch",                  /* 0xA360 */
  "MaxBlock",                     /* 0xA361 */
  "PurgeSpace",                   /* 0xA362 */
  "MaxApplZone",                  /* 0xA363 */
  "MoveHHi",                      /* 0xA364 */
  "StackSpace",                   /* 0xA365 */
  "NewEmptyHandle",               /* 0xA366 */
  "HSetRBit",                     /* 0xA367 */
  "HClrRBit",                     /* 0xA368 */
  "HGetState",                    /* 0xA369 */
  "HSetState",                    /* 0xA36A */
  "A36B",                          /* 0xA36B */
  "InitFS",                       /* 0xA36C */
  "InitEvents",                   /* 0xA36D */
  "SlotManager",                  /* 0xA36E */
  "SlotVInstall",                 /* 0xA36F */
  "SlotVRemove",                  /* 0xA370 */
  "AttachVBL",                    /* 0xA371 */
  "DoVBLTask",                    /* 0xA372 */
  "A373",                          /* 0xA373 */
  "A374",                          /* 0xA374 */
  "DTInstall",                    /* 0xA375 */
  "SIntRemove",                   /* 0xA376 */
  "CountADBs",                    /* 0xA377 */
  "GetIndADB",                    /* 0xA378 */
  "GetADBInfo",                   /* 0xA379 */
  "SetADBInfo",                   /* 0xA37A */
  "ADBReInit",                    /* 0xA37B */
  "ADBOp",                        /* 0xA37C */
  "GetDefaultStartup",            /* 0xA37D */
  "SetDefaultStartup",            /* 0xA37E */
  "InternalWait",                 /* 0xA37F */
  "GetVideoDefault",              /* 0xA380 */
  "SetVideoDefault",              /* 0xA381 */
  "SIntInstall",                  /* 0xA382 */
  "SetOSDefault",                 /* 0xA383 */
  "GetOSDefault",                 /* 0xA384 */
  "IMVI_IdleUpdate",              /* 0xA385 */
  "IOPInfoAccess",                /* 0xA386 */
  "IOPMsgRequest",                /* 0xA387 */
  "IOPMoveData",                  /* 0xA388 */
  "SCSIAtomic",                   /* 0xA389 */
  "IMVI_SlpQInstall",             /* 0xA38A */
  "CommToolboxDispatch",          /* 0xA38B */
  "A38C",                          /* 0xA38C */
  "IMVI_DebugUtil",               /* 0xA38D */
  "A38E",                          /* 0xA38E */
  "IMVI_DeferUserFn",             /* 0xA38F */
  "SysEnvirons",                  /* 0xA390 */
  "IMVI_Translate24To32",         /* 0xA391 */
  "EgretDispatch",                /* 0xA392 */
  "Microseconds",                 /* 0xA393 */
  "A394",                          /* 0xA394 */
  "A395",                          /* 0xA395 */
  "A396",                          /* 0xA396 */
  "A397",                          /* 0xA397 */
  "HWPriv",                       /* 0xA398 */
  "A399",                          /* 0xA399 */
  "A39A",                          /* 0xA39A */
  "A39B",                          /* 0xA39B */
  "A39C",                          /* 0xA39C */
  "A39D",                          /* 0xA39D */
  "A39E",                          /* 0xA39E */
  "PowerDispatch",                /* 0xA39F */
  "A3A0",                          /* 0xA3A0 */
  "A3A1",                          /* 0xA3A1 */
  "A3A2",                          /* 0xA3A2 */
  "A3A3",                          /* 0xA3A3 */
  "HeapDispatch",                 /* 0xA3A4 */
  "A3A5",                          /* 0xA3A5 */
  "A3A6",                          /* 0xA3A6 */
  "A3A7",                          /* 0xA3A7 */
  "A3A8",                          /* 0xA3A8 */
  "A3A9",                          /* 0xA3A9 */
  "A3AA",                          /* 0xA3AA */
  "A3AB",                          /* 0xA3AB */
  "A3AC",                          /* 0xA3AC */
  "Gestalt",                      /* 0xA3AD */
  "VADBProc",                     /* 0xA3AE */
  "A3AF",                          /* 0xA3AF */
  "A3B0",                          /* 0xA3B0 */
  "A3B1",                          /* 0xA3B1 */
  "A3B2",                          /* 0xA3B2 */
  "A3B3",                          /* 0xA3B3 */
  "A3B4",                          /* 0xA3B4 */
  "A3B5",                          /* 0xA3B5 */
  "A3B6",                          /* 0xA3B6 */
  "A3B7",                          /* 0xA3B7 */
  "A3B8",                          /* 0xA3B8 */
  "A3B9",                          /* 0xA3B9 */
  "A3BA",                          /* 0xA3BA */
  "A3BB",                          /* 0xA3BB */
  "A3BC",                          /* 0xA3BC */
  "flushcache",                   /* 0xA3BD */
  "A3BE",                          /* 0xA3BE */
  "A3BF",                          /* 0xA3BF */
  "A3C0",                          /* 0xA3C0 */
  "A3C1",                          /* 0xA3C1 */
  "A3C2",                          /* 0xA3C2 */
  "A3C3",                          /* 0xA3C3 */
  "A3C4",                          /* 0xA3C4 */
  "A3C5",                          /* 0xA3C5 */
  "A3C6",                          /* 0xA3C6 */
  "A3C7",                          /* 0xA3C7 */
  "A3C8",                          /* 0xA3C8 */
  "A3C9",                          /* 0xA3C9 */
  "A3CA",                          /* 0xA3CA */
  "A3CB",                          /* 0xA3CB */
  "A3CC",                          /* 0xA3CC */
  "A3CD",                          /* 0xA3CD */
  "A3CE",                          /* 0xA3CE */
  "A3CF",                          /* 0xA3CF */
  "A3D0",                          /* 0xA3D0 */
  "A3D1",                          /* 0xA3D1 */
  "A3D2",                          /* 0xA3D2 */
  "A3D3",                          /* 0xA3D3 */
  "A3D4",                          /* 0xA3D4 */
  "A3D5",                          /* 0xA3D5 */
  "A3D6",                          /* 0xA3D6 */
  "A3D7",                          /* 0xA3D7 */
  "A3D8",                          /* 0xA3D8 */
  "A3D9",                          /* 0xA3D9 */
  "A3DA",                          /* 0xA3DA */
  "A3DB",                          /* 0xA3DB */
  "A3DC",                          /* 0xA3DC */
  "IMVI_PPC",                     /* 0xA3DD */
  "A3DE",                          /* 0xA3DE */
  "A3DF",                          /* 0xA3DF */
  "A3E0",                          /* 0xA3E0 */
  "A3E1",                          /* 0xA3E1 */
  "A3E2",                          /* 0xA3E2 */
  "A3E3",                          /* 0xA3E3 */
  "A3E4",                          /* 0xA3E4 */
  "A3E5",                          /* 0xA3E5 */
  "A3E6",                          /* 0xA3E6 */
  "A3E7",                          /* 0xA3E7 */
  "A3E8",                          /* 0xA3E8 */
  "A3E9",                          /* 0xA3E9 */
  "A3EA",                          /* 0xA3EA */
  "A3EB",                          /* 0xA3EB */
  "A3EC",                          /* 0xA3EC */
  "A3ED",                          /* 0xA3ED */
  "A3EE",                          /* 0xA3EE */
  "A3EF",                          /* 0xA3EF */
  "A3F0",                          /* 0xA3F0 */
  "A3F1",                          /* 0xA3F1 */
  "A3F2",                          /* 0xA3F2 */
  "A3F3",                          /* 0xA3F3 */
  "A3F4",                          /* 0xA3F4 */
  "A3F5",                          /* 0xA3F5 */
  "A3F6",                          /* 0xA3F6 */
  "A3F7",                          /* 0xA3F7 */
  "A3F8",                          /* 0xA3F8 */
  "A3F9",                          /* 0xA3F9 */
  "A3FA",                          /* 0xA3FA */
  "A3FB",                          /* 0xA3FB */
  "ResourceStub",                 /* 0xA3FC */
  "A3FD",                          /* 0xA3FD */
  "TEFindWord",                   /* 0xA3FE */
  "TEFindLine",                   /* 0xA3FF */
  "PBOpen",                       /* 0xA400 */
  "PBClose",                      /* 0xA401 */
  "PBRead",                       /* 0xA402 */
  "PBWrite",                      /* 0xA403 */
  "PBControl",                    /* 0xA404 */
  "PBStatus",                     /* 0xA405 */
  "PBKillIO",                     /* 0xA406 */
  "PBGetVInfo",                   /* 0xA407 */
  "PBCreate",                     /* 0xA408 */
  "PBDelete",                     /* 0xA409 */
  "PBOpenRF",                     /* 0xA40A */
  "PBRename",                     /* 0xA40B */
  "PBGetFInfo",                   /* 0xA40C */
  "PBSetFInfo",                   /* 0xA40D */
  "PBUnmountVol",                 /* 0xA40E */
  "PBMountVol",                   /* 0xA40F */
  "PBAllocate",                   /* 0xA410 */
  "PBGetEOF",                     /* 0xA411 */
  "PBSetEOF",                     /* 0xA412 */
  "PBFlushVol",                   /* 0xA413 */
  "PBGetVol",                     /* 0xA414 */
  "PBSetVol",                     /* 0xA415 */
  "FInitQueue",                   /* 0xA416 */
  "PBEject",                      /* 0xA417 */
  "PBGetFPos",                    /* 0xA418 */
  "InitZone",                     /* 0xA419 */
  "GetZone",                      /* 0xA41A */
  "SetZone",                      /* 0xA41B */
  "FreeMem",                      /* 0xA41C */
  "MaxMem",                       /* 0xA41D */
  "NewPtr",                       /* 0xA41E */
  "DisposPtr",                    /* 0xA41F */
  "SetPtrSize",                   /* 0xA420 */
  "GetPtrSize",                   /* 0xA421 */
  "NewHandle",                    /* 0xA422 */
  "DisposHandle",                 /* 0xA423 */
  "SetHandleSize",                /* 0xA424 */
  "GetHandleSize",                /* 0xA425 */
  "HandleZone",                   /* 0xA426 */
  "ReallocHandle",                /* 0xA427 */
  "RecoverHandle",                /* 0xA428 */
  "HLock",                        /* 0xA429 */
  "HUnlock",                      /* 0xA42A */
  "EmptyHandle",                  /* 0xA42B */
  "InitApplZone",                 /* 0xA42C */
  "SetApplLimit",                 /* 0xA42D */
  "BlockMove",                    /* 0xA42E */
  "PostEvent",                    /* 0xA42F */
  "OSEventAvail",                 /* 0xA430 */
  "GetOSEvent",                   /* 0xA431 */
  "FlushEvents",                  /* 0xA432 */
  "VInstall",                     /* 0xA433 */
  "VRemove",                      /* 0xA434 */
  "PBOffLine",                    /* 0xA435 */
  "MoreMasters",                  /* 0xA436 */
  "A437",                          /* 0xA437 */
  "WriteParam",                   /* 0xA438 */
  "ReadDateTime",                 /* 0xA439 */
  "SetDateTime",                  /* 0xA43A */
  "Delay",                        /* 0xA43B */
  "EqualString",                  /* 0xA43C */
  "DrvrInstall",                  /* 0xA43D */
  "DrvrRemove",                   /* 0xA43E */
  "InitUtil",                     /* 0xA43F */
  "ResrvMem",                     /* 0xA440 */
  "PBSetFLock",                   /* 0xA441 */
  "PBRstFLock",                   /* 0xA442 */
  "PBSetFVers",                   /* 0xA443 */
  "PBSetFPos",                    /* 0xA444 */
  "PBFlushFile",                  /* 0xA445 */
  "GetTrapAddress",               /* 0xA446 */
  "SetTrapAddress",               /* 0xA447 */
  "PtrZone",                      /* 0xA448 */
  "HPurge",                       /* 0xA449 */
  "HNoPurge",                     /* 0xA44A */
  "SetGrowZone",                  /* 0xA44B */
  "CompactMem",                   /* 0xA44C */
  "PurgeMem",                     /* 0xA44D */
  "AddDrive",                     /* 0xA44E */
  "RDrvrInstall",                 /* 0xA44F */
  "RelString",                    /* 0xA450 */
  "IMVI_ReadXPRam",               /* 0xA451 */
  "IMVI_WriteXPRam",              /* 0xA452 */
  "A453",                          /* 0xA453 */
  "UprString",                    /* 0xA454 */
  "StripAddress",                 /* 0xA455 */
  "IMVI_LowerText",               /* 0xA456 */
  "SetApplBase",                  /* 0xA457 */
  "InsTime",                      /* 0xA458 */
  "RmvTime",                      /* 0xA459 */
  "PrimeTime",                    /* 0xA45A */
  "PowerOff",                     /* 0xA45B */
  "IMVI_MemoryDispatch",          /* 0xA45C */
  "SwapMMUMode",                  /* 0xA45D */
  "NMInstall",                    /* 0xA45E */
  "NMRemove",                     /* 0xA45F */
  "HFSDispatch",                  /* 0xA460 */
  "MaxBlock",                     /* 0xA461 */
  "PurgeSpace",                   /* 0xA462 */
  "MaxApplZone",                  /* 0xA463 */
  "MoveHHi",                      /* 0xA464 */
  "StackSpace",                   /* 0xA465 */
  "NewEmptyHandle",               /* 0xA466 */
  "HSetRBit",                     /* 0xA467 */
  "HClrRBit",                     /* 0xA468 */
  "HGetState",                    /* 0xA469 */
  "HSetState",                    /* 0xA46A */
  "A46B",                          /* 0xA46B */
  "InitFS",                       /* 0xA46C */
  "InitEvents",                   /* 0xA46D */
  "SlotManager",                  /* 0xA46E */
  "SlotVInstall",                 /* 0xA46F */
  "SlotVRemove",                  /* 0xA470 */
  "AttachVBL",                    /* 0xA471 */
  "DoVBLTask",                    /* 0xA472 */
  "A473",                          /* 0xA473 */
  "A474",                          /* 0xA474 */
  "DTInstall",                    /* 0xA475 */
  "SIntRemove",                   /* 0xA476 */
  "CountADBs",                    /* 0xA477 */
  "GetIndADB",                    /* 0xA478 */
  "GetADBInfo",                   /* 0xA479 */
  "SetADBInfo",                   /* 0xA47A */
  "ADBReInit",                    /* 0xA47B */
  "ADBOp",                        /* 0xA47C */
  "GetDefaultStartup",            /* 0xA47D */
  "SetDefaultStartup",            /* 0xA47E */
  "InternalWait",                 /* 0xA47F */
  "GetVideoDefault",              /* 0xA480 */
  "SetVideoDefault",              /* 0xA481 */
  "SIntInstall",                  /* 0xA482 */
  "SetOSDefault",                 /* 0xA483 */
  "GetOSDefault",                 /* 0xA484 */
  "IMVI_IdleUpdate",              /* 0xA485 */
  "IOPInfoAccess",                /* 0xA486 */
  "IOPMsgRequest",                /* 0xA487 */
  "IOPMoveData",                  /* 0xA488 */
  "SCSIAtomic",                   /* 0xA489 */
  "IMVI_SlpQInstall",             /* 0xA48A */
  "CommToolboxDispatch",          /* 0xA48B */
  "A48C",                          /* 0xA48C */
  "IMVI_DebugUtil",               /* 0xA48D */
  "A48E",                          /* 0xA48E */
  "IMVI_DeferUserFn",             /* 0xA48F */
  "SysEnvirons",                  /* 0xA490 */
  "IMVI_Translate24To32",         /* 0xA491 */
  "EgretDispatch",                /* 0xA492 */
  "Microseconds",                 /* 0xA493 */
  "A494",                          /* 0xA494 */
  "A495",                          /* 0xA495 */
  "A496",                          /* 0xA496 */
  "A497",                          /* 0xA497 */
  "HWPriv",                       /* 0xA498 */
  "A499",                          /* 0xA499 */
  "A49A",                          /* 0xA49A */
  "A49B",                          /* 0xA49B */
  "A49C",                          /* 0xA49C */
  "A49D",                          /* 0xA49D */
  "A49E",                          /* 0xA49E */
  "PowerDispatch",                /* 0xA49F */
  "A4A0",                          /* 0xA4A0 */
  "A4A1",                          /* 0xA4A1 */
  "A4A2",                          /* 0xA4A2 */
  "A4A3",                          /* 0xA4A3 */
  "HeapDispatch",                 /* 0xA4A4 */
  "A4A5",                          /* 0xA4A5 */
  "A4A6",                          /* 0xA4A6 */
  "A4A7",                          /* 0xA4A7 */
  "A4A8",                          /* 0xA4A8 */
  "A4A9",                          /* 0xA4A9 */
  "A4AA",                          /* 0xA4AA */
  "A4AB",                          /* 0xA4AB */
  "A4AC",                          /* 0xA4AC */
  "Gestalt",                      /* 0xA4AD */
  "VADBProc",                     /* 0xA4AE */
  "A4AF",                          /* 0xA4AF */
  "A4B0",                          /* 0xA4B0 */
  "A4B1",                          /* 0xA4B1 */
  "A4B2",                          /* 0xA4B2 */
  "A4B3",                          /* 0xA4B3 */
  "A4B4",                          /* 0xA4B4 */
  "A4B5",                          /* 0xA4B5 */
  "A4B6",                          /* 0xA4B6 */
  "A4B7",                          /* 0xA4B7 */
  "A4B8",                          /* 0xA4B8 */
  "A4B9",                          /* 0xA4B9 */
  "A4BA",                          /* 0xA4BA */
  "A4BB",                          /* 0xA4BB */
  "A4BC",                          /* 0xA4BC */
  "flushcache",                   /* 0xA4BD */
  "A4BE",                          /* 0xA4BE */
  "A4BF",                          /* 0xA4BF */
  "A4C0",                          /* 0xA4C0 */
  "A4C1",                          /* 0xA4C1 */
  "A4C2",                          /* 0xA4C2 */
  "A4C3",                          /* 0xA4C3 */
  "A4C4",                          /* 0xA4C4 */
  "A4C5",                          /* 0xA4C5 */
  "A4C6",                          /* 0xA4C6 */
  "A4C7",                          /* 0xA4C7 */
  "A4C8",                          /* 0xA4C8 */
  "A4C9",                          /* 0xA4C9 */
  "A4CA",                          /* 0xA4CA */
  "A4CB",                          /* 0xA4CB */
  "A4CC",                          /* 0xA4CC */
  "A4CD",                          /* 0xA4CD */
  "A4CE",                          /* 0xA4CE */
  "A4CF",                          /* 0xA4CF */
  "A4D0",                          /* 0xA4D0 */
  "A4D1",                          /* 0xA4D1 */
  "A4D2",                          /* 0xA4D2 */
  "A4D3",                          /* 0xA4D3 */
  "A4D4",                          /* 0xA4D4 */
  "A4D5",                          /* 0xA4D5 */
  "A4D6",                          /* 0xA4D6 */
  "A4D7",                          /* 0xA4D7 */
  "A4D8",                          /* 0xA4D8 */
  "A4D9",                          /* 0xA4D9 */
  "A4DA",                          /* 0xA4DA */
  "A4DB",                          /* 0xA4DB */
  "A4DC",                          /* 0xA4DC */
  "IMVI_PPC",                     /* 0xA4DD */
  "A4DE",                          /* 0xA4DE */
  "A4DF",                          /* 0xA4DF */
  "A4E0",                          /* 0xA4E0 */
  "A4E1",                          /* 0xA4E1 */
  "A4E2",                          /* 0xA4E2 */
  "A4E3",                          /* 0xA4E3 */
  "A4E4",                          /* 0xA4E4 */
  "A4E5",                          /* 0xA4E5 */
  "A4E6",                          /* 0xA4E6 */
  "A4E7",                          /* 0xA4E7 */
  "A4E8",                          /* 0xA4E8 */
  "A4E9",                          /* 0xA4E9 */
  "A4EA",                          /* 0xA4EA */
  "A4EB",                          /* 0xA4EB */
  "A4EC",                          /* 0xA4EC */
  "A4ED",                          /* 0xA4ED */
  "A4EE",                          /* 0xA4EE */
  "A4EF",                          /* 0xA4EF */
  "A4F0",                          /* 0xA4F0 */
  "A4F1",                          /* 0xA4F1 */
  "A4F2",                          /* 0xA4F2 */
  "A4F3",                          /* 0xA4F3 */
  "A4F4",                          /* 0xA4F4 */
  "A4F5",                          /* 0xA4F5 */
  "A4F6",                          /* 0xA4F6 */
  "A4F7",                          /* 0xA4F7 */
  "A4F8",                          /* 0xA4F8 */
  "A4F9",                          /* 0xA4F9 */
  "A4FA",                          /* 0xA4FA */
  "A4FB",                          /* 0xA4FB */
  "ResourceStub",                 /* 0xA4FC */
  "A4FD",                          /* 0xA4FD */
  "TEFindWord",                   /* 0xA4FE */
  "TEFindLine",                   /* 0xA4FF */
  "PBOpen",                       /* 0xA500 */
  "PBClose",                      /* 0xA501 */
  "PBRead",                       /* 0xA502 */
  "PBWrite",                      /* 0xA503 */
  "PBControl",                    /* 0xA504 */
  "PBStatus",                     /* 0xA505 */
  "PBKillIO",                     /* 0xA506 */
  "PBGetVInfo",                   /* 0xA507 */
  "PBCreate",                     /* 0xA508 */
  "PBDelete",                     /* 0xA509 */
  "PBOpenRF",                     /* 0xA50A */
  "PBRename",                     /* 0xA50B */
  "PBGetFInfo",                   /* 0xA50C */
  "PBSetFInfo",                   /* 0xA50D */
  "PBUnmountVol",                 /* 0xA50E */
  "PBMountVol",                   /* 0xA50F */
  "PBAllocate",                   /* 0xA510 */
  "PBGetEOF",                     /* 0xA511 */
  "PBSetEOF",                     /* 0xA512 */
  "PBFlushVol",                   /* 0xA513 */
  "PBGetVol",                     /* 0xA514 */
  "PBSetVol",                     /* 0xA515 */
  "FInitQueue",                   /* 0xA516 */
  "PBEject",                      /* 0xA517 */
  "PBGetFPos",                    /* 0xA518 */
  "InitZone",                     /* 0xA519 */
  "GetZone",                      /* 0xA51A */
  "SetZone",                      /* 0xA51B */
  "FreeMem",                      /* 0xA51C */
  "MaxMem",                       /* 0xA51D */
  "NewPtr",                       /* 0xA51E */
  "DisposPtr",                    /* 0xA51F */
  "SetPtrSize",                   /* 0xA520 */
  "GetPtrSize",                   /* 0xA521 */
  "NewHandle",                    /* 0xA522 */
  "DisposHandle",                 /* 0xA523 */
  "SetHandleSize",                /* 0xA524 */
  "GetHandleSize",                /* 0xA525 */
  "HandleZone",                   /* 0xA526 */
  "ReallocHandle",                /* 0xA527 */
  "RecoverHandle",                /* 0xA528 */
  "HLock",                        /* 0xA529 */
  "HUnlock",                      /* 0xA52A */
  "EmptyHandle",                  /* 0xA52B */
  "InitApplZone",                 /* 0xA52C */
  "SetApplLimit",                 /* 0xA52D */
  "BlockMove",                    /* 0xA52E */
  "PostEvent",                    /* 0xA52F */
  "OSEventAvail",                 /* 0xA530 */
  "GetOSEvent",                   /* 0xA531 */
  "FlushEvents",                  /* 0xA532 */
  "VInstall",                     /* 0xA533 */
  "VRemove",                      /* 0xA534 */
  "PBOffLine",                    /* 0xA535 */
  "MoreMasters",                  /* 0xA536 */
  "A537",                          /* 0xA537 */
  "WriteParam",                   /* 0xA538 */
  "ReadDateTime",                 /* 0xA539 */
  "SetDateTime",                  /* 0xA53A */
  "Delay",                        /* 0xA53B */
  "EqualString",                  /* 0xA53C */
  "DrvrInstall",                  /* 0xA53D */
  "DrvrRemove",                   /* 0xA53E */
  "InitUtil",                     /* 0xA53F */
  "ResrvMem",                     /* 0xA540 */
  "PBSetFLock",                   /* 0xA541 */
  "PBRstFLock",                   /* 0xA542 */
  "PBSetFVers",                   /* 0xA543 */
  "PBSetFPos",                    /* 0xA544 */
  "PBFlushFile",                  /* 0xA545 */
  "GetTrapAddress",               /* 0xA546 */
  "SetTrapAddress",               /* 0xA547 */
  "PtrZone",                      /* 0xA548 */
  "HPurge",                       /* 0xA549 */
  "HNoPurge",                     /* 0xA54A */
  "SetGrowZone",                  /* 0xA54B */
  "CompactMem",                   /* 0xA54C */
  "PurgeMem",                     /* 0xA54D */
  "AddDrive",                     /* 0xA54E */
  "RDrvrInstall",                 /* 0xA54F */
  "RelString",                    /* 0xA550 */
  "IMVI_ReadXPRam",               /* 0xA551 */
  "IMVI_WriteXPRam",              /* 0xA552 */
  "A553",                          /* 0xA553 */
  "UprString",                    /* 0xA554 */
  "StripAddress",                 /* 0xA555 */
  "IMVI_LowerText",               /* 0xA556 */
  "SetApplBase",                  /* 0xA557 */
  "InsTime",                      /* 0xA558 */
  "RmvTime",                      /* 0xA559 */
  "PrimeTime",                    /* 0xA55A */
  "PowerOff",                     /* 0xA55B */
  "IMVI_MemoryDispatch",          /* 0xA55C */
  "SwapMMUMode",                  /* 0xA55D */
  "NMInstall",                    /* 0xA55E */
  "NMRemove",                     /* 0xA55F */
  "HFSDispatch",                  /* 0xA560 */
  "MaxBlock",                     /* 0xA561 */
  "PurgeSpace",                   /* 0xA562 */
  "MaxApplZone",                  /* 0xA563 */
  "MoveHHi",                      /* 0xA564 */
  "StackSpace",                   /* 0xA565 */
  "NewEmptyHandle",               /* 0xA566 */
  "HSetRBit",                     /* 0xA567 */
  "HClrRBit",                     /* 0xA568 */
  "HGetState",                    /* 0xA569 */
  "HSetState",                    /* 0xA56A */
  "A56B",                          /* 0xA56B */
  "InitFS",                       /* 0xA56C */
  "InitEvents",                   /* 0xA56D */
  "SlotManager",                  /* 0xA56E */
  "SlotVInstall",                 /* 0xA56F */
  "SlotVRemove",                  /* 0xA570 */
  "AttachVBL",                    /* 0xA571 */
  "DoVBLTask",                    /* 0xA572 */
  "A573",                          /* 0xA573 */
  "A574",                          /* 0xA574 */
  "DTInstall",                    /* 0xA575 */
  "SIntRemove",                   /* 0xA576 */
  "CountADBs",                    /* 0xA577 */
  "GetIndADB",                    /* 0xA578 */
  "GetADBInfo",                   /* 0xA579 */
  "SetADBInfo",                   /* 0xA57A */
  "ADBReInit",                    /* 0xA57B */
  "ADBOp",                        /* 0xA57C */
  "GetDefaultStartup",            /* 0xA57D */
  "SetDefaultStartup",            /* 0xA57E */
  "InternalWait",                 /* 0xA57F */
  "GetVideoDefault",              /* 0xA580 */
  "SetVideoDefault",              /* 0xA581 */
  "SIntInstall",                  /* 0xA582 */
  "SetOSDefault",                 /* 0xA583 */
  "GetOSDefault",                 /* 0xA584 */
  "IMVI_IdleUpdate",              /* 0xA585 */
  "IOPInfoAccess",                /* 0xA586 */
  "IOPMsgRequest",                /* 0xA587 */
  "IOPMoveData",                  /* 0xA588 */
  "SCSIAtomic",                   /* 0xA589 */
  "IMVI_SlpQInstall",             /* 0xA58A */
  "CommToolboxDispatch",          /* 0xA58B */
  "A58C",                          /* 0xA58C */
  "IMVI_DebugUtil",               /* 0xA58D */
  "A58E",                          /* 0xA58E */
  "IMVI_DeferUserFn",             /* 0xA58F */
  "SysEnvirons",                  /* 0xA590 */
  "IMVI_Translate24To32",         /* 0xA591 */
  "EgretDispatch",                /* 0xA592 */
  "Microseconds",                 /* 0xA593 */
  "A594",                          /* 0xA594 */
  "A595",                          /* 0xA595 */
  "A596",                          /* 0xA596 */
  "A597",                          /* 0xA597 */
  "HWPriv",                       /* 0xA598 */
  "A599",                          /* 0xA599 */
  "A59A",                          /* 0xA59A */
  "A59B",                          /* 0xA59B */
  "A59C",                          /* 0xA59C */
  "A59D",                          /* 0xA59D */
  "A59E",                          /* 0xA59E */
  "PowerDispatch",                /* 0xA59F */
  "A5A0",                          /* 0xA5A0 */
  "A5A1",                          /* 0xA5A1 */
  "A5A2",                          /* 0xA5A2 */
  "A5A3",                          /* 0xA5A3 */
  "HeapDispatch",                 /* 0xA5A4 */
  "A5A5",                          /* 0xA5A5 */
  "A5A6",                          /* 0xA5A6 */
  "A5A7",                          /* 0xA5A7 */
  "A5A8",                          /* 0xA5A8 */
  "A5A9",                          /* 0xA5A9 */
  "A5AA",                          /* 0xA5AA */
  "A5AB",                          /* 0xA5AB */
  "A5AC",                          /* 0xA5AC */
  "Gestalt",                      /* 0xA5AD */
  "VADBProc",                     /* 0xA5AE */
  "A5AF",                          /* 0xA5AF */
  "A5B0",                          /* 0xA5B0 */
  "A5B1",                          /* 0xA5B1 */
  "A5B2",                          /* 0xA5B2 */
  "A5B3",                          /* 0xA5B3 */
  "A5B4",                          /* 0xA5B4 */
  "A5B5",                          /* 0xA5B5 */
  "A5B6",                          /* 0xA5B6 */
  "A5B7",                          /* 0xA5B7 */
  "A5B8",                          /* 0xA5B8 */
  "A5B9",                          /* 0xA5B9 */
  "A5BA",                          /* 0xA5BA */
  "A5BB",                          /* 0xA5BB */
  "A5BC",                          /* 0xA5BC */
  "flushcache",                   /* 0xA5BD */
  "A5BE",                          /* 0xA5BE */
  "A5BF",                          /* 0xA5BF */
  "A5C0",                          /* 0xA5C0 */
  "A5C1",                          /* 0xA5C1 */
  "A5C2",                          /* 0xA5C2 */
  "A5C3",                          /* 0xA5C3 */
  "A5C4",                          /* 0xA5C4 */
  "A5C5",                          /* 0xA5C5 */
  "A5C6",                          /* 0xA5C6 */
  "A5C7",                          /* 0xA5C7 */
  "A5C8",                          /* 0xA5C8 */
  "A5C9",                          /* 0xA5C9 */
  "A5CA",                          /* 0xA5CA */
  "A5CB",                          /* 0xA5CB */
  "A5CC",                          /* 0xA5CC */
  "A5CD",                          /* 0xA5CD */
  "A5CE",                          /* 0xA5CE */
  "A5CF",                          /* 0xA5CF */
  "A5D0",                          /* 0xA5D0 */
  "A5D1",                          /* 0xA5D1 */
  "A5D2",                          /* 0xA5D2 */
  "A5D3",                          /* 0xA5D3 */
  "A5D4",                          /* 0xA5D4 */
  "A5D5",                          /* 0xA5D5 */
  "A5D6",                          /* 0xA5D6 */
  "A5D7",                          /* 0xA5D7 */
  "A5D8",                          /* 0xA5D8 */
  "A5D9",                          /* 0xA5D9 */
  "A5DA",                          /* 0xA5DA */
  "A5DB",                          /* 0xA5DB */
  "A5DC",                          /* 0xA5DC */
  "IMVI_PPC",                     /* 0xA5DD */
  "A5DE",                          /* 0xA5DE */
  "A5DF",                          /* 0xA5DF */
  "A5E0",                          /* 0xA5E0 */
  "A5E1",                          /* 0xA5E1 */
  "A5E2",                          /* 0xA5E2 */
  "A5E3",                          /* 0xA5E3 */
  "A5E4",                          /* 0xA5E4 */
  "A5E5",                          /* 0xA5E5 */
  "A5E6",                          /* 0xA5E6 */
  "A5E7",                          /* 0xA5E7 */
  "A5E8",                          /* 0xA5E8 */
  "A5E9",                          /* 0xA5E9 */
  "A5EA",                          /* 0xA5EA */
  "A5EB",                          /* 0xA5EB */
  "A5EC",                          /* 0xA5EC */
  "A5ED",                          /* 0xA5ED */
  "A5EE",                          /* 0xA5EE */
  "A5EF",                          /* 0xA5EF */
  "A5F0",                          /* 0xA5F0 */
  "A5F1",                          /* 0xA5F1 */
  "A5F2",                          /* 0xA5F2 */
  "A5F3",                          /* 0xA5F3 */
  "A5F4",                          /* 0xA5F4 */
  "A5F5",                          /* 0xA5F5 */
  "A5F6",                          /* 0xA5F6 */
  "A5F7",                          /* 0xA5F7 */
  "A5F8",                          /* 0xA5F8 */
  "A5F9",                          /* 0xA5F9 */
  "A5FA",                          /* 0xA5FA */
  "A5FB",                          /* 0xA5FB */
  "ResourceStub",                 /* 0xA5FC */
  "A5FD",                          /* 0xA5FD */
  "TEFindWord",                   /* 0xA5FE */
  "TEFindLine",                   /* 0xA5FF */
  "PBOpen",                       /* 0xA600 */
  "PBClose",                      /* 0xA601 */
  "PBRead",                       /* 0xA602 */
  "PBWrite",                      /* 0xA603 */
  "PBControl",                    /* 0xA604 */
  "PBStatus",                     /* 0xA605 */
  "PBKillIO",                     /* 0xA606 */
  "PBGetVInfo",                   /* 0xA607 */
  "PBCreate",                     /* 0xA608 */
  "PBDelete",                     /* 0xA609 */
  "PBOpenRF",                     /* 0xA60A */
  "PBRename",                     /* 0xA60B */
  "PBGetFInfo",                   /* 0xA60C */
  "PBSetFInfo",                   /* 0xA60D */
  "PBUnmountVol",                 /* 0xA60E */
  "PBMountVol",                   /* 0xA60F */
  "PBAllocate",                   /* 0xA610 */
  "PBGetEOF",                     /* 0xA611 */
  "PBSetEOF",                     /* 0xA612 */
  "PBFlushVol",                   /* 0xA613 */
  "PBGetVol",                     /* 0xA614 */
  "PBSetVol",                     /* 0xA615 */
  "FInitQueue",                   /* 0xA616 */
  "PBEject",                      /* 0xA617 */
  "PBGetFPos",                    /* 0xA618 */
  "InitZone",                     /* 0xA619 */
  "GetZone",                      /* 0xA61A */
  "SetZone",                      /* 0xA61B */
  "FreeMem",                      /* 0xA61C */
  "MaxMem",                       /* 0xA61D */
  "NewPtr",                       /* 0xA61E */
  "DisposPtr",                    /* 0xA61F */
  "SetPtrSize",                   /* 0xA620 */
  "GetPtrSize",                   /* 0xA621 */
  "NewHandle",                    /* 0xA622 */
  "DisposHandle",                 /* 0xA623 */
  "SetHandleSize",                /* 0xA624 */
  "GetHandleSize",                /* 0xA625 */
  "HandleZone",                   /* 0xA626 */
  "ReallocHandle",                /* 0xA627 */
  "RecoverHandle",                /* 0xA628 */
  "HLock",                        /* 0xA629 */
  "HUnlock",                      /* 0xA62A */
  "EmptyHandle",                  /* 0xA62B */
  "InitApplZone",                 /* 0xA62C */
  "SetApplLimit",                 /* 0xA62D */
  "BlockMove",                    /* 0xA62E */
  "PostEvent",                    /* 0xA62F */
  "OSEventAvail",                 /* 0xA630 */
  "GetOSEvent",                   /* 0xA631 */
  "FlushEvents",                  /* 0xA632 */
  "VInstall",                     /* 0xA633 */
  "VRemove",                      /* 0xA634 */
  "PBOffLine",                    /* 0xA635 */
  "MoreMasters",                  /* 0xA636 */
  "A637",                          /* 0xA637 */
  "WriteParam",                   /* 0xA638 */
  "ReadDateTime",                 /* 0xA639 */
  "SetDateTime",                  /* 0xA63A */
  "Delay",                        /* 0xA63B */
  "EqualString",                  /* 0xA63C */
  "DrvrInstall",                  /* 0xA63D */
  "DrvrRemove",                   /* 0xA63E */
  "InitUtil",                     /* 0xA63F */
  "ResrvMem",                     /* 0xA640 */
  "PBSetFLock",                   /* 0xA641 */
  "PBRstFLock",                   /* 0xA642 */
  "PBSetFVers",                   /* 0xA643 */
  "PBSetFPos",                    /* 0xA644 */
  "PBFlushFile",                  /* 0xA645 */
  "GetTrapAddress",               /* 0xA646 */
  "SetTrapAddress",               /* 0xA647 */
  "PtrZone",                      /* 0xA648 */
  "HPurge",                       /* 0xA649 */
  "HNoPurge",                     /* 0xA64A */
  "SetGrowZone",                  /* 0xA64B */
  "CompactMem",                   /* 0xA64C */
  "PurgeMem",                     /* 0xA64D */
  "AddDrive",                     /* 0xA64E */
  "RDrvrInstall",                 /* 0xA64F */
  "RelString",                    /* 0xA650 */
  "IMVI_ReadXPRam",               /* 0xA651 */
  "IMVI_WriteXPRam",              /* 0xA652 */
  "A653",                          /* 0xA653 */
  "UprString",                    /* 0xA654 */
  "StripAddress",                 /* 0xA655 */
  "IMVI_LowerText",               /* 0xA656 */
  "SetApplBase",                  /* 0xA657 */
  "InsTime",                      /* 0xA658 */
  "RmvTime",                      /* 0xA659 */
  "PrimeTime",                    /* 0xA65A */
  "PowerOff",                     /* 0xA65B */
  "IMVI_MemoryDispatch",          /* 0xA65C */
  "SwapMMUMode",                  /* 0xA65D */
  "NMInstall",                    /* 0xA65E */
  "NMRemove",                     /* 0xA65F */
  "HFSDispatch",                  /* 0xA660 */
  "MaxBlock",                     /* 0xA661 */
  "PurgeSpace",                   /* 0xA662 */
  "MaxApplZone",                  /* 0xA663 */
  "MoveHHi",                      /* 0xA664 */
  "StackSpace",                   /* 0xA665 */
  "NewEmptyHandle",               /* 0xA666 */
  "HSetRBit",                     /* 0xA667 */
  "HClrRBit",                     /* 0xA668 */
  "HGetState",                    /* 0xA669 */
  "HSetState",                    /* 0xA66A */
  "A66B",                          /* 0xA66B */
  "InitFS",                       /* 0xA66C */
  "InitEvents",                   /* 0xA66D */
  "SlotManager",                  /* 0xA66E */
  "SlotVInstall",                 /* 0xA66F */
  "SlotVRemove",                  /* 0xA670 */
  "AttachVBL",                    /* 0xA671 */
  "DoVBLTask",                    /* 0xA672 */
  "A673",                          /* 0xA673 */
  "A674",                          /* 0xA674 */
  "DTInstall",                    /* 0xA675 */
  "SIntRemove",                   /* 0xA676 */
  "CountADBs",                    /* 0xA677 */
  "GetIndADB",                    /* 0xA678 */
  "GetADBInfo",                   /* 0xA679 */
  "SetADBInfo",                   /* 0xA67A */
  "ADBReInit",                    /* 0xA67B */
  "ADBOp",                        /* 0xA67C */
  "GetDefaultStartup",            /* 0xA67D */
  "SetDefaultStartup",            /* 0xA67E */
  "InternalWait",                 /* 0xA67F */
  "GetVideoDefault",              /* 0xA680 */
  "SetVideoDefault",              /* 0xA681 */
  "SIntInstall",                  /* 0xA682 */
  "SetOSDefault",                 /* 0xA683 */
  "GetOSDefault",                 /* 0xA684 */
  "IMVI_IdleUpdate",              /* 0xA685 */
  "IOPInfoAccess",                /* 0xA686 */
  "IOPMsgRequest",                /* 0xA687 */
  "IOPMoveData",                  /* 0xA688 */
  "SCSIAtomic",                   /* 0xA689 */
  "IMVI_SlpQInstall",             /* 0xA68A */
  "CommToolboxDispatch",          /* 0xA68B */
  "A68C",                          /* 0xA68C */
  "IMVI_DebugUtil",               /* 0xA68D */
  "A68E",                          /* 0xA68E */
  "IMVI_DeferUserFn",             /* 0xA68F */
  "SysEnvirons",                  /* 0xA690 */
  "IMVI_Translate24To32",         /* 0xA691 */
  "EgretDispatch",                /* 0xA692 */
  "Microseconds",                 /* 0xA693 */
  "A694",                          /* 0xA694 */
  "A695",                          /* 0xA695 */
  "A696",                          /* 0xA696 */
  "A697",                          /* 0xA697 */
  "HWPriv",                       /* 0xA698 */
  "A699",                          /* 0xA699 */
  "A69A",                          /* 0xA69A */
  "A69B",                          /* 0xA69B */
  "A69C",                          /* 0xA69C */
  "A69D",                          /* 0xA69D */
  "A69E",                          /* 0xA69E */
  "PowerDispatch",                /* 0xA69F */
  "A6A0",                          /* 0xA6A0 */
  "A6A1",                          /* 0xA6A1 */
  "A6A2",                          /* 0xA6A2 */
  "A6A3",                          /* 0xA6A3 */
  "HeapDispatch",                 /* 0xA6A4 */
  "A6A5",                          /* 0xA6A5 */
  "A6A6",                          /* 0xA6A6 */
  "A6A7",                          /* 0xA6A7 */
  "A6A8",                          /* 0xA6A8 */
  "A6A9",                          /* 0xA6A9 */
  "A6AA",                          /* 0xA6AA */
  "A6AB",                          /* 0xA6AB */
  "A6AC",                          /* 0xA6AC */
  "Gestalt",                      /* 0xA6AD */
  "VADBProc",                     /* 0xA6AE */
  "A6AF",                          /* 0xA6AF */
  "A6B0",                          /* 0xA6B0 */
  "A6B1",                          /* 0xA6B1 */
  "A6B2",                          /* 0xA6B2 */
  "A6B3",                          /* 0xA6B3 */
  "A6B4",                          /* 0xA6B4 */
  "A6B5",                          /* 0xA6B5 */
  "A6B6",                          /* 0xA6B6 */
  "A6B7",                          /* 0xA6B7 */
  "A6B8",                          /* 0xA6B8 */
  "A6B9",                          /* 0xA6B9 */
  "A6BA",                          /* 0xA6BA */
  "A6BB",                          /* 0xA6BB */
  "A6BC",                          /* 0xA6BC */
  "flushcache",                   /* 0xA6BD */
  "A6BE",                          /* 0xA6BE */
  "A6BF",                          /* 0xA6BF */
  "A6C0",                          /* 0xA6C0 */
  "A6C1",                          /* 0xA6C1 */
  "A6C2",                          /* 0xA6C2 */
  "A6C3",                          /* 0xA6C3 */
  "A6C4",                          /* 0xA6C4 */
  "A6C5",                          /* 0xA6C5 */
  "A6C6",                          /* 0xA6C6 */
  "A6C7",                          /* 0xA6C7 */
  "A6C8",                          /* 0xA6C8 */
  "A6C9",                          /* 0xA6C9 */
  "A6CA",                          /* 0xA6CA */
  "A6CB",                          /* 0xA6CB */
  "A6CC",                          /* 0xA6CC */
  "A6CD",                          /* 0xA6CD */
  "A6CE",                          /* 0xA6CE */
  "A6CF",                          /* 0xA6CF */
  "A6D0",                          /* 0xA6D0 */
  "A6D1",                          /* 0xA6D1 */
  "A6D2",                          /* 0xA6D2 */
  "A6D3",                          /* 0xA6D3 */
  "A6D4",                          /* 0xA6D4 */
  "A6D5",                          /* 0xA6D5 */
  "A6D6",                          /* 0xA6D6 */
  "A6D7",                          /* 0xA6D7 */
  "A6D8",                          /* 0xA6D8 */
  "A6D9",                          /* 0xA6D9 */
  "A6DA",                          /* 0xA6DA */
  "A6DB",                          /* 0xA6DB */
  "A6DC",                          /* 0xA6DC */
  "IMVI_PPC",                     /* 0xA6DD */
  "A6DE",                          /* 0xA6DE */
  "A6DF",                          /* 0xA6DF */
  "A6E0",                          /* 0xA6E0 */
  "A6E1",                          /* 0xA6E1 */
  "A6E2",                          /* 0xA6E2 */
  "A6E3",                          /* 0xA6E3 */
  "A6E4",                          /* 0xA6E4 */
  "A6E5",                          /* 0xA6E5 */
  "A6E6",                          /* 0xA6E6 */
  "A6E7",                          /* 0xA6E7 */
  "A6E8",                          /* 0xA6E8 */
  "A6E9",                          /* 0xA6E9 */
  "A6EA",                          /* 0xA6EA */
  "A6EB",                          /* 0xA6EB */
  "A6EC",                          /* 0xA6EC */
  "A6ED",                          /* 0xA6ED */
  "A6EE",                          /* 0xA6EE */
  "A6EF",                          /* 0xA6EF */
  "A6F0",                          /* 0xA6F0 */
  "A6F1",                          /* 0xA6F1 */
  "A6F2",                          /* 0xA6F2 */
  "A6F3",                          /* 0xA6F3 */
  "A6F4",                          /* 0xA6F4 */
  "A6F5",                          /* 0xA6F5 */
  "A6F6",                          /* 0xA6F6 */
  "A6F7",                          /* 0xA6F7 */
  "A6F8",                          /* 0xA6F8 */
  "A6F9",                          /* 0xA6F9 */
  "A6FA",                          /* 0xA6FA */
  "A6FB",                          /* 0xA6FB */
  "ResourceStub",                 /* 0xA6FC */
  "A6FD",                          /* 0xA6FD */
  "TEFindWord",                   /* 0xA6FE */
  "TEFindLine",                   /* 0xA6FF */
  "PBOpen",                       /* 0xA700 */
  "PBClose",                      /* 0xA701 */
  "PBRead",                       /* 0xA702 */
  "PBWrite",                      /* 0xA703 */
  "PBControl",                    /* 0xA704 */
  "PBStatus",                     /* 0xA705 */
  "PBKillIO",                     /* 0xA706 */
  "PBGetVInfo",                   /* 0xA707 */
  "PBCreate",                     /* 0xA708 */
  "PBDelete",                     /* 0xA709 */
  "PBOpenRF",                     /* 0xA70A */
  "PBRename",                     /* 0xA70B */
  "PBGetFInfo",                   /* 0xA70C */
  "PBSetFInfo",                   /* 0xA70D */
  "PBUnmountVol",                 /* 0xA70E */
  "PBMountVol",                   /* 0xA70F */
  "PBAllocate",                   /* 0xA710 */
  "PBGetEOF",                     /* 0xA711 */
  "PBSetEOF",                     /* 0xA712 */
  "PBFlushVol",                   /* 0xA713 */
  "PBGetVol",                     /* 0xA714 */
  "PBSetVol",                     /* 0xA715 */
  "FInitQueue",                   /* 0xA716 */
  "PBEject",                      /* 0xA717 */
  "PBGetFPos",                    /* 0xA718 */
  "InitZone",                     /* 0xA719 */
  "GetZone",                      /* 0xA71A */
  "SetZone",                      /* 0xA71B */
  "FreeMem",                      /* 0xA71C */
  "MaxMem",                       /* 0xA71D */
  "NewPtr",                       /* 0xA71E */
  "DisposPtr",                    /* 0xA71F */
  "SetPtrSize",                   /* 0xA720 */
  "GetPtrSize",                   /* 0xA721 */
  "NewHandle",                    /* 0xA722 */
  "DisposHandle",                 /* 0xA723 */
  "SetHandleSize",                /* 0xA724 */
  "GetHandleSize",                /* 0xA725 */
  "HandleZone",                   /* 0xA726 */
  "ReallocHandle",                /* 0xA727 */
  "RecoverHandle",                /* 0xA728 */
  "HLock",                        /* 0xA729 */
  "HUnlock",                      /* 0xA72A */
  "EmptyHandle",                  /* 0xA72B */
  "InitApplZone",                 /* 0xA72C */
  "SetApplLimit",                 /* 0xA72D */
  "BlockMove",                    /* 0xA72E */
  "PostEvent",                    /* 0xA72F */
  "OSEventAvail",                 /* 0xA730 */
  "GetOSEvent",                   /* 0xA731 */
  "FlushEvents",                  /* 0xA732 */
  "VInstall",                     /* 0xA733 */
  "VRemove",                      /* 0xA734 */
  "PBOffLine",                    /* 0xA735 */
  "MoreMasters",                  /* 0xA736 */
  "A737",                          /* 0xA737 */
  "WriteParam",                   /* 0xA738 */
  "ReadDateTime",                 /* 0xA739 */
  "SetDateTime",                  /* 0xA73A */
  "Delay",                        /* 0xA73B */
  "EqualString",                  /* 0xA73C */
  "DrvrInstall",                  /* 0xA73D */
  "DrvrRemove",                   /* 0xA73E */
  "InitUtil",                     /* 0xA73F */
  "ResrvMem",                     /* 0xA740 */
  "PBSetFLock",                   /* 0xA741 */
  "PBRstFLock",                   /* 0xA742 */
  "PBSetFVers",                   /* 0xA743 */
  "PBSetFPos",                    /* 0xA744 */
  "PBFlushFile",                  /* 0xA745 */
  "GetTrapAddress",               /* 0xA746 */
  "SetTrapAddress",               /* 0xA747 */
  "PtrZone",                      /* 0xA748 */
  "HPurge",                       /* 0xA749 */
  "HNoPurge",                     /* 0xA74A */
  "SetGrowZone",                  /* 0xA74B */
  "CompactMem",                   /* 0xA74C */
  "PurgeMem",                     /* 0xA74D */
  "AddDrive",                     /* 0xA74E */
  "RDrvrInstall",                 /* 0xA74F */
  "RelString",                    /* 0xA750 */
  "IMVI_ReadXPRam",               /* 0xA751 */
  "IMVI_WriteXPRam",              /* 0xA752 */
  "A753",                          /* 0xA753 */
  "UprString",                    /* 0xA754 */
  "StripAddress",                 /* 0xA755 */
  "IMVI_LowerText",               /* 0xA756 */
  "SetApplBase",                  /* 0xA757 */
  "InsTime",                      /* 0xA758 */
  "RmvTime",                      /* 0xA759 */
  "PrimeTime",                    /* 0xA75A */
  "PowerOff",                     /* 0xA75B */
  "IMVI_MemoryDispatch",          /* 0xA75C */
  "SwapMMUMode",                  /* 0xA75D */
  "NMInstall",                    /* 0xA75E */
  "NMRemove",                     /* 0xA75F */
  "HFSDispatch",                  /* 0xA760 */
  "MaxBlock",                     /* 0xA761 */
  "PurgeSpace",                   /* 0xA762 */
  "MaxApplZone",                  /* 0xA763 */
  "MoveHHi",                      /* 0xA764 */
  "StackSpace",                   /* 0xA765 */
  "NewEmptyHandle",               /* 0xA766 */
  "HSetRBit",                     /* 0xA767 */
  "HClrRBit",                     /* 0xA768 */
  "HGetState",                    /* 0xA769 */
  "HSetState",                    /* 0xA76A */
  "A76B",                          /* 0xA76B */
  "InitFS",                       /* 0xA76C */
  "InitEvents",                   /* 0xA76D */
  "SlotManager",                  /* 0xA76E */
  "SlotVInstall",                 /* 0xA76F */
  "SlotVRemove",                  /* 0xA770 */
  "AttachVBL",                    /* 0xA771 */
  "DoVBLTask",                    /* 0xA772 */
  "A773",                          /* 0xA773 */
  "A774",                          /* 0xA774 */
  "DTInstall",                    /* 0xA775 */
  "SIntRemove",                   /* 0xA776 */
  "CountADBs",                    /* 0xA777 */
  "GetIndADB",                    /* 0xA778 */
  "GetADBInfo",                   /* 0xA779 */
  "SetADBInfo",                   /* 0xA77A */
  "ADBReInit",                    /* 0xA77B */
  "ADBOp",                        /* 0xA77C */
  "GetDefaultStartup",            /* 0xA77D */
  "SetDefaultStartup",            /* 0xA77E */
  "InternalWait",                 /* 0xA77F */
  "GetVideoDefault",              /* 0xA780 */
  "SetVideoDefault",              /* 0xA781 */
  "SIntInstall",                  /* 0xA782 */
  "SetOSDefault",                 /* 0xA783 */
  "GetOSDefault",                 /* 0xA784 */
  "IMVI_IdleUpdate",              /* 0xA785 */
  "IOPInfoAccess",                /* 0xA786 */
  "IOPMsgRequest",                /* 0xA787 */
  "IOPMoveData",                  /* 0xA788 */
  "SCSIAtomic",                   /* 0xA789 */
  "IMVI_SlpQInstall",             /* 0xA78A */
  "CommToolboxDispatch",          /* 0xA78B */
  "A78C",                          /* 0xA78C */
  "IMVI_DebugUtil",               /* 0xA78D */
  "A78E",                          /* 0xA78E */
  "IMVI_DeferUserFn",             /* 0xA78F */
  "SysEnvirons",                  /* 0xA790 */
  "IMVI_Translate24To32",         /* 0xA791 */
  "EgretDispatch",                /* 0xA792 */
  "Microseconds",                 /* 0xA793 */
  "A794",                          /* 0xA794 */
  "A795",                          /* 0xA795 */
  "A796",                          /* 0xA796 */
  "A797",                          /* 0xA797 */
  "HWPriv",                       /* 0xA798 */
  "A799",                          /* 0xA799 */
  "A79A",                          /* 0xA79A */
  "A79B",                          /* 0xA79B */
  "A79C",                          /* 0xA79C */
  "A79D",                          /* 0xA79D */
  "A79E",                          /* 0xA79E */
  "PowerDispatch",                /* 0xA79F */
  "A7A0",                          /* 0xA7A0 */
  "A7A1",                          /* 0xA7A1 */
  "A7A2",                          /* 0xA7A2 */
  "A7A3",                          /* 0xA7A3 */
  "HeapDispatch",                 /* 0xA7A4 */
  "A7A5",                          /* 0xA7A5 */
  "A7A6",                          /* 0xA7A6 */
  "A7A7",                          /* 0xA7A7 */
  "A7A8",                          /* 0xA7A8 */
  "A7A9",                          /* 0xA7A9 */
  "A7AA",                          /* 0xA7AA */
  "A7AB",                          /* 0xA7AB */
  "A7AC",                          /* 0xA7AC */
  "Gestalt",                      /* 0xA7AD */
  "VADBProc",                     /* 0xA7AE */
  "A7AF",                          /* 0xA7AF */
  "A7B0",                          /* 0xA7B0 */
  "A7B1",                          /* 0xA7B1 */
  "A7B2",                          /* 0xA7B2 */
  "A7B3",                          /* 0xA7B3 */
  "A7B4",                          /* 0xA7B4 */
  "A7B5",                          /* 0xA7B5 */
  "A7B6",                          /* 0xA7B6 */
  "A7B7",                          /* 0xA7B7 */
  "A7B8",                          /* 0xA7B8 */
  "A7B9",                          /* 0xA7B9 */
  "A7BA",                          /* 0xA7BA */
  "A7BB",                          /* 0xA7BB */
  "A7BC",                          /* 0xA7BC */
  "flushcache",                   /* 0xA7BD */
  "A7BE",                          /* 0xA7BE */
  "A7BF",                          /* 0xA7BF */
  "A7C0",                          /* 0xA7C0 */
  "A7C1",                          /* 0xA7C1 */
  "A7C2",                          /* 0xA7C2 */
  "A7C3",                          /* 0xA7C3 */
  "A7C4",                          /* 0xA7C4 */
  "A7C5",                          /* 0xA7C5 */
  "A7C6",                          /* 0xA7C6 */
  "A7C7",                          /* 0xA7C7 */
  "A7C8",                          /* 0xA7C8 */
  "A7C9",                          /* 0xA7C9 */
  "A7CA",                          /* 0xA7CA */
  "A7CB",                          /* 0xA7CB */
  "A7CC",                          /* 0xA7CC */
  "A7CD",                          /* 0xA7CD */
  "A7CE",                          /* 0xA7CE */
  "A7CF",                          /* 0xA7CF */
  "A7D0",                          /* 0xA7D0 */
  "A7D1",                          /* 0xA7D1 */
  "A7D2",                          /* 0xA7D2 */
  "A7D3",                          /* 0xA7D3 */
  "A7D4",                          /* 0xA7D4 */
  "A7D5",                          /* 0xA7D5 */
  "A7D6",                          /* 0xA7D6 */
  "A7D7",                          /* 0xA7D7 */
  "A7D8",                          /* 0xA7D8 */
  "A7D9",                          /* 0xA7D9 */
  "A7DA",                          /* 0xA7DA */
  "A7DB",                          /* 0xA7DB */
  "A7DC",                          /* 0xA7DC */
  "IMVI_PPC",                     /* 0xA7DD */
  "A7DE",                          /* 0xA7DE */
  "A7DF",                          /* 0xA7DF */
  "A7E0",                          /* 0xA7E0 */
  "A7E1",                          /* 0xA7E1 */
  "A7E2",                          /* 0xA7E2 */
  "A7E3",                          /* 0xA7E3 */
  "A7E4",                          /* 0xA7E4 */
  "A7E5",                          /* 0xA7E5 */
  "A7E6",                          /* 0xA7E6 */
  "A7E7",                          /* 0xA7E7 */
  "A7E8",                          /* 0xA7E8 */
  "A7E9",                          /* 0xA7E9 */
  "A7EA",                          /* 0xA7EA */
  "A7EB",                          /* 0xA7EB */
  "A7EC",                          /* 0xA7EC */
  "A7ED",                          /* 0xA7ED */
  "A7EE",                          /* 0xA7EE */
  "A7EF",                          /* 0xA7EF */
  "A7F0",                          /* 0xA7F0 */
  "A7F1",                          /* 0xA7F1 */
  "A7F2",                          /* 0xA7F2 */
  "A7F3",                          /* 0xA7F3 */
  "A7F4",                          /* 0xA7F4 */
  "A7F5",                          /* 0xA7F5 */
  "A7F6",                          /* 0xA7F6 */
  "A7F7",                          /* 0xA7F7 */
  "A7F8",                          /* 0xA7F8 */
  "A7F9",                          /* 0xA7F9 */
  "A7FA",                          /* 0xA7FA */
  "A7FB",                          /* 0xA7FB */
  "ResourceStub",                 /* 0xA7FC */
  "A7FD",                          /* 0xA7FD */
  "TEFindWord",                   /* 0xA7FE */
  "TEFindLine",                   /* 0xA7FF */
  "SoundDispatch",                /* 0xA800 */
  "SndDisposeChannel",            /* 0xA801 */
  "SndAddModifier",               /* 0xA802 */
  "SndDoCommand",                 /* 0xA803 */
  "SndDoImmediate",               /* 0xA804 */
  "SndPlay",                      /* 0xA805 */
  "SndControl",                   /* 0xA806 */
  "SndNewChannel",                /* 0xA807 */
  "InitProcMenu",                 /* 0xA808 */
  "GetCVariant",                  /* 0xA809 */
  "GetWVariant",                  /* 0xA80A */
  "PopUpMenuSelect",              /* 0xA80B */
  "RGetResource",                 /* 0xA80C */
  "Count1Resources",              /* 0xA80D */
  "Get1IndResource",              /* 0xA80E */
  "Get1IndType",                  /* 0xA80F */
  "Unique1ID",                    /* 0xA810 */
  "TESelView",                    /* 0xA811 */
  "TEPinScroll",                  /* 0xA812 */
  "TEAutoView",                   /* 0xA813 */
  "SetFractEnable",               /* 0xA814 */
  "SCSIDispatch",                 /* 0xA815 */
  "Pack8",                        /* 0xA816 */
  "CopyMask",                     /* 0xA817 */
  "FixAtan2",                     /* 0xA818 */
  "A819",                          /* 0xA819 */
  "HOpenResFile",                 /* 0xA81A */
  "IMVI_HCreateResFile",          /* 0xA81B */
  "Count1Types",                  /* 0xA81C */
  "InvalMenuBar",                 /* 0xA81D */
  "A81E",                          /* 0xA81E */
  "Get1Resource",                 /* 0xA81F */
  "Get1NamedResource",            /* 0xA820 */
  "MaxSizeRsrc",                  /* 0xA821 */
  "ResourceDispatch",             /* 0xA822 */
  "AliasDispatch",                /* 0xA823 */
  "A824",                          /* 0xA824 */
  "A825",                          /* 0xA825 */
  "InsMenuItem",                  /* 0xA826 */
  "HideDItem",                    /* 0xA827 */
  "ShowDItem",                    /* 0xA828 */
  "A829",                          /* 0xA829 */
  "A82A",                          /* 0xA82A */
  "Pack9",                        /* 0xA82B */
  "Pack10",                       /* 0xA82C */
  "Pack11",                       /* 0xA82D */
  "Pack12",                       /* 0xA82E */
  "Pack13",                       /* 0xA82F */
  "Pack14",                       /* 0xA830 */
  "Pack15",                       /* 0xA831 */
  "A832",                          /* 0xA832 */
  "A833",                          /* 0xA833 */
  "SetFScaleDisable",             /* 0xA834 */
  "FontMetrics",                  /* 0xA835 */
  "GetMaskTable",                 /* 0xA836 */
  "MeasureText",                  /* 0xA837 */
  "CalcMask",                     /* 0xA838 */
  "SeedFill",                     /* 0xA839 */
  "ZoomWindow",                   /* 0xA83A */
  "TrackBox",                     /* 0xA83B */
  "TEGetOffset",                  /* 0xA83C */
  "TEDispatch",                   /* 0xA83D */
  "TEStylNew",                    /* 0xA83E */
  "Long2Fix",                     /* 0xA83F */
  "Fix2Long",                     /* 0xA840 */
  "Fix2Frac",                     /* 0xA841 */
  "Frac2Fix",                     /* 0xA842 */
  "Fix2X",                        /* 0xA843 */
  "X2Fix",                        /* 0xA844 */
  "Frac2X",                       /* 0xA845 */
  "X2Frac",                       /* 0xA846 */
  "FracCos",                      /* 0xA847 */
  "FracSin",                      /* 0xA848 */
  "FracSqrt",                     /* 0xA849 */
  "FracMul",                      /* 0xA84A */
  "FracDiv",                      /* 0xA84B */
  "A84C",                          /* 0xA84C */
  "FixDiv",                       /* 0xA84D */
  "GetItemCmd",                   /* 0xA84E */
  "SetItemCmd",                   /* 0xA84F */
  "InitCursor",                   /* 0xA850 */
  "SetCursor",                    /* 0xA851 */
  "HideCursor",                   /* 0xA852 */
  "ShowCursor",                   /* 0xA853 */
  "FontDispatch",                 /* 0xA854 */
  "ShieldCursor",                 /* 0xA855 */
  "ObscureCursor",                /* 0xA856 */
  "A857",                          /* 0xA857 */
  "BitAnd",                       /* 0xA858 */
  "BitXor",                       /* 0xA859 */
  "BitNot",                       /* 0xA85A */
  "BitOr",                        /* 0xA85B */
  "BitShift",                     /* 0xA85C */
  "BitTst",                       /* 0xA85D */
  "BitSet",                       /* 0xA85E */
  "BitClr",                       /* 0xA85F */
  "WaitNextEvent",                /* 0xA860 */
  "Random",                       /* 0xA861 */
  "ForeColor",                    /* 0xA862 */
  "BackColor",                    /* 0xA863 */
  "ColorBit",                     /* 0xA864 */
  "GetPixel",                     /* 0xA865 */
  "StuffHex",                     /* 0xA866 */
  "LongMul",                      /* 0xA867 */
  "FixMul",                       /* 0xA868 */
  "FixRatio",                     /* 0xA869 */
  "HiWord",                       /* 0xA86A */
  "LoWord",                       /* 0xA86B */
  "FixRound",                     /* 0xA86C */
  "InitPort",                     /* 0xA86D */
  "InitGraf",                     /* 0xA86E */
  "OpenPort",                     /* 0xA86F */
  "LocalToGlobal",                /* 0xA870 */
  "GlobalToLocal",                /* 0xA871 */
  "GrafDevice",                   /* 0xA872 */
  "SetPort",                      /* 0xA873 */
  "GetPort",                      /* 0xA874 */
  "SetPortBits",                  /* 0xA875 */
  "PortSize",                     /* 0xA876 */
  "MovePortTo",                   /* 0xA877 */
  "SetOrigin",                    /* 0xA878 */
  "SetClip",                      /* 0xA879 */
  "GetClip",                      /* 0xA87A */
  "ClipRect",                     /* 0xA87B */
  "BackPat",                      /* 0xA87C */
  "ClosePort",                    /* 0xA87D */
  "AddPt",                        /* 0xA87E */
  "SubPt",                        /* 0xA87F */
  "SetPt",                        /* 0xA880 */
  "EqualPt",                      /* 0xA881 */
  "StdText",                      /* 0xA882 */
  "DrawChar",                     /* 0xA883 */
  "DrawString",                   /* 0xA884 */
  "DrawText",                     /* 0xA885 */
  "TextWidth",                    /* 0xA886 */
  "TextFont",                     /* 0xA887 */
  "TextFace",                     /* 0xA888 */
  "TextMode",                     /* 0xA889 */
  "TextSize",                     /* 0xA88A */
  "GetFontInfo",                  /* 0xA88B */
  "StringWidth",                  /* 0xA88C */
  "CharWidth",                    /* 0xA88D */
  "SpaceExtra",                   /* 0xA88E */
  "OSDispatch",                   /* 0xA88F */
  "StdLine",                      /* 0xA890 */
  "LineTo",                       /* 0xA891 */
  "Line",                         /* 0xA892 */
  "MoveTo",                       /* 0xA893 */
  "Move",                         /* 0xA894 */
  "ShutDown",                     /* 0xA895 */
  "HidePen",                      /* 0xA896 */
  "ShowPen",                      /* 0xA897 */
  "GetPenState",                  /* 0xA898 */
  "SetPenState",                  /* 0xA899 */
  "GetPen",                       /* 0xA89A */
  "PenSize",                      /* 0xA89B */
  "PenMode",                      /* 0xA89C */
  "PenPat",                       /* 0xA89D */
  "PenNormal",                    /* 0xA89E */
  "Unimplemented",                /* 0xA89F */
  "StdRect",                      /* 0xA8A0 */
  "FrameRect",                    /* 0xA8A1 */
  "PaintRect",                    /* 0xA8A2 */
  "EraseRect",                    /* 0xA8A3 */
  "InvertRect",                   /* 0xA8A4 */
  "FillRect",                     /* 0xA8A5 */
  "EqualRect",                    /* 0xA8A6 */
  "SetRect",                      /* 0xA8A7 */
  "OffsetRect",                   /* 0xA8A8 */
  "InsetRect",                    /* 0xA8A9 */
  "SectRect",                     /* 0xA8AA */
  "UnionRect",                    /* 0xA8AB */
  "Pt2Rect",                      /* 0xA8AC */
  "PtInRect",                     /* 0xA8AD */
  "EmptyRect",                    /* 0xA8AE */
  "StdRRect",                     /* 0xA8AF */
  "FrameRoundRect",               /* 0xA8B0 */
  "PaintRoundRect",               /* 0xA8B1 */
  "EraseRoundRect",               /* 0xA8B2 */
  "InvertRoundRect",              /* 0xA8B3 */
  "FillRoundRect",                /* 0xA8B4 */
  "ScriptUtil",                   /* 0xA8B5 */
  "StdOval",                      /* 0xA8B6 */
  "FrameOval",                    /* 0xA8B7 */
  "PaintOval",                    /* 0xA8B8 */
  "EraseOval",                    /* 0xA8B9 */
  "InvertOval",                   /* 0xA8BA */
  "FillOval",                     /* 0xA8BB */
  "SlopeFromAngle",               /* 0xA8BC */
  "StdArc",                       /* 0xA8BD */
  "FrameArc",                     /* 0xA8BE */
  "PaintArc",                     /* 0xA8BF */
  "EraseArc",                     /* 0xA8C0 */
  "InvertArc",                    /* 0xA8C1 */
  "FillArc",                      /* 0xA8C2 */
  "PtToAngle",                    /* 0xA8C3 */
  "AngleFromSlope",               /* 0xA8C4 */
  "StdPoly",                      /* 0xA8C5 */
  "FramePoly",                    /* 0xA8C6 */
  "PaintPoly",                    /* 0xA8C7 */
  "ErasePoly",                    /* 0xA8C8 */
  "InvertPoly",                   /* 0xA8C9 */
  "FillPoly",                     /* 0xA8CA */
  "OpenPoly",                     /* 0xA8CB */
  "ClosePoly",                    /* 0xA8CC */
  "KillPoly",                     /* 0xA8CD */
  "OffsetPoly",                   /* 0xA8CE */
  "PackBits",                     /* 0xA8CF */
  "UnpackBits",                   /* 0xA8D0 */
  "StdRgn",                       /* 0xA8D1 */
  "FrameRgn",                     /* 0xA8D2 */
  "PaintRgn",                     /* 0xA8D3 */
  "EraseRgn",                     /* 0xA8D4 */
  "InvertRgn",                    /* 0xA8D5 */
  "FillRgn",                      /* 0xA8D6 */
  "BitMapToRegion",               /* 0xA8D7 */
  "NewRgn",                       /* 0xA8D8 */
  "DisposeRgn",                   /* 0xA8D9 */
  "OpenRgn",                      /* 0xA8DA */
  "CloseRgn",                     /* 0xA8DB */
  "CopyRgn",                      /* 0xA8DC */
  "SetEmptyRgn",                  /* 0xA8DD */
  "SetRectRgn",                   /* 0xA8DE */
  "RectRgn",                      /* 0xA8DF */
  "OffsetRgn",                    /* 0xA8E0 */
  "InsetRgn",                     /* 0xA8E1 */
  "EmptyRgn",                     /* 0xA8E2 */
  "EqualRgn",                     /* 0xA8E3 */
  "SectRgn",                      /* 0xA8E4 */
  "UnionRgn",                     /* 0xA8E5 */
  "DiffRgn",                      /* 0xA8E6 */
  "XorRgn",                       /* 0xA8E7 */
  "PtInRgn",                      /* 0xA8E8 */
  "RectInRgn",                    /* 0xA8E9 */
  "SetStdProcs",                  /* 0xA8EA */
  "StdBits",                      /* 0xA8EB */
  "CopyBits",                     /* 0xA8EC */
  "StdTxMeas",                    /* 0xA8ED */
  "StdGetPic",                    /* 0xA8EE */
  "ScrollRect",                   /* 0xA8EF */
  "StdPutPic",                    /* 0xA8F0 */
  "StdComment",                   /* 0xA8F1 */
  "PicComment",                   /* 0xA8F2 */
  "OpenPicture",                  /* 0xA8F3 */
  "ClosePicture",                 /* 0xA8F4 */
  "KillPicture",                  /* 0xA8F5 */
  "DrawPicture",                  /* 0xA8F6 */
  "A8F7",                          /* 0xA8F7 */
  "ScalePt",                      /* 0xA8F8 */
  "MapPt",                        /* 0xA8F9 */
  "MapRect",                      /* 0xA8FA */
  "MapRgn",                       /* 0xA8FB */
  "MapPoly",                      /* 0xA8FC */
  "PrGlue",                       /* 0xA8FD */
  "InitFonts",                    /* 0xA8FE */
  "GetFontName",                  /* 0xA8FF */
  "GetFNum",                      /* 0xA900 */
  "FMSwapFont",                   /* 0xA901 */
  "RealFont",                     /* 0xA902 */
  "SetFontLock",                  /* 0xA903 */
  "DrawGrowIcon",                 /* 0xA904 */
  "DragGrayRgn",                  /* 0xA905 */
  "NewString",                    /* 0xA906 */
  "SetString",                    /* 0xA907 */
  "ShowHide",                     /* 0xA908 */
  "CalcVis",                      /* 0xA909 */
  "CalcVisBehind",                /* 0xA90A */
  "ClipAbove",                    /* 0xA90B */
  "PaintOne",                     /* 0xA90C */
  "PaintBehind",                  /* 0xA90D */
  "SaveOld",                      /* 0xA90E */
  "DrawNew",                      /* 0xA90F */
  "GetWMgrPort",                  /* 0xA910 */
  "CheckUpdate",                  /* 0xA911 */
  "InitWindows",                  /* 0xA912 */
  "NewWindow",                    /* 0xA913 */
  "DisposeWindow",                /* 0xA914 */
  "ShowWindow",                   /* 0xA915 */
  "HideWindow",                   /* 0xA916 */
  "GetWRefCon",                   /* 0xA917 */
  "SetWRefCon",                   /* 0xA918 */
  "GetWTitle",                    /* 0xA919 */
  "SetWTitle",                    /* 0xA91A */
  "MoveWindow",                   /* 0xA91B */
  "HiliteWindow",                 /* 0xA91C */
  "SizeWindow",                   /* 0xA91D */
  "TrackGoAway",                  /* 0xA91E */
  "SelectWindow",                 /* 0xA91F */
  "BringToFront",                 /* 0xA920 */
  "SendBehind",                   /* 0xA921 */
  "BeginUpdate",                  /* 0xA922 */
  "EndUpdate",                    /* 0xA923 */
  "FrontWindow",                  /* 0xA924 */
  "DragWindow",                   /* 0xA925 */
  "DragTheRgn",                   /* 0xA926 */
  "InvalRgn",                     /* 0xA927 */
  "InvalRect",                    /* 0xA928 */
  "ValidRgn",                     /* 0xA929 */
  "ValidRect",                    /* 0xA92A */
  "GrowWindow",                   /* 0xA92B */
  "FindWindow",                   /* 0xA92C */
  "CloseWindow",                  /* 0xA92D */
  "SetWindowPic",                 /* 0xA92E */
  "GetWindowPic",                 /* 0xA92F */
  "InitMenus",                    /* 0xA930 */
  "NewMenu",                      /* 0xA931 */
  "DisposeMenu",                  /* 0xA932 */
  "AppendMenu",                   /* 0xA933 */
  "ClearMenuBar",                 /* 0xA934 */
  "InsertMenu",                   /* 0xA935 */
  "DeleteMenu",                   /* 0xA936 */
  "DrawMenuBar",                  /* 0xA937 */
  "HiliteMenu",                   /* 0xA938 */
  "EnableItem",                   /* 0xA939 */
  "DisableItem",                  /* 0xA93A */
  "GetMenuBar",                   /* 0xA93B */
  "SetMenuBar",                   /* 0xA93C */
  "MenuSelect",                   /* 0xA93D */
  "MenuKey",                      /* 0xA93E */
  "GetItemIcon",                  /* 0xA93F */
  "SetItemIcon",                  /* 0xA940 */
  "GetItemStyle",                 /* 0xA941 */
  "SetItemStyle",                 /* 0xA942 */
  "GetItemMark",                  /* 0xA943 */
  "SetItemMark",                  /* 0xA944 */
  "CheckItem",                    /* 0xA945 */
  "GetItem",                      /* 0xA946 */
  "SetItem",                      /* 0xA947 */
  "CalcMenuSize",                 /* 0xA948 */
  "GetMHandle",                   /* 0xA949 */
  "SetMenuFlash",                 /* 0xA94A */
  "PlotIcon",                     /* 0xA94B */
  "FlashMenuBar",                 /* 0xA94C */
  "AddResMenu",                   /* 0xA94D */
  "PinRect",                      /* 0xA94E */
  "DeltaPoint",                   /* 0xA94F */
  "CountMItems",                  /* 0xA950 */
  "InsertResMenu",                /* 0xA951 */
  "DelMenuItem",                  /* 0xA952 */
  "UpdtControl",                  /* 0xA953 */
  "NewControl",                   /* 0xA954 */
  "DisposeControl",               /* 0xA955 */
  "KillControls",                 /* 0xA956 */
  "ShowControl",                  /* 0xA957 */
  "HideControl",                  /* 0xA958 */
  "MoveControl",                  /* 0xA959 */
  "GetCRefCon",                   /* 0xA95A */
  "SetCRefCon",                   /* 0xA95B */
  "SizeControl",                  /* 0xA95C */
  "HiliteControl",                /* 0xA95D */
  "GetCTitle",                    /* 0xA95E */
  "SetCTitle",                    /* 0xA95F */
  "GetCtlValue",                  /* 0xA960 */
  "GetCtlMin",                    /* 0xA961 */
  "GetCtlMax",                    /* 0xA962 */
  "SetCtlValue",                  /* 0xA963 */
  "SetCtlMin",                    /* 0xA964 */
  "SetCtlMax",                    /* 0xA965 */
  "TestControl",                  /* 0xA966 */
  "DragControl",                  /* 0xA967 */
  "TrackControl",                 /* 0xA968 */
  "DrawControls",                 /* 0xA969 */
  "GetCtlAction",                 /* 0xA96A */
  "SetCtlAction",                 /* 0xA96B */
  "FindControl",                  /* 0xA96C */
  "Draw1Control",                 /* 0xA96D */
  "Dequeue",                      /* 0xA96E */
  "Enqueue",                      /* 0xA96F */
  "GetNextEvent",                 /* 0xA970 */
  "EventAvail",                   /* 0xA971 */
  "GetMouse",                     /* 0xA972 */
  "StillDown",                    /* 0xA973 */
  "Button",                       /* 0xA974 */
  "TickCount",                    /* 0xA975 */
  "GetKeys",                      /* 0xA976 */
  "WaitMouseUp",                  /* 0xA977 */
  "UpdtDialog",                   /* 0xA978 */
  "CouldDialog",                  /* 0xA979 */
  "FreeDialog",                   /* 0xA97A */
  "InitDialogs",                  /* 0xA97B */
  "GetNewDialog",                 /* 0xA97C */
  "NewDialog",                    /* 0xA97D */
  "SelIText",                     /* 0xA97E */
  "IsDialogEvent",                /* 0xA97F */
  "DialogSelect",                 /* 0xA980 */
  "DrawDialog",                   /* 0xA981 */
  "CloseDialog",                  /* 0xA982 */
  "DisposDialog",                 /* 0xA983 */
  "FindDItem",                    /* 0xA984 */
  "Alert",                        /* 0xA985 */
  "StopAlert",                    /* 0xA986 */
  "NoteAlert",                    /* 0xA987 */
  "CautionAlert",                 /* 0xA988 */
  "CouldAlert",                   /* 0xA989 */
  "FreeAlert",                    /* 0xA98A */
  "ParamText",                    /* 0xA98B */
  "ErrorSound",                   /* 0xA98C */
  "GetDItem",                     /* 0xA98D */
  "SetDItem",                     /* 0xA98E */
  "SetIText",                     /* 0xA98F */
  "GetIText",                     /* 0xA990 */
  "ModalDialog",                  /* 0xA991 */
  "DetachResource",               /* 0xA992 */
  "SetResPurge",                  /* 0xA993 */
  "CurResFile",                   /* 0xA994 */
  "InitResources",                /* 0xA995 */
  "RsrcZoneInit",                 /* 0xA996 */
  "OpenResFile",                  /* 0xA997 */
  "UseResFile",                   /* 0xA998 */
  "UpdateResFile",                /* 0xA999 */
  "CloseResFile",                 /* 0xA99A */
  "SetResLoad",                   /* 0xA99B */
  "CountResources",               /* 0xA99C */
  "GetIndResource",               /* 0xA99D */
  "CountTypes",                   /* 0xA99E */
  "GetIndType",                   /* 0xA99F */
  "GetResource",                  /* 0xA9A0 */
  "GetNamedResource",             /* 0xA9A1 */
  "LoadResource",                 /* 0xA9A2 */
  "ReleaseResource",              /* 0xA9A3 */
  "HomeResFile",                  /* 0xA9A4 */
  "SizeResource",                 /* 0xA9A5 */
  "GetResAttrs",                  /* 0xA9A6 */
  "SetResAttrs",                  /* 0xA9A7 */
  "GetResInfo",                   /* 0xA9A8 */
  "SetResInfo",                   /* 0xA9A9 */
  "ChangedResource",              /* 0xA9AA */
  "AddResource",                  /* 0xA9AB */
  "AddReference",                 /* 0xA9AC */
  "RmveResource",                 /* 0xA9AD */
  "RmveReference",                /* 0xA9AE */
  "ResError",                     /* 0xA9AF */
  "WriteResource",                /* 0xA9B0 */
  "CreateResFile",                /* 0xA9B1 */
  "SystemEvent",                  /* 0xA9B2 */
  "SystemClick",                  /* 0xA9B3 */
  "SystemTask",                   /* 0xA9B4 */
  "SystemMenu",                   /* 0xA9B5 */
  "OpenDeskAcc",                  /* 0xA9B6 */
  "CloseDeskAcc",                 /* 0xA9B7 */
  "GetPattern",                   /* 0xA9B8 */
  "GetCursor",                    /* 0xA9B9 */
  "GetString",                    /* 0xA9BA */
  "GetIcon",                      /* 0xA9BB */
  "GetPicture",                   /* 0xA9BC */
  "GetNewWindow",                 /* 0xA9BD */
  "GetNewControl",                /* 0xA9BE */
  "GetMenu",                      /* 0xA9BF */
  "GetNewMBar",                   /* 0xA9C0 */
  "UniqueID",                     /* 0xA9C1 */
  "SystemEdit",                   /* 0xA9C2 */
  "KeyTrans",                     /* 0xA9C3 */
  "OpenRFPerm",                   /* 0xA9C4 */
  "RsrcMapEntry",                 /* 0xA9C5 */
  "Secs2Date",                    /* 0xA9C6 */
  "Date2Secs",                    /* 0xA9C7 */
  "SysBeep",                      /* 0xA9C8 */
  "SysError",                     /* 0xA9C9 */
  "PutIcon",                      /* 0xA9CA */
  "TEGetText",                    /* 0xA9CB */
  "TEInit",                       /* 0xA9CC */
  "TEDispose",                    /* 0xA9CD */
  "TextBox",                      /* 0xA9CE */
  "TESetText",                    /* 0xA9CF */
  "TECalText",                    /* 0xA9D0 */
  "TESetSelect",                  /* 0xA9D1 */
  "TENew",                        /* 0xA9D2 */
  "TEUpdate",                     /* 0xA9D3 */
  "TEClick",                      /* 0xA9D4 */
  "TECopy",                       /* 0xA9D5 */
  "TECut",                        /* 0xA9D6 */
  "TEDelete",                     /* 0xA9D7 */
  "TEActivate",                   /* 0xA9D8 */
  "TEDeactivate",                 /* 0xA9D9 */
  "TEIdle",                       /* 0xA9DA */
  "TEPaste",                      /* 0xA9DB */
  "TEKey",                        /* 0xA9DC */
  "TEScroll",                     /* 0xA9DD */
  "TEInsert",                     /* 0xA9DE */
  "TESetJust",                    /* 0xA9DF */
  "Munger",                       /* 0xA9E0 */
  "HandToHand",                   /* 0xA9E1 */
  "PtrToXHand",                   /* 0xA9E2 */
  "PtrToHand",                    /* 0xA9E3 */
  "HandAndHand",                  /* 0xA9E4 */
  "InitPack",                     /* 0xA9E5 */
  "InitAllPacks",                 /* 0xA9E6 */
  "Pack0",                        /* 0xA9E7 */
  "Pack1",                        /* 0xA9E8 */
  "Pack2",                        /* 0xA9E9 */
  "Pack3",                        /* 0xA9EA */
  "Pack4",                        /* 0xA9EB */
  "Pack5",                        /* 0xA9EC */
  "Pack6",                        /* 0xA9ED */
  "Pack7",                        /* 0xA9EE */
  "PtrAndHand",                   /* 0xA9EF */
  "LoadSeg",                      /* 0xA9F0 */
  "UnloadSeg",                    /* 0xA9F1 */
  "Launch",                       /* 0xA9F2 */
  "Chain",                        /* 0xA9F3 */
  "ExitToShell",                  /* 0xA9F4 */
  "GetAppParms",                  /* 0xA9F5 */
  "GetResFileAttrs",              /* 0xA9F6 */
  "SetResFileAttrs",              /* 0xA9F7 */
  "MethodDispatch",               /* 0xA9F8 */
  "InfoScrap",                    /* 0xA9F9 */
  "UnloadScrap",                  /* 0xA9FA */
  "LoadScrap",                    /* 0xA9FB */
  "ZeroScrap",                    /* 0xA9FC */
  "GetScrap",                     /* 0xA9FD */
  "PutScrap",                     /* 0xA9FE */
  "Debugger",                     /* 0xA9FF */
  "OpenCPort",                    /* 0xAA00 */
  "InitCPort",                    /* 0xAA01 */
  "CloseCPort",                   /* 0xAA02 */
  "NewPixMap",                    /* 0xAA03 */
  "DisposPixMap",                 /* 0xAA04 */
  "CopyPixMap",                   /* 0xAA05 */
  "SetPortPix",                   /* 0xAA06 */
  "NewPixPat",                    /* 0xAA07 */
  "DisposPixPat",                 /* 0xAA08 */
  "CopyPixPat",                   /* 0xAA09 */
  "PenPixPat",                    /* 0xAA0A */
  "BackPixPat",                   /* 0xAA0B */
  "GetPixPat",                    /* 0xAA0C */
  "MakeRGBPat",                   /* 0xAA0D */
  "FillCRect",                    /* 0xAA0E */
  "FillCOval",                    /* 0xAA0F */
  "FillCRoundRect",               /* 0xAA10 */
  "FillCArc",                     /* 0xAA11 */
  "FillCRgn",                     /* 0xAA12 */
  "FillCPoly",                    /* 0xAA13 */
  "RGBForeColor",                 /* 0xAA14 */
  "RGBBackColor",                 /* 0xAA15 */
  "SetCPixel",                    /* 0xAA16 */
  "GetCPixel",                    /* 0xAA17 */
  "GetCTable",                    /* 0xAA18 */
  "GetForeColor",                 /* 0xAA19 */
  "GetBackColor",                 /* 0xAA1A */
  "GetCCursor",                   /* 0xAA1B */
  "SetCCursor",                   /* 0xAA1C */
  "AllocCursor",                  /* 0xAA1D */
  "GetCIcon",                     /* 0xAA1E */
  "PlotCIcon",                    /* 0xAA1F */
  "IMVI_OpenCPicture",            /* 0xAA20 */
  "OpColor",                      /* 0xAA21 */
  "HiliteColor",                  /* 0xAA22 */
  "CharExtra",                    /* 0xAA23 */
  "DisposCTable",                 /* 0xAA24 */
  "DisposeCIcon",                 /* 0xAA25 */
  "DisposCCursor",                /* 0xAA26 */
  "GetMaxDevice",                 /* 0xAA27 */
  "GetCTSeed",                    /* 0xAA28 */
  "GetDeviceList",                /* 0xAA29 */
  "GetMainDevice",                /* 0xAA2A */
  "GetNextDevice",                /* 0xAA2B */
  "TestDeviceAttribute",          /* 0xAA2C */
  "SetDeviceAttribute",           /* 0xAA2D */
  "InitGDevice",                  /* 0xAA2E */
  "NewGDevice",                   /* 0xAA2F */
  "DisposGDevice",                /* 0xAA30 */
  "SetGDevice",                   /* 0xAA31 */
  "GetGDevice",                   /* 0xAA32 */
  "Color2Index",                  /* 0xAA33 */
  "Index2Color",                  /* 0xAA34 */
  "InvertColor",                  /* 0xAA35 */
  "RealColor",                    /* 0xAA36 */
  "GetSubTable",                  /* 0xAA37 */
  "AA38",                          /* 0xAA38 */
  "MakeITable",                   /* 0xAA39 */
  "AddSearch",                    /* 0xAA3A */
  "AddComp",                      /* 0xAA3B */
  "SetClientID",                  /* 0xAA3C */
  "ProtectEntry",                 /* 0xAA3D */
  "ReserveEntry",                 /* 0xAA3E */
  "SetEntries",                   /* 0xAA3F */
  "QDError",                      /* 0xAA40 */
  "SetWinColor",                  /* 0xAA41 */
  "GetAuxWin",                    /* 0xAA42 */
  "SetCtlColor",                  /* 0xAA43 */
  "GetAuxCtl",                    /* 0xAA44 */
  "NewCWindow",                   /* 0xAA45 */
  "GetNewCWindow",                /* 0xAA46 */
  "SetDeskCPat",                  /* 0xAA47 */
  "GetCWMgrPort",                 /* 0xAA48 */
  "SaveEntries",                  /* 0xAA49 */
  "RestoreEntries",               /* 0xAA4A */
  "NewCDialog",                   /* 0xAA4B */
  "DelSearch",                    /* 0xAA4C */
  "DelComp",                      /* 0xAA4D */
  "SetStdCProcs",                 /* 0xAA4E */
  "CalcCMask",                    /* 0xAA4F */
  "SeedCFill",                    /* 0xAA50 */
  "IMVI_CopyDeepMask",            /* 0xAA51 */
  "HighLevelFSDispatch",          /* 0xAA52 */
  "AA53",                          /* 0xAA53 */
  "AA54",                          /* 0xAA54 */
  "AA55",                          /* 0xAA55 */
  "AA56",                          /* 0xAA56 */
  "DockingDispatch",              /* 0xAA57 */
  "AA58",                          /* 0xAA58 */
  "AA59",                          /* 0xAA59 */
  "AA5A",                          /* 0xAA5A */
  "AA5B",                          /* 0xAA5B */
  "AA5C",                          /* 0xAA5C */
  "AA5D",                          /* 0xAA5D */
  "AA5E",                          /* 0xAA5E */
  "AA5F",                          /* 0xAA5F */
  "DelMCEntries",                 /* 0xAA60 */
  "GetMCInfo",                    /* 0xAA61 */
  "SetMCInfo",                    /* 0xAA62 */
  "DispMCInfo",                   /* 0xAA63 */
  "GetMCEntry",                   /* 0xAA64 */
  "SetMCEntries",                 /* 0xAA65 */
  "MenuChoice",                   /* 0xAA66 */
  "AA67",                          /* 0xAA67 */
  "DialogDispatch",               /* 0xAA68 */
  "AA69",                          /* 0xAA69 */
  "AA6A",                          /* 0xAA6A */
  "AA6B",                          /* 0xAA6B */
  "AA6C",                          /* 0xAA6C */
  "AA6D",                          /* 0xAA6D */
  "AA6E",                          /* 0xAA6E */
  "AA6F",                          /* 0xAA6F */
  "AA70",                          /* 0xAA70 */
  "AA71",                          /* 0xAA71 */
  "AA72",                          /* 0xAA72 */
  "AA73",                          /* 0xAA73 */
  "AA74",                          /* 0xAA74 */
  "AA75",                          /* 0xAA75 */
  "AA76",                          /* 0xAA76 */
  "AA77",                          /* 0xAA77 */
  "AA78",                          /* 0xAA78 */
  "AA79",                          /* 0xAA79 */
  "AA7A",                          /* 0xAA7A */
  "AA7B",                          /* 0xAA7B */
  "AA7C",                          /* 0xAA7C */
  "AA7D",                          /* 0xAA7D */
  "AA7E",                          /* 0xAA7E */
  "AA7F",                          /* 0xAA7F */
  "AA80",                          /* 0xAA80 */
  "AA81",                          /* 0xAA81 */
  "AA82",                          /* 0xAA82 */
  "AA83",                          /* 0xAA83 */
  "AA84",                          /* 0xAA84 */
  "AA85",                          /* 0xAA85 */
  "AA86",                          /* 0xAA86 */
  "AA87",                          /* 0xAA87 */
  "AA88",                          /* 0xAA88 */
  "AA89",                          /* 0xAA89 */
  "AA8A",                          /* 0xAA8A */
  "AA8B",                          /* 0xAA8B */
  "AA8C",                          /* 0xAA8C */
  "AA8D",                          /* 0xAA8D */
  "AA8E",                          /* 0xAA8E */
  "AA8F",                          /* 0xAA8F */
  "InitPalettes",                 /* 0xAA90 */
  "NewPalette",                   /* 0xAA91 */
  "GetNewPalette",                /* 0xAA92 */
  "DisposePalette",               /* 0xAA93 */
  "ActivatePalette",              /* 0xAA94 */
  "NSetPalette",                   /* 0xAA95 */
  "GetPalette",                   /* 0xAA96 */
  "PmForeColor",                  /* 0xAA97 */
  "PmBackColor",                  /* 0xAA98 */
  "AnimateEntry",                 /* 0xAA99 */
  "AnimatePalette",               /* 0xAA9A */
  "GetEntryColor",                /* 0xAA9B */
  "SetEntryColor",                /* 0xAA9C */
  "GetEntryUsage",                /* 0xAA9D */
  "SetEntryUsage",                /* 0xAA9E */
  "CTab2Palette",                 /* 0xAA9F */
  "Palette2CTab",                 /* 0xAAA0 */
  "CopyPalette",                  /* 0xAAA1 */
  "PaletteDispatch",              /* 0xAAA2 */
  "AAA3",                          /* 0xAAA3 */
  "AAA4",                          /* 0xAAA4 */
  "AAA5",                          /* 0xAAA5 */
  "AAA6",                          /* 0xAAA6 */
  "AAA7",                          /* 0xAAA7 */
  "AAA8",                          /* 0xAAA8 */
  "AAA9",                          /* 0xAAA9 */
  "QuickTime",                     /* 0xAAAA */
  "AAAB",                          /* 0xAAAB */
  "AAAC",                          /* 0xAAAC */
  "AAAD",                          /* 0xAAAD */
  "AAAE",                          /* 0xAAAE */
  "AAAF",                          /* 0xAAAF */
  "AAB0",                          /* 0xAAB0 */
  "AAB1",                          /* 0xAAB1 */
  "AAB2",                          /* 0xAAB2 */
  "AAB3",                          /* 0xAAB3 */
  "AAB4",                          /* 0xAAB4 */
  "AAB5",                          /* 0xAAB5 */
  "AAB6",                          /* 0xAAB6 */
  "AAB7",                          /* 0xAAB7 */
  "AAB8",                          /* 0xAAB8 */
  "AAB9",                          /* 0xAAB9 */
  "AABA",                          /* 0xAABA */
  "AABB",                          /* 0xAABB */
  "AABC",                          /* 0xAABC */
  "AABD",                          /* 0xAABD */
  "AABE",                          /* 0xAABE */
  "AABF",                          /* 0xAABF */
  "AAC0",                          /* 0xAAC0 */
  "AAC1",                          /* 0xAAC1 */
  "AAC2",                          /* 0xAAC2 */
  "AAC3",                          /* 0xAAC3 */
  "AAC4",                          /* 0xAAC4 */
  "AAC5",                          /* 0xAAC5 */
  "AAC6",                          /* 0xAAC6 */
  "AAC7",                          /* 0xAAC7 */
  "AAC8",                          /* 0xAAC8 */
  "AAC9",                          /* 0xAAC9 */
  "AACA",                          /* 0xAACA */
  "AACB",                          /* 0xAACB */
  "AACC",                          /* 0xAACC */
  "AACD",                          /* 0xAACD */
  "AACE",                          /* 0xAACE */
  "AACF",                          /* 0xAACF */
  "AAD0",                          /* 0xAAD0 */
  "AAD1",                          /* 0xAAD1 */
  "AAD2",                          /* 0xAAD2 */
  "AAD3",                          /* 0xAAD3 */
  "AAD4",                          /* 0xAAD4 */
  "AAD5",                          /* 0xAAD5 */
  "AAD6",                          /* 0xAAD6 */
  "AAD7",                          /* 0xAAD7 */
  "AAD8",                          /* 0xAAD8 */
  "AAD9",                          /* 0xAAD9 */
  "AADA",                          /* 0xAADA */
  "AADB",                          /* 0xAADB */
  "AADC",                          /* 0xAADC */
  "AADD",                          /* 0xAADD */
  "AADE",                          /* 0xAADE */
  "AADF",                          /* 0xAADF */
  "AAE0",                          /* 0xAAE0 */
  "AAE1",                          /* 0xAAE1 */
  "AAE2",                          /* 0xAAE2 */
  "AAE3",                          /* 0xAAE3 */
  "AAE4",                          /* 0xAAE4 */
  "AAE5",                          /* 0xAAE5 */
  "AAE6",                          /* 0xAAE6 */
  "AAE7",                          /* 0xAAE7 */
  "AAE8",                          /* 0xAAE8 */
  "AAE9",                          /* 0xAAE9 */
  "AAEA",                          /* 0xAAEA */
  "AAEB",                          /* 0xAAEB */
  "AAEC",                          /* 0xAAEC */
  "AAED",                          /* 0xAAED */
  "AAEE",                          /* 0xAAEE */
  "AAEF",                          /* 0xAAEF */
  "AAF0",                          /* 0xAAF0 */
  "AAF1",                          /* 0xAAF1 */
  "AAF2",                          /* 0xAAF2 */
  "AAF3",                          /* 0xAAF3 */
  "AAF4",                          /* 0xAAF4 */
  "AAF5",                          /* 0xAAF5 */
  "AAF6",                          /* 0xAAF6 */
  "AAF7",                          /* 0xAAF7 */
  "AAF8",                          /* 0xAAF8 */
  "AAF9",                          /* 0xAAF9 */
  "AAFA",                          /* 0xAAFA */
  "AAFB",                          /* 0xAAFB */
  "AAFC",                          /* 0xAAFC */
  "AAFD",                          /* 0xAAFD */
  "AAFE",                          /* 0xAAFE */
  "AAFF",                          /* 0xAAFF */
  "AB00",                          /* 0xAB00 */
  "AB01",                          /* 0xAB01 */
  "AB02",                          /* 0xAB02 */
  "WackyQD32Trap",                /* 0xAB03 */
  "AB04",                          /* 0xAB04 */
  "AB05",                          /* 0xAB05 */
  "AB06",                          /* 0xAB06 */
  "AB07",                          /* 0xAB07 */
  "AB08",                          /* 0xAB08 */
  "AB09",                          /* 0xAB09 */
  "AB0A",                          /* 0xAB0A */
  "AB0B",                          /* 0xAB0B */
  "AB0C",                          /* 0xAB0C */
  "AB0D",                          /* 0xAB0D */
  "AB0E",                          /* 0xAB0E */
  "AB0F",                          /* 0xAB0F */
  "AB10",                          /* 0xAB10 */
  "AB11",                          /* 0xAB11 */
  "AB12",                          /* 0xAB12 */
  "AB13",                          /* 0xAB13 */
  "AB14",                          /* 0xAB14 */
  "AB15",                          /* 0xAB15 */
  "AB16",                          /* 0xAB16 */
  "AB17",                          /* 0xAB17 */
  "AB18",                          /* 0xAB18 */
  "AB19",                          /* 0xAB19 */
  "AB1A",                          /* 0xAB1A */
  "AB1B",                          /* 0xAB1B */
  "AB1C",                          /* 0xAB1C */
  "QDExtensions",                 /* 0xAB1D */
  "AB1E",                          /* 0xAB1E */
  "AB1F",                          /* 0xAB1F */
  "AB20",                          /* 0xAB20 */
  "AB21",                          /* 0xAB21 */
  "AB22",                          /* 0xAB22 */
  "AB23",                          /* 0xAB23 */
  "AB24",                          /* 0xAB24 */
  "AB25",                          /* 0xAB25 */
  "AB26",                          /* 0xAB26 */
  "AB27",                          /* 0xAB27 */
  "AB28",                          /* 0xAB28 */
  "AB29",                          /* 0xAB29 */
  "AB2A",                          /* 0xAB2A */
  "AB2B",                          /* 0xAB2B */
  "AB2C",                          /* 0xAB2C */
  "AB2D",                          /* 0xAB2D */
  "AB2E",                          /* 0xAB2E */
  "AB2F",                          /* 0xAB2F */
  "AB30",                          /* 0xAB30 */
  "AB31",                          /* 0xAB31 */
  "AB32",                          /* 0xAB32 */
  "AB33",                          /* 0xAB33 */
  "AB34",                          /* 0xAB34 */
  "AB35",                          /* 0xAB35 */
  "AB36",                          /* 0xAB36 */
  "AB37",                          /* 0xAB37 */
  "AB38",                          /* 0xAB38 */
  "AB39",                          /* 0xAB39 */
  "AB3A",                          /* 0xAB3A */
  "AB3B",                          /* 0xAB3B */
  "AB3C",                          /* 0xAB3C */
  "AB3D",                          /* 0xAB3D */
  "AB3E",                          /* 0xAB3E */
  "AB3F",                          /* 0xAB3F */
  "AB40",                          /* 0xAB40 */
  "AB41",                          /* 0xAB41 */
  "AB42",                          /* 0xAB42 */
  "AB43",                          /* 0xAB43 */
  "AB44",                          /* 0xAB44 */
  "AB45",                          /* 0xAB45 */
  "AB46",                          /* 0xAB46 */
  "AB47",                          /* 0xAB47 */
  "AB48",                          /* 0xAB48 */
  "AB49",                          /* 0xAB49 */
  "AB4A",                          /* 0xAB4A */
  "AB4B",                          /* 0xAB4B */
  "AB4C",                          /* 0xAB4C */
  "AB4D",                          /* 0xAB4D */
  "AB4E",                          /* 0xAB4E */
  "AB4F",                          /* 0xAB4F */
  "AB50",                          /* 0xAB50 */
  "AB51",                          /* 0xAB51 */
  "AB52",                          /* 0xAB52 */
  "AB53",                          /* 0xAB53 */
  "AB54",                          /* 0xAB54 */
  "AB55",                          /* 0xAB55 */
  "AB56",                          /* 0xAB56 */
  "AB57",                          /* 0xAB57 */
  "AB58",                          /* 0xAB58 */
  "AB59",                          /* 0xAB59 */
  "AB5A",                          /* 0xAB5A */
  "AB5B",                          /* 0xAB5B */
  "AB5C",                          /* 0xAB5C */
  "AB5D",                          /* 0xAB5D */
  "AB5E",                          /* 0xAB5E */
  "AB5F",                          /* 0xAB5F */
  "AB60",                          /* 0xAB60 */
  "AB61",                          /* 0xAB61 */
  "AB62",                          /* 0xAB62 */
  "AB63",                          /* 0xAB63 */
  "AB64",                          /* 0xAB64 */
  "AB65",                          /* 0xAB65 */
  "AB66",                          /* 0xAB66 */
  "AB67",                          /* 0xAB67 */
  "AB68",                          /* 0xAB68 */
  "AB69",                          /* 0xAB69 */
  "AB6A",                          /* 0xAB6A */
  "AB6B",                          /* 0xAB6B */
  "AB6C",                          /* 0xAB6C */
  "AB6D",                          /* 0xAB6D */
  "AB6E",                          /* 0xAB6E */
  "AB6F",                          /* 0xAB6F */
  "AB70",                          /* 0xAB70 */
  "AB71",                          /* 0xAB71 */
  "AB72",                          /* 0xAB72 */
  "AB73",                          /* 0xAB73 */
  "AB74",                          /* 0xAB74 */
  "AB75",                          /* 0xAB75 */
  "AB76",                          /* 0xAB76 */
  "AB77",                          /* 0xAB77 */
  "AB78",                          /* 0xAB78 */
  "AB79",                          /* 0xAB79 */
  "AB7A",                          /* 0xAB7A */
  "AB7B",                          /* 0xAB7B */
  "AB7C",                          /* 0xAB7C */
  "AB7D",                          /* 0xAB7D */
  "AB7E",                          /* 0xAB7E */
  "AB7F",                          /* 0xAB7F */
  "AB80",                          /* 0xAB80 */
  "AB81",                          /* 0xAB81 */
  "AB82",                          /* 0xAB82 */
  "AB83",                          /* 0xAB83 */
  "AB84",                          /* 0xAB84 */
  "AB85",                          /* 0xAB85 */
  "AB86",                          /* 0xAB86 */
  "AB87",                          /* 0xAB87 */
  "AB88",                          /* 0xAB88 */
  "AB89",                          /* 0xAB89 */
  "AB8A",                          /* 0xAB8A */
  "AB8B",                          /* 0xAB8B */
  "AB8C",                          /* 0xAB8C */
  "AB8D",                          /* 0xAB8D */
  "AB8E",                          /* 0xAB8E */
  "AB8F",                          /* 0xAB8F */
  "AB90",                          /* 0xAB90 */
  "AB91",                          /* 0xAB91 */
  "AB92",                          /* 0xAB92 */
  "AB93",                          /* 0xAB93 */
  "AB94",                          /* 0xAB94 */
  "AB95",                          /* 0xAB95 */
  "AB96",                          /* 0xAB96 */
  "AB97",                          /* 0xAB97 */
  "AB98",                          /* 0xAB98 */
  "AB99",                          /* 0xAB99 */
  "AB9A",                          /* 0xAB9A */
  "AB9B",                          /* 0xAB9B */
  "AB9C",                          /* 0xAB9C */
  "AB9D",                          /* 0xAB9D */
  "AB9E",                          /* 0xAB9E */
  "AB9F",                          /* 0xAB9F */
  "ABA0",                          /* 0xABA0 */
  "ABA1",                          /* 0xABA1 */
  "ABA2",                          /* 0xABA2 */
  "ABA3",                          /* 0xABA3 */
  "ABA4",                          /* 0xABA4 */
  "ABA5",                          /* 0xABA5 */
  "ABA6",                          /* 0xABA6 */
  "ABA7",                          /* 0xABA7 */
  "ABA8",                          /* 0xABA8 */
  "ABA9",                          /* 0xABA9 */
  "ABAA",                          /* 0xABAA */
  "ABAB",                          /* 0xABAB */
  "ABAC",                          /* 0xABAC */
  "ABAD",                          /* 0xABAD */
  "ABAE",                          /* 0xABAE */
  "ABAF",                          /* 0xABAF */
  "ABB0",                          /* 0xABB0 */
  "ABB1",                          /* 0xABB1 */
  "ABB2",                          /* 0xABB2 */
  "ABB3",                          /* 0xABB3 */
  "ABB4",                          /* 0xABB4 */
  "ABB5",                          /* 0xABB5 */
  "ABB6",                          /* 0xABB6 */
  "ABB7",                          /* 0xABB7 */
  "ABB8",                          /* 0xABB8 */
  "ABB9",                          /* 0xABB9 */
  "ABBA",                          /* 0xABBA */
  "ABBB",                          /* 0xABBB */
  "ABBC",                          /* 0xABBC */
  "ABBD",                          /* 0xABBD */
  "ABBE",                          /* 0xABBE */
  "ABBF",                          /* 0xABBF */
  "ABC0",                          /* 0xABC0 */
  "ABC1",                          /* 0xABC1 */
  "ABC2",                          /* 0xABC2 */
  "ABC3",                          /* 0xABC3 */
  "ABC4",                          /* 0xABC4 */
  "ABC5",                          /* 0xABC5 */
  "ABC6",                          /* 0xABC6 */
  "ABC7",                          /* 0xABC7 */
  "ABC8",                          /* 0xABC8 */
  "IconDispatch",                 /* 0xABC9 */
  "DeviceLoop",              /* 0xABCA */
  "ABCB",                          /* 0xABCB */
  "ABCC",                          /* 0xABCC */
  "ABCD",                          /* 0xABCD */
  "ABCE",                          /* 0xABCE */
  "ABCF",                          /* 0xABCF */
  "ABD0",                          /* 0xABD0 */
  "ABD1",                          /* 0xABD1 */
  "ABD2",                          /* 0xABD2 */
  "ABD3",                          /* 0xABD3 */
  "ABD4",                          /* 0xABD4 */
  "ABD5",                          /* 0xABD5 */
  "ABD6",                          /* 0xABD6 */
  "ABD7",                          /* 0xABD7 */
  "ABD8",                          /* 0xABD8 */
  "ABD9",                          /* 0xABD9 */
  "ABDA",                          /* 0xABDA */
  "ABDB",                          /* 0xABDB */
  "ABDC",                          /* 0xABDC */
  "ABDD",                          /* 0xABDD */
  "ABDE",                          /* 0xABDE */
  "ABDF",                          /* 0xABDF */
  "ABE0",                          /* 0xABE0 */
  "ABE1",                          /* 0xABE1 */
  "ABE2",                          /* 0xABE2 */
  "ABE3",                          /* 0xABE3 */
  "ABE4",                          /* 0xABE4 */
  "ABE5",                          /* 0xABE5 */
  "ABE6",                          /* 0xABE6 */
  "ABE7",                          /* 0xABE7 */
  "ABE8",                          /* 0xABE8 */
  "ABE9",                          /* 0xABE9 */
  "ABEA",                          /* 0xABEA */
  "ABEB",                          /* 0xABEB */
  "ABEC",                          /* 0xABEC */
  "DragManagerDispatch",          /* 0xABED */
  "ABEE",                          /* 0xABEE */
  "ABEF",                          /* 0xABEF */
  "ABF0",                          /* 0xABF0 */
  "ABF1",                          /* 0xABF1 */
  "ABF2",                          /* 0xABF2 */
  "ABF3",                          /* 0xABF3 */
  "ABF4",                          /* 0xABF4 */
  "ABF5",                          /* 0xABF5 */
  "ABF6",                          /* 0xABF6 */
  "ABF7",                          /* 0xABF7 */
  "StdOpcodeProc",                /* 0xABF8 */
  "ABF9",                          /* 0xABF9 */
  "ABFA",                          /* 0xABFA */
  "ABFB",                          /* 0xABFB */
  "ABFC",                          /* 0xABFC */
  "ABFD",                          /* 0xABFD */
  "ABFE",                          /* 0xABFE */
  "DebugStr",                     /* 0xABFF */
  "SoundDispatch",                /* 0xAC00 */
  "SndDisposeChannel",            /* 0xAC01 */
  "SndAddModifier",               /* 0xAC02 */
  "SndDoCommand",                 /* 0xAC03 */
  "SndDoImmediate",               /* 0xAC04 */
  "SndPlay",                      /* 0xAC05 */
  "SndControl",                   /* 0xAC06 */
  "SndNewChannel",                /* 0xAC07 */
  "InitProcMenu",                 /* 0xAC08 */
  "GetCVariant",                  /* 0xAC09 */
  "GetWVariant",                  /* 0xAC0A */
  "PopUpMenuSelect",              /* 0xAC0B */
  "RGetResource",                 /* 0xAC0C */
  "Count1Resources",              /* 0xAC0D */
  "Get1IndResource",              /* 0xAC0E */
  "Get1IndType",                  /* 0xAC0F */
  "Unique1ID",                    /* 0xAC10 */
  "TESelView",                    /* 0xAC11 */
  "TEPinScroll",                  /* 0xAC12 */
  "TEAutoView",                   /* 0xAC13 */
  "SetFractEnable",               /* 0xAC14 */
  "SCSIDispatch",                 /* 0xAC15 */
  "Pack8",                        /* 0xAC16 */
  "CopyMask",                     /* 0xAC17 */
  "FixAtan2",                     /* 0xAC18 */
  "AC19",                          /* 0xAC19 */
  "HOpenResFile",                 /* 0xAC1A */
  "IMVI_HCreateResFile",          /* 0xAC1B */
  "Count1Types",                  /* 0xAC1C */
  "InvalMenuBar",                 /* 0xAC1D */
  "AC1E",                          /* 0xAC1E */
  "Get1Resource",                 /* 0xAC1F */
  "Get1NamedResource",            /* 0xAC20 */
  "MaxSizeRsrc",                  /* 0xAC21 */
  "ResourceDispatch",             /* 0xAC22 */
  "AliasDispatch",                /* 0xAC23 */
  "AC24",                          /* 0xAC24 */
  "AC25",                          /* 0xAC25 */
  "InsMenuItem",                  /* 0xAC26 */
  "HideDItem",                    /* 0xAC27 */
  "ShowDItem",                    /* 0xAC28 */
  "AC29",                          /* 0xAC29 */
  "AC2A",                          /* 0xAC2A */
  "Pack9",                        /* 0xAC2B */
  "Pack10",                       /* 0xAC2C */
  "Pack11",                       /* 0xAC2D */
  "Pack12",                       /* 0xAC2E */
  "Pack13",                       /* 0xAC2F */
  "Pack14",                       /* 0xAC30 */
  "Pack15",                       /* 0xAC31 */
  "AC32",                          /* 0xAC32 */
  "AC33",                          /* 0xAC33 */
  "SetFScaleDisable",             /* 0xAC34 */
  "FontMetrics",                  /* 0xAC35 */
  "GetMaskTable",                 /* 0xAC36 */
  "MeasureText",                  /* 0xAC37 */
  "CalcMask",                     /* 0xAC38 */
  "SeedFill",                     /* 0xAC39 */
  "ZoomWindow",                   /* 0xAC3A */
  "TrackBox",                     /* 0xAC3B */
  "TEGetOffset",                  /* 0xAC3C */
  "TEDispatch",                   /* 0xAC3D */
  "TEStylNew",                    /* 0xAC3E */
  "Long2Fix",                     /* 0xAC3F */
  "Fix2Long",                     /* 0xAC40 */
  "Fix2Frac",                     /* 0xAC41 */
  "Frac2Fix",                     /* 0xAC42 */
  "Fix2X",                        /* 0xAC43 */
  "X2Fix",                        /* 0xAC44 */
  "Frac2X",                       /* 0xAC45 */
  "X2Frac",                       /* 0xAC46 */
  "FracCos",                      /* 0xAC47 */
  "FracSin",                      /* 0xAC48 */
  "FracSqrt",                     /* 0xAC49 */
  "FracMul",                      /* 0xAC4A */
  "FracDiv",                      /* 0xAC4B */
  "AC4C",                          /* 0xAC4C */
  "FixDiv",                       /* 0xAC4D */
  "GetItemCmd",                   /* 0xAC4E */
  "SetItemCmd",                   /* 0xAC4F */
  "InitCursor",                   /* 0xAC50 */
  "SetCursor",                    /* 0xAC51 */
  "HideCursor",                   /* 0xAC52 */
  "ShowCursor",                   /* 0xAC53 */
  "FontDispatch",                 /* 0xAC54 */
  "ShieldCursor",                 /* 0xAC55 */
  "ObscureCursor",                /* 0xAC56 */
  "AC57",                          /* 0xAC57 */
  "BitAnd",                       /* 0xAC58 */
  "BitXor",                       /* 0xAC59 */
  "BitNot",                       /* 0xAC5A */
  "BitOr",                        /* 0xAC5B */
  "BitShift",                     /* 0xAC5C */
  "BitTst",                       /* 0xAC5D */
  "BitSet",                       /* 0xAC5E */
  "BitClr",                       /* 0xAC5F */
  "WaitNextEvent",                /* 0xAC60 */
  "Random",                       /* 0xAC61 */
  "ForeColor",                    /* 0xAC62 */
  "BackColor",                    /* 0xAC63 */
  "ColorBit",                     /* 0xAC64 */
  "GetPixel",                     /* 0xAC65 */
  "StuffHex",                     /* 0xAC66 */
  "LongMul",                      /* 0xAC67 */
  "FixMul",                       /* 0xAC68 */
  "FixRatio",                     /* 0xAC69 */
  "HiWord",                       /* 0xAC6A */
  "LoWord",                       /* 0xAC6B */
  "FixRound",                     /* 0xAC6C */
  "InitPort",                     /* 0xAC6D */
  "InitGraf",                     /* 0xAC6E */
  "OpenPort",                     /* 0xAC6F */
  "LocalToGlobal",                /* 0xAC70 */
  "GlobalToLocal",                /* 0xAC71 */
  "GrafDevice",                   /* 0xAC72 */
  "SetPort",                      /* 0xAC73 */
  "GetPort",                      /* 0xAC74 */
  "SetPortBits",                  /* 0xAC75 */
  "PortSize",                     /* 0xAC76 */
  "MovePortTo",                   /* 0xAC77 */
  "SetOrigin",                    /* 0xAC78 */
  "SetClip",                      /* 0xAC79 */
  "GetClip",                      /* 0xAC7A */
  "ClipRect",                     /* 0xAC7B */
  "BackPat",                      /* 0xAC7C */
  "ClosePort",                    /* 0xAC7D */
  "AddPt",                        /* 0xAC7E */
  "SubPt",                        /* 0xAC7F */
  "SetPt",                        /* 0xAC80 */
  "EqualPt",                      /* 0xAC81 */
  "StdText",                      /* 0xAC82 */
  "DrawChar",                     /* 0xAC83 */
  "DrawString",                   /* 0xAC84 */
  "DrawText",                     /* 0xAC85 */
  "TextWidth",                    /* 0xAC86 */
  "TextFont",                     /* 0xAC87 */
  "TextFace",                     /* 0xAC88 */
  "TextMode",                     /* 0xAC89 */
  "TextSize",                     /* 0xAC8A */
  "GetFontInfo",                  /* 0xAC8B */
  "StringWidth",                  /* 0xAC8C */
  "CharWidth",                    /* 0xAC8D */
  "SpaceExtra",                   /* 0xAC8E */
  "OSDispatch",                   /* 0xAC8F */
  "StdLine",                      /* 0xAC90 */
  "LineTo",                       /* 0xAC91 */
  "Line",                         /* 0xAC92 */
  "MoveTo",                       /* 0xAC93 */
  "Move",                         /* 0xAC94 */
  "ShutDown",                     /* 0xAC95 */
  "HidePen",                      /* 0xAC96 */
  "ShowPen",                      /* 0xAC97 */
  "GetPenState",                  /* 0xAC98 */
  "SetPenState",                  /* 0xAC99 */
  "GetPen",                       /* 0xAC9A */
  "PenSize",                      /* 0xAC9B */
  "PenMode",                      /* 0xAC9C */
  "PenPat",                       /* 0xAC9D */
  "PenNormal",                    /* 0xAC9E */
  "Unimplemented",                /* 0xAC9F */
  "StdRect",                      /* 0xACA0 */
  "FrameRect",                    /* 0xACA1 */
  "PaintRect",                    /* 0xACA2 */
  "EraseRect",                    /* 0xACA3 */
  "InvertRect",                   /* 0xACA4 */
  "FillRect",                     /* 0xACA5 */
  "EqualRect",                    /* 0xACA6 */
  "SetRect",                      /* 0xACA7 */
  "OffsetRect",                   /* 0xACA8 */
  "InsetRect",                    /* 0xACA9 */
  "SectRect",                     /* 0xACAA */
  "UnionRect",                    /* 0xACAB */
  "Pt2Rect",                      /* 0xACAC */
  "PtInRect",                     /* 0xACAD */
  "EmptyRect",                    /* 0xACAE */
  "StdRRect",                     /* 0xACAF */
  "FrameRoundRect",               /* 0xACB0 */
  "PaintRoundRect",               /* 0xACB1 */
  "EraseRoundRect",               /* 0xACB2 */
  "InvertRoundRect",              /* 0xACB3 */
  "FillRoundRect",                /* 0xACB4 */
  "ScriptUtil",                   /* 0xACB5 */
  "StdOval",                      /* 0xACB6 */
  "FrameOval",                    /* 0xACB7 */
  "PaintOval",                    /* 0xACB8 */
  "EraseOval",                    /* 0xACB9 */
  "InvertOval",                   /* 0xACBA */
  "FillOval",                     /* 0xACBB */
  "SlopeFromAngle",               /* 0xACBC */
  "StdArc",                       /* 0xACBD */
  "FrameArc",                     /* 0xACBE */
  "PaintArc",                     /* 0xACBF */
  "EraseArc",                     /* 0xACC0 */
  "InvertArc",                    /* 0xACC1 */
  "FillArc",                      /* 0xACC2 */
  "PtToAngle",                    /* 0xACC3 */
  "AngleFromSlope",               /* 0xACC4 */
  "StdPoly",                      /* 0xACC5 */
  "FramePoly",                    /* 0xACC6 */
  "PaintPoly",                    /* 0xACC7 */
  "ErasePoly",                    /* 0xACC8 */
  "InvertPoly",                   /* 0xACC9 */
  "FillPoly",                     /* 0xACCA */
  "OpenPoly",                     /* 0xACCB */
  "ClosePoly",                    /* 0xACCC */
  "KillPoly",                     /* 0xACCD */
  "OffsetPoly",                   /* 0xACCE */
  "PackBits",                     /* 0xACCF */
  "UnpackBits",                   /* 0xACD0 */
  "StdRgn",                       /* 0xACD1 */
  "FrameRgn",                     /* 0xACD2 */
  "PaintRgn",                     /* 0xACD3 */
  "EraseRgn",                     /* 0xACD4 */
  "InvertRgn",                    /* 0xACD5 */
  "FillRgn",                      /* 0xACD6 */
  "BitMapToRegion",               /* 0xACD7 */
  "NewRgn",                       /* 0xACD8 */
  "DisposeRgn",                   /* 0xACD9 */
  "OpenRgn",                      /* 0xACDA */
  "CloseRgn",                     /* 0xACDB */
  "CopyRgn",                      /* 0xACDC */
  "SetEmptyRgn",                  /* 0xACDD */
  "SetRectRgn",                   /* 0xACDE */
  "RectRgn",                      /* 0xACDF */
  "OffsetRgn",                    /* 0xACE0 */
  "InsetRgn",                     /* 0xACE1 */
  "EmptyRgn",                     /* 0xACE2 */
  "EqualRgn",                     /* 0xACE3 */
  "SectRgn",                      /* 0xACE4 */
  "UnionRgn",                     /* 0xACE5 */
  "DiffRgn",                      /* 0xACE6 */
  "XorRgn",                       /* 0xACE7 */
  "PtInRgn",                      /* 0xACE8 */
  "RectInRgn",                    /* 0xACE9 */
  "SetStdProcs",                  /* 0xACEA */
  "StdBits",                      /* 0xACEB */
  "CopyBits",                     /* 0xACEC */
  "StdTxMeas",                    /* 0xACED */
  "StdGetPic",                    /* 0xACEE */
  "ScrollRect",                   /* 0xACEF */
  "StdPutPic",                    /* 0xACF0 */
  "StdComment",                   /* 0xACF1 */
  "PicComment",                   /* 0xACF2 */
  "OpenPicture",                  /* 0xACF3 */
  "ClosePicture",                 /* 0xACF4 */
  "KillPicture",                  /* 0xACF5 */
  "DrawPicture",                  /* 0xACF6 */
  "ACF7",                          /* 0xACF7 */
  "ScalePt",                      /* 0xACF8 */
  "MapPt",                        /* 0xACF9 */
  "MapRect",                      /* 0xACFA */
  "MapRgn",                       /* 0xACFB */
  "MapPoly",                      /* 0xACFC */
  "PrGlue",                       /* 0xACFD */
  "InitFonts",                    /* 0xACFE */
  "GetFontName",                  /* 0xACFF */
  "GetFNum",                      /* 0xAD00 */
  "FMSwapFont",                   /* 0xAD01 */
  "RealFont",                     /* 0xAD02 */
  "SetFontLock",                  /* 0xAD03 */
  "DrawGrowIcon",                 /* 0xAD04 */
  "DragGrayRgn",                  /* 0xAD05 */
  "NewString",                    /* 0xAD06 */
  "SetString",                    /* 0xAD07 */
  "ShowHide",                     /* 0xAD08 */
  "CalcVis",                      /* 0xAD09 */
  "CalcVisBehind",                /* 0xAD0A */
  "ClipAbove",                    /* 0xAD0B */
  "PaintOne",                     /* 0xAD0C */
  "PaintBehind",                  /* 0xAD0D */
  "SaveOld",                      /* 0xAD0E */
  "DrawNew",                      /* 0xAD0F */
  "GetWMgrPort",                  /* 0xAD10 */
  "CheckUpdate",                  /* 0xAD11 */
  "InitWindows",                  /* 0xAD12 */
  "NewWindow",                    /* 0xAD13 */
  "DisposeWindow",                /* 0xAD14 */
  "ShowWindow",                   /* 0xAD15 */
  "HideWindow",                   /* 0xAD16 */
  "GetWRefCon",                   /* 0xAD17 */
  "SetWRefCon",                   /* 0xAD18 */
  "GetWTitle",                    /* 0xAD19 */
  "SetWTitle",                    /* 0xAD1A */
  "MoveWindow",                   /* 0xAD1B */
  "HiliteWindow",                 /* 0xAD1C */
  "SizeWindow",                   /* 0xAD1D */
  "TrackGoAway",                  /* 0xAD1E */
  "SelectWindow",                 /* 0xAD1F */
  "BringToFront",                 /* 0xAD20 */
  "SendBehind",                   /* 0xAD21 */
  "BeginUpdate",                  /* 0xAD22 */
  "EndUpdate",                    /* 0xAD23 */
  "FrontWindow",                  /* 0xAD24 */
  "DragWindow",                   /* 0xAD25 */
  "DragTheRgn",                   /* 0xAD26 */
  "InvalRgn",                     /* 0xAD27 */
  "InvalRect",                    /* 0xAD28 */
  "ValidRgn",                     /* 0xAD29 */
  "ValidRect",                    /* 0xAD2A */
  "GrowWindow",                   /* 0xAD2B */
  "FindWindow",                   /* 0xAD2C */
  "CloseWindow",                  /* 0xAD2D */
  "SetWindowPic",                 /* 0xAD2E */
  "GetWindowPic",                 /* 0xAD2F */
  "InitMenus",                    /* 0xAD30 */
  "NewMenu",                      /* 0xAD31 */
  "DisposeMenu",                  /* 0xAD32 */
  "AppendMenu",                   /* 0xAD33 */
  "ClearMenuBar",                 /* 0xAD34 */
  "InsertMenu",                   /* 0xAD35 */
  "DeleteMenu",                   /* 0xAD36 */
  "DrawMenuBar",                  /* 0xAD37 */
  "HiliteMenu",                   /* 0xAD38 */
  "EnableItem",                   /* 0xAD39 */
  "DisableItem",                  /* 0xAD3A */
  "GetMenuBar",                   /* 0xAD3B */
  "SetMenuBar",                   /* 0xAD3C */
  "MenuSelect",                   /* 0xAD3D */
  "MenuKey",                      /* 0xAD3E */
  "GetItemIcon",                  /* 0xAD3F */
  "SetItemIcon",                  /* 0xAD40 */
  "GetItemStyle",                 /* 0xAD41 */
  "SetItemStyle",                 /* 0xAD42 */
  "GetItemMark",                  /* 0xAD43 */
  "SetItemMark",                  /* 0xAD44 */
  "CheckItem",                    /* 0xAD45 */
  "GetItem",                      /* 0xAD46 */
  "SetItem",                      /* 0xAD47 */
  "CalcMenuSize",                 /* 0xAD48 */
  "GetMHandle",                   /* 0xAD49 */
  "SetMenuFlash",                 /* 0xAD4A */
  "PlotIcon",                     /* 0xAD4B */
  "FlashMenuBar",                 /* 0xAD4C */
  "AddResMenu",                   /* 0xAD4D */
  "PinRect",                      /* 0xAD4E */
  "DeltaPoint",                   /* 0xAD4F */
  "CountMItems",                  /* 0xAD50 */
  "InsertResMenu",                /* 0xAD51 */
  "DelMenuItem",                  /* 0xAD52 */
  "UpdtControl",                  /* 0xAD53 */
  "NewControl",                   /* 0xAD54 */
  "DisposeControl",               /* 0xAD55 */
  "KillControls",                 /* 0xAD56 */
  "ShowControl",                  /* 0xAD57 */
  "HideControl",                  /* 0xAD58 */
  "MoveControl",                  /* 0xAD59 */
  "GetCRefCon",                   /* 0xAD5A */
  "SetCRefCon",                   /* 0xAD5B */
  "SizeControl",                  /* 0xAD5C */
  "HiliteControl",                /* 0xAD5D */
  "GetCTitle",                    /* 0xAD5E */
  "SetCTitle",                    /* 0xAD5F */
  "GetCtlValue",                  /* 0xAD60 */
  "GetCtlMin",                    /* 0xAD61 */
  "GetCtlMax",                    /* 0xAD62 */
  "SetCtlValue",                  /* 0xAD63 */
  "SetCtlMin",                    /* 0xAD64 */
  "SetCtlMax",                    /* 0xAD65 */
  "TestControl",                  /* 0xAD66 */
  "DragControl",                  /* 0xAD67 */
  "TrackControl",                 /* 0xAD68 */
  "DrawControls",                 /* 0xAD69 */
  "GetCtlAction",                 /* 0xAD6A */
  "SetCtlAction",                 /* 0xAD6B */
  "FindControl",                  /* 0xAD6C */
  "Draw1Control",                 /* 0xAD6D */
  "Dequeue",                      /* 0xAD6E */
  "Enqueue",                      /* 0xAD6F */
  "GetNextEvent",                 /* 0xAD70 */
  "EventAvail",                   /* 0xAD71 */
  "GetMouse",                     /* 0xAD72 */
  "StillDown",                    /* 0xAD73 */
  "Button",                       /* 0xAD74 */
  "TickCount",                    /* 0xAD75 */
  "GetKeys",                      /* 0xAD76 */
  "WaitMouseUp",                  /* 0xAD77 */
  "UpdtDialog",                   /* 0xAD78 */
  "CouldDialog",                  /* 0xAD79 */
  "FreeDialog",                   /* 0xAD7A */
  "InitDialogs",                  /* 0xAD7B */
  "GetNewDialog",                 /* 0xAD7C */
  "NewDialog",                    /* 0xAD7D */
  "SelIText",                     /* 0xAD7E */
  "IsDialogEvent",                /* 0xAD7F */
  "DialogSelect",                 /* 0xAD80 */
  "DrawDialog",                   /* 0xAD81 */
  "CloseDialog",                  /* 0xAD82 */
  "DisposDialog",                 /* 0xAD83 */
  "FindDItem",                    /* 0xAD84 */
  "Alert",                        /* 0xAD85 */
  "StopAlert",                    /* 0xAD86 */
  "NoteAlert",                    /* 0xAD87 */
  "CautionAlert",                 /* 0xAD88 */
  "CouldAlert",                   /* 0xAD89 */
  "FreeAlert",                    /* 0xAD8A */
  "ParamText",                    /* 0xAD8B */
  "ErrorSound",                   /* 0xAD8C */
  "GetDItem",                     /* 0xAD8D */
  "SetDItem",                     /* 0xAD8E */
  "SetIText",                     /* 0xAD8F */
  "GetIText",                     /* 0xAD90 */
  "ModalDialog",                  /* 0xAD91 */
  "DetachResource",               /* 0xAD92 */
  "SetResPurge",                  /* 0xAD93 */
  "CurResFile",                   /* 0xAD94 */
  "InitResources",                /* 0xAD95 */
  "RsrcZoneInit",                 /* 0xAD96 */
  "OpenResFile",                  /* 0xAD97 */
  "UseResFile",                   /* 0xAD98 */
  "UpdateResFile",                /* 0xAD99 */
  "CloseResFile",                 /* 0xAD9A */
  "SetResLoad",                   /* 0xAD9B */
  "CountResources",               /* 0xAD9C */
  "GetIndResource",               /* 0xAD9D */
  "CountTypes",                   /* 0xAD9E */
  "GetIndType",                   /* 0xAD9F */
  "GetResource",                  /* 0xADA0 */
  "GetNamedResource",             /* 0xADA1 */
  "LoadResource",                 /* 0xADA2 */
  "ReleaseResource",              /* 0xADA3 */
  "HomeResFile",                  /* 0xADA4 */
  "SizeResource",                 /* 0xADA5 */
  "GetResAttrs",                  /* 0xADA6 */
  "SetResAttrs",                  /* 0xADA7 */
  "GetResInfo",                   /* 0xADA8 */
  "SetResInfo",                   /* 0xADA9 */
  "ChangedResource",              /* 0xADAA */
  "AddResource",                  /* 0xADAB */
  "AddReference",                 /* 0xADAC */
  "RmveResource",                 /* 0xADAD */
  "RmveReference",                /* 0xADAE */
  "ResError",                     /* 0xADAF */
  "WriteResource",                /* 0xADB0 */
  "CreateResFile",                /* 0xADB1 */
  "SystemEvent",                  /* 0xADB2 */
  "SystemClick",                  /* 0xADB3 */
  "SystemTask",                   /* 0xADB4 */
  "SystemMenu",                   /* 0xADB5 */
  "OpenDeskAcc",                  /* 0xADB6 */
  "CloseDeskAcc",                 /* 0xADB7 */
  "GetPattern",                   /* 0xADB8 */
  "GetCursor",                    /* 0xADB9 */
  "GetString",                    /* 0xADBA */
  "GetIcon",                      /* 0xADBB */
  "GetPicture",                   /* 0xADBC */
  "GetNewWindow",                 /* 0xADBD */
  "GetNewControl",                /* 0xADBE */
  "GetMenu",                      /* 0xADBF */
  "GetNewMBar",                   /* 0xADC0 */
  "UniqueID",                     /* 0xADC1 */
  "SystemEdit",                   /* 0xADC2 */
  "KeyTrans",                     /* 0xADC3 */
  "OpenRFPerm",                   /* 0xADC4 */
  "RsrcMapEntry",                 /* 0xADC5 */
  "Secs2Date",                    /* 0xADC6 */
  "Date2Secs",                    /* 0xADC7 */
  "SysBeep",                      /* 0xADC8 */
  "SysError",                     /* 0xADC9 */
  "PutIcon",                      /* 0xADCA */
  "TEGetText",                    /* 0xADCB */
  "TEInit",                       /* 0xADCC */
  "TEDispose",                    /* 0xADCD */
  "TextBox",                      /* 0xADCE */
  "TESetText",                    /* 0xADCF */
  "TECalText",                    /* 0xADD0 */
  "TESetSelect",                  /* 0xADD1 */
  "TENew",                        /* 0xADD2 */
  "TEUpdate",                     /* 0xADD3 */
  "TEClick",                      /* 0xADD4 */
  "TECopy",                       /* 0xADD5 */
  "TECut",                        /* 0xADD6 */
  "TEDelete",                     /* 0xADD7 */
  "TEActivate",                   /* 0xADD8 */
  "TEDeactivate",                 /* 0xADD9 */
  "TEIdle",                       /* 0xADDA */
  "TEPaste",                      /* 0xADDB */
  "TEKey",                        /* 0xADDC */
  "TEScroll",                     /* 0xADDD */
  "TEInsert",                     /* 0xADDE */
  "TESetJust",                    /* 0xADDF */
  "Munger",                       /* 0xADE0 */
  "HandToHand",                   /* 0xADE1 */
  "PtrToXHand",                   /* 0xADE2 */
  "PtrToHand",                    /* 0xADE3 */
  "HandAndHand",                  /* 0xADE4 */
  "InitPack",                     /* 0xADE5 */
  "InitAllPacks",                 /* 0xADE6 */
  "Pack0",                        /* 0xADE7 */
  "Pack1",                        /* 0xADE8 */
  "Pack2",                        /* 0xADE9 */
  "Pack3",                        /* 0xADEA */
  "Pack4",                        /* 0xADEB */
  "Pack5",                        /* 0xADEC */
  "Pack6",                        /* 0xADED */
  "Pack7",                        /* 0xADEE */
  "PtrAndHand",                   /* 0xADEF */
  "LoadSeg",                      /* 0xADF0 */
  "UnloadSeg",                    /* 0xADF1 */
  "Launch",                       /* 0xADF2 */
  "Chain",                        /* 0xADF3 */
  "ExitToShell",                  /* 0xADF4 */
  "GetAppParms",                  /* 0xADF5 */
  "GetResFileAttrs",              /* 0xADF6 */
  "SetResFileAttrs",              /* 0xADF7 */
  "MethodDispatch",               /* 0xADF8 */
  "InfoScrap",                    /* 0xADF9 */
  "UnloadScrap",                  /* 0xADFA */
  "LoadScrap",                    /* 0xADFB */
  "ZeroScrap",                    /* 0xADFC */
  "GetScrap",                     /* 0xADFD */
  "PutScrap",                     /* 0xADFE */
  "Debugger",                     /* 0xADFF */
  "OpenCPort",                    /* 0xAE00 */
  "InitCPort",                    /* 0xAE01 */
  "AE02",                          /* 0xAE02 */
  "NewPixMap",                    /* 0xAE03 */
  "DisposPixMap",                 /* 0xAE04 */
  "CopyPixMap",                   /* 0xAE05 */
  "SetPortPix",                   /* 0xAE06 */
  "NewPixPat",                    /* 0xAE07 */
  "DisposPixPat",                 /* 0xAE08 */
  "CopyPixPat",                   /* 0xAE09 */
  "PenPixPat",                    /* 0xAE0A */
  "BackPixPat",                   /* 0xAE0B */
  "GetPixPat",                    /* 0xAE0C */
  "MakeRGBPat",                   /* 0xAE0D */
  "FillCRect",                    /* 0xAE0E */
  "FillCOval",                    /* 0xAE0F */
  "FillCRoundRect",               /* 0xAE10 */
  "FillCArc",                     /* 0xAE11 */
  "FillCRgn",                     /* 0xAE12 */
  "FillCPoly",                    /* 0xAE13 */
  "RGBForeColor",                 /* 0xAE14 */
  "RGBBackColor",                 /* 0xAE15 */
  "SetCPixel",                    /* 0xAE16 */
  "GetCPixel",                    /* 0xAE17 */
  "GetCTable",                    /* 0xAE18 */
  "GetForeColor",                 /* 0xAE19 */
  "GetBackColor",                 /* 0xAE1A */
  "GetCCursor",                   /* 0xAE1B */
  "SetCCursor",                   /* 0xAE1C */
  "AllocCursor",                  /* 0xAE1D */
  "GetCIcon",                     /* 0xAE1E */
  "PlotCIcon",                    /* 0xAE1F */
  "IMVI_OpenCPicture",            /* 0xAE20 */
  "OpColor",                      /* 0xAE21 */
  "HiliteColor",                  /* 0xAE22 */
  "CharExtra",                    /* 0xAE23 */
  "DisposCTable",                 /* 0xAE24 */
  "DisposeCIcon",                 /* 0xAE25 */
  "DisposCCursor",                /* 0xAE26 */
  "GetMaxDevice",                 /* 0xAE27 */
  "GetCTSeed",                    /* 0xAE28 */
  "GetDeviceList",                /* 0xAE29 */
  "GetMainDevice",                /* 0xAE2A */
  "GetNextDevice",                /* 0xAE2B */
  "TestDeviceAttribute",          /* 0xAE2C */
  "SetDeviceAttribute",           /* 0xAE2D */
  "InitGDevice",                  /* 0xAE2E */
  "NewGDevice",                   /* 0xAE2F */
  "DisposGDevice",                /* 0xAE30 */
  "SetGDevice",                   /* 0xAE31 */
  "GetGDevice",                   /* 0xAE32 */
  "Color2Index",                  /* 0xAE33 */
  "Index2Color",                  /* 0xAE34 */
  "InvertColor",                  /* 0xAE35 */
  "RealColor",                    /* 0xAE36 */
  "GetSubTable",                  /* 0xAE37 */
  "AE38",                          /* 0xAE38 */
  "MakeITable",                   /* 0xAE39 */
  "AddSearch",                    /* 0xAE3A */
  "AddComp",                      /* 0xAE3B */
  "SetClientID",                  /* 0xAE3C */
  "ProtectEntry",                 /* 0xAE3D */
  "ReserveEntry",                 /* 0xAE3E */
  "SetEntries",                   /* 0xAE3F */
  "QDError",                      /* 0xAE40 */
  "SetWinColor",                  /* 0xAE41 */
  "GetAuxWin",                    /* 0xAE42 */
  "SetCtlColor",                  /* 0xAE43 */
  "GetAuxCtl",                    /* 0xAE44 */
  "NewCWindow",                   /* 0xAE45 */
  "GetNewCWindow",                /* 0xAE46 */
  "SetDeskCPat",                  /* 0xAE47 */
  "GetCWMgrPort",                 /* 0xAE48 */
  "SaveEntries",                  /* 0xAE49 */
  "RestoreEntries",               /* 0xAE4A */
  "NewCDialog",                   /* 0xAE4B */
  "DelSearch",                    /* 0xAE4C */
  "DelComp",                      /* 0xAE4D */
  "SetStdCProcs",                 /* 0xAE4E */
  "CalcCMask",                    /* 0xAE4F */
  "SeedCFill",                    /* 0xAE50 */
  "IMVI_CopyDeepMask",            /* 0xAE51 */
  "HighLevelFSDispatch",          /* 0xAE52 */
  "AE53",                          /* 0xAE53 */
  "AE54",                          /* 0xAE54 */
  "AE55",                          /* 0xAE55 */
  "AE56",                          /* 0xAE56 */
  "DockingDispatch",              /* 0xAE57 */
  "AE58",                          /* 0xAE58 */
  "AE59",                          /* 0xAE59 */
  "AE5A",                          /* 0xAE5A */
  "AE5B",                          /* 0xAE5B */
  "AE5C",                          /* 0xAE5C */
  "AE5D",                          /* 0xAE5D */
  "AE5E",                          /* 0xAE5E */
  "AE5F",                          /* 0xAE5F */
  "DelMCEntries",                 /* 0xAE60 */
  "GetMCInfo",                    /* 0xAE61 */
  "SetMCInfo",                    /* 0xAE62 */
  "DispMCInfo",                   /* 0xAE63 */
  "GetMCEntry",                   /* 0xAE64 */
  "SetMCEntries",                 /* 0xAE65 */
  "MenuChoice",                   /* 0xAE66 */
  "AE67",                          /* 0xAE67 */
  "DialogDispatch",               /* 0xAE68 */
  "AE69",                          /* 0xAE69 */
  "AE6A",                          /* 0xAE6A */
  "AE6B",                          /* 0xAE6B */
  "AE6C",                          /* 0xAE6C */
  "AE6D",                          /* 0xAE6D */
  "AE6E",                          /* 0xAE6E */
  "AE6F",                          /* 0xAE6F */
  "AE70",                          /* 0xAE70 */
  "AE71",                          /* 0xAE71 */
  "AE72",                          /* 0xAE72 */
  "AE73",                          /* 0xAE73 */
  "AE74",                          /* 0xAE74 */
  "AE75",                          /* 0xAE75 */
  "AE76",                          /* 0xAE76 */
  "AE77",                          /* 0xAE77 */
  "AE78",                          /* 0xAE78 */
  "AE79",                          /* 0xAE79 */
  "AE7A",                          /* 0xAE7A */
  "AE7B",                          /* 0xAE7B */
  "AE7C",                          /* 0xAE7C */
  "AE7D",                          /* 0xAE7D */
  "AE7E",                          /* 0xAE7E */
  "AE7F",                          /* 0xAE7F */
  "AE80",                          /* 0xAE80 */
  "AE81",                          /* 0xAE81 */
  "AE82",                          /* 0xAE82 */
  "AE83",                          /* 0xAE83 */
  "AE84",                          /* 0xAE84 */
  "AE85",                          /* 0xAE85 */
  "AE86",                          /* 0xAE86 */
  "AE87",                          /* 0xAE87 */
  "AE88",                          /* 0xAE88 */
  "AE89",                          /* 0xAE89 */
  "AE8A",                          /* 0xAE8A */
  "AE8B",                          /* 0xAE8B */
  "AE8C",                          /* 0xAE8C */
  "AE8D",                          /* 0xAE8D */
  "AE8E",                          /* 0xAE8E */
  "AE8F",                          /* 0xAE8F */
  "InitPalettes",                 /* 0xAE90 */
  "NewPalette",                   /* 0xAE91 */
  "GetNewPalette",                /* 0xAE92 */
  "DisposePalette",               /* 0xAE93 */
  "ActivatePalette",              /* 0xAE94 */
  "NSetPalette",                   /* 0xAE95 */
  "GetPalette",                   /* 0xAE96 */
  "PmForeColor",                  /* 0xAE97 */
  "PmBackColor",                  /* 0xAE98 */
  "AnimateEntry",                 /* 0xAE99 */
  "AnimatePalette",               /* 0xAE9A */
  "GetEntryColor",                /* 0xAE9B */
  "SetEntryColor",                /* 0xAE9C */
  "GetEntryUsage",                /* 0xAE9D */
  "SetEntryUsage",                /* 0xAE9E */
  "CTab2Palette",                 /* 0xAE9F */
  "Palette2CTab",                 /* 0xAEA0 */
  "CopyPalette",                  /* 0xAEA1 */
  "PaletteDispatch",              /* 0xAEA2 */
  "AEA3",                          /* 0xAEA3 */
  "AEA4",                          /* 0xAEA4 */
  "AEA5",                          /* 0xAEA5 */
  "AEA6",                          /* 0xAEA6 */
  "AEA7",                          /* 0xAEA7 */
  "AEA8",                          /* 0xAEA8 */
  "AEA9",                          /* 0xAEA9 */
  "AEAA",                          /* 0xAEAA */
  "AEAB",                          /* 0xAEAB */
  "AEAC",                          /* 0xAEAC */
  "AEAD",                          /* 0xAEAD */
  "AEAE",                          /* 0xAEAE */
  "AEAF",                          /* 0xAEAF */
  "AEB0",                          /* 0xAEB0 */
  "AEB1",                          /* 0xAEB1 */
  "AEB2",                          /* 0xAEB2 */
  "AEB3",                          /* 0xAEB3 */
  "AEB4",                          /* 0xAEB4 */
  "AEB5",                          /* 0xAEB5 */
  "AEB6",                          /* 0xAEB6 */
  "AEB7",                          /* 0xAEB7 */
  "AEB8",                          /* 0xAEB8 */
  "AEB9",                          /* 0xAEB9 */
  "AEBA",                          /* 0xAEBA */
  "AEBB",                          /* 0xAEBB */
  "AEBC",                          /* 0xAEBC */
  "AEBD",                          /* 0xAEBD */
  "AEBE",                          /* 0xAEBE */
  "AEBF",                          /* 0xAEBF */
  "AEC0",                          /* 0xAEC0 */
  "AEC1",                          /* 0xAEC1 */
  "AEC2",                          /* 0xAEC2 */
  "AEC3",                          /* 0xAEC3 */
  "AEC4",                          /* 0xAEC4 */
  "AEC5",                          /* 0xAEC5 */
  "AEC6",                          /* 0xAEC6 */
  "AEC7",                          /* 0xAEC7 */
  "AEC8",                          /* 0xAEC8 */
  "AEC9",                          /* 0xAEC9 */
  "AECA",                          /* 0xAECA */
  "AECB",                          /* 0xAECB */
  "AECC",                          /* 0xAECC */
  "AECD",                          /* 0xAECD */
  "AECE",                          /* 0xAECE */
  "AECF",                          /* 0xAECF */
  "AED0",                          /* 0xAED0 */
  "AED1",                          /* 0xAED1 */
  "AED2",                          /* 0xAED2 */
  "AED3",                          /* 0xAED3 */
  "AED4",                          /* 0xAED4 */
  "AED5",                          /* 0xAED5 */
  "AED6",                          /* 0xAED6 */
  "AED7",                          /* 0xAED7 */
  "AED8",                          /* 0xAED8 */
  "AED9",                          /* 0xAED9 */
  "AEDA",                          /* 0xAEDA */
  "AEDB",                          /* 0xAEDB */
  "AEDC",                          /* 0xAEDC */
  "AEDD",                          /* 0xAEDD */
  "AEDE",                          /* 0xAEDE */
  "AEDF",                          /* 0xAEDF */
  "AEE0",                          /* 0xAEE0 */
  "AEE1",                          /* 0xAEE1 */
  "AEE2",                          /* 0xAEE2 */
  "AEE3",                          /* 0xAEE3 */
  "AEE4",                          /* 0xAEE4 */
  "AEE5",                          /* 0xAEE5 */
  "AEE6",                          /* 0xAEE6 */
  "AEE7",                          /* 0xAEE7 */
  "AEE8",                          /* 0xAEE8 */
  "AEE9",                          /* 0xAEE9 */
  "AEEA",                          /* 0xAEEA */
  "AEEB",                          /* 0xAEEB */
  "AEEC",                          /* 0xAEEC */
  "AEED",                          /* 0xAEED */
  "AEEE",                          /* 0xAEEE */
  "AEEF",                          /* 0xAEEF */
  "AEF0",                          /* 0xAEF0 */
  "AEF1",                          /* 0xAEF1 */
  "AEF2",                          /* 0xAEF2 */
  "AEF3",                          /* 0xAEF3 */
  "AEF4",                          /* 0xAEF4 */
  "AEF5",                          /* 0xAEF5 */
  "AEF6",                          /* 0xAEF6 */
  "AEF7",                          /* 0xAEF7 */
  "AEF8",                          /* 0xAEF8 */
  "AEF9",                          /* 0xAEF9 */
  "AEFA",                          /* 0xAEFA */
  "AEFB",                          /* 0xAEFB */
  "AEFC",                          /* 0xAEFC */
  "AEFD",                          /* 0xAEFD */
  "AEFE",                          /* 0xAEFE */
  "AEFF",                          /* 0xAEFF */
  "AF00",                          /* 0xAF00 */
  "AF01",                          /* 0xAF01 */
  "AF02",                          /* 0xAF02 */
  "WackyQD32Trap",                /* 0xAF03 */
  "AF04",                          /* 0xAF04 */
  "AF05",                          /* 0xAF05 */
  "AF06",                          /* 0xAF06 */
  "AF07",                          /* 0xAF07 */
  "AF08",                          /* 0xAF08 */
  "AF09",                          /* 0xAF09 */
  "AF0A",                          /* 0xAF0A */
  "AF0B",                          /* 0xAF0B */
  "AF0C",                          /* 0xAF0C */
  "AF0D",                          /* 0xAF0D */
  "AF0E",                          /* 0xAF0E */
  "AF0F",                          /* 0xAF0F */
  "AF10",                          /* 0xAF10 */
  "AF11",                          /* 0xAF11 */
  "AF12",                          /* 0xAF12 */
  "AF13",                          /* 0xAF13 */
  "AF14",                          /* 0xAF14 */
  "AF15",                          /* 0xAF15 */
  "AF16",                          /* 0xAF16 */
  "AF17",                          /* 0xAF17 */
  "AF18",                          /* 0xAF18 */
  "AF19",                          /* 0xAF19 */
  "AF1A",                          /* 0xAF1A */
  "AF1B",                          /* 0xAF1B */
  "AF1C",                          /* 0xAF1C */
  "QDExtensions",                 /* 0xAF1D */
  "AF1E",                          /* 0xAF1E */
  "AF1F",                          /* 0xAF1F */
  "AF20",                          /* 0xAF20 */
  "AF21",                          /* 0xAF21 */
  "AF22",                          /* 0xAF22 */
  "AF23",                          /* 0xAF23 */
  "AF24",                          /* 0xAF24 */
  "AF25",                          /* 0xAF25 */
  "AF26",                          /* 0xAF26 */
  "AF27",                          /* 0xAF27 */
  "AF28",                          /* 0xAF28 */
  "AF29",                          /* 0xAF29 */
  "AF2A",                          /* 0xAF2A */
  "AF2B",                          /* 0xAF2B */
  "AF2C",                          /* 0xAF2C */
  "AF2D",                          /* 0xAF2D */
  "AF2E",                          /* 0xAF2E */
  "AF2F",                          /* 0xAF2F */
  "AF30",                          /* 0xAF30 */
  "AF31",                          /* 0xAF31 */
  "AF32",                          /* 0xAF32 */
  "AF33",                          /* 0xAF33 */
  "AF34",                          /* 0xAF34 */
  "AF35",                          /* 0xAF35 */
  "AF36",                          /* 0xAF36 */
  "AF37",                          /* 0xAF37 */
  "AF38",                          /* 0xAF38 */
  "AF39",                          /* 0xAF39 */
  "AF3A",                          /* 0xAF3A */
  "AF3B",                          /* 0xAF3B */
  "AF3C",                          /* 0xAF3C */
  "AF3D",                          /* 0xAF3D */
  "AF3E",                          /* 0xAF3E */
  "AF3F",                          /* 0xAF3F */
  "AF40",                          /* 0xAF40 */
  "AF41",                          /* 0xAF41 */
  "AF42",                          /* 0xAF42 */
  "AF43",                          /* 0xAF43 */
  "AF44",                          /* 0xAF44 */
  "AF45",                          /* 0xAF45 */
  "AF46",                          /* 0xAF46 */
  "AF47",                          /* 0xAF47 */
  "AF48",                          /* 0xAF48 */
  "AF49",                          /* 0xAF49 */
  "AF4A",                          /* 0xAF4A */
  "AF4B",                          /* 0xAF4B */
  "AF4C",                          /* 0xAF4C */
  "AF4D",                          /* 0xAF4D */
  "AF4E",                          /* 0xAF4E */
  "AF4F",                          /* 0xAF4F */
  "AF50",                          /* 0xAF50 */
  "AF51",                          /* 0xAF51 */
  "AF52",                          /* 0xAF52 */
  "AF53",                          /* 0xAF53 */
  "AF54",                          /* 0xAF54 */
  "AF55",                          /* 0xAF55 */
  "AF56",                          /* 0xAF56 */
  "AF57",                          /* 0xAF57 */
  "AF58",                          /* 0xAF58 */
  "AF59",                          /* 0xAF59 */
  "AF5A",                          /* 0xAF5A */
  "AF5B",                          /* 0xAF5B */
  "AF5C",                          /* 0xAF5C */
  "AF5D",                          /* 0xAF5D */
  "AF5E",                          /* 0xAF5E */
  "AF5F",                          /* 0xAF5F */
  "AF60",                          /* 0xAF60 */
  "AF61",                          /* 0xAF61 */
  "AF62",                          /* 0xAF62 */
  "AF63",                          /* 0xAF63 */
  "AF64",                          /* 0xAF64 */
  "AF65",                          /* 0xAF65 */
  "AF66",                          /* 0xAF66 */
  "AF67",                          /* 0xAF67 */
  "AF68",                          /* 0xAF68 */
  "AF69",                          /* 0xAF69 */
  "AF6A",                          /* 0xAF6A */
  "AF6B",                          /* 0xAF6B */
  "AF6C",                          /* 0xAF6C */
  "AF6D",                          /* 0xAF6D */
  "AF6E",                          /* 0xAF6E */
  "AF6F",                          /* 0xAF6F */
  "AF70",                          /* 0xAF70 */
  "AF71",                          /* 0xAF71 */
  "AF72",                          /* 0xAF72 */
  "AF73",                          /* 0xAF73 */
  "AF74",                          /* 0xAF74 */
  "AF75",                          /* 0xAF75 */
  "AF76",                          /* 0xAF76 */
  "AF77",                          /* 0xAF77 */
  "AF78",                          /* 0xAF78 */
  "AF79",                          /* 0xAF79 */
  "AF7A",                          /* 0xAF7A */
  "AF7B",                          /* 0xAF7B */
  "AF7C",                          /* 0xAF7C */
  "AF7D",                          /* 0xAF7D */
  "AF7E",                          /* 0xAF7E */
  "AF7F",                          /* 0xAF7F */
  "AF80",                          /* 0xAF80 */
  "AF81",                          /* 0xAF81 */
  "AF82",                          /* 0xAF82 */
  "AF83",                          /* 0xAF83 */
  "AF84",                          /* 0xAF84 */
  "AF85",                          /* 0xAF85 */
  "AF86",                          /* 0xAF86 */
  "AF87",                          /* 0xAF87 */
  "AF88",                          /* 0xAF88 */
  "AF89",                          /* 0xAF89 */
  "AF8A",                          /* 0xAF8A */
  "AF8B",                          /* 0xAF8B */
  "AF8C",                          /* 0xAF8C */
  "AF8D",                          /* 0xAF8D */
  "AF8E",                          /* 0xAF8E */
  "AF8F",                          /* 0xAF8F */
  "AF90",                          /* 0xAF90 */
  "AF91",                          /* 0xAF91 */
  "AF92",                          /* 0xAF92 */
  "AF93",                          /* 0xAF93 */
  "AF94",                          /* 0xAF94 */
  "AF95",                          /* 0xAF95 */
  "AF96",                          /* 0xAF96 */
  "AF97",                          /* 0xAF97 */
  "AF98",                          /* 0xAF98 */
  "AF99",                          /* 0xAF99 */
  "AF9A",                          /* 0xAF9A */
  "AF9B",                          /* 0xAF9B */
  "AF9C",                          /* 0xAF9C */
  "AF9D",                          /* 0xAF9D */
  "AF9E",                          /* 0xAF9E */
  "AF9F",                          /* 0xAF9F */
  "AFA0",                          /* 0xAFA0 */
  "AFA1",                          /* 0xAFA1 */
  "AFA2",                          /* 0xAFA2 */
  "AFA3",                          /* 0xAFA3 */
  "AFA4",                          /* 0xAFA4 */
  "AFA5",                          /* 0xAFA5 */
  "AFA6",                          /* 0xAFA6 */
  "AFA7",                          /* 0xAFA7 */
  "AFA8",                          /* 0xAFA8 */
  "AFA9",                          /* 0xAFA9 */
  "AFAA",                          /* 0xAFAA */
  "AFAB",                          /* 0xAFAB */
  "AFAC",                          /* 0xAFAC */
  "AFAD",                          /* 0xAFAD */
  "AFAE",                          /* 0xAFAE */
  "AFAF",                          /* 0xAFAF */
  "AFB0",                          /* 0xAFB0 */
  "AFB1",                          /* 0xAFB1 */
  "AFB2",                          /* 0xAFB2 */
  "AFB3",                          /* 0xAFB3 */
  "AFB4",                          /* 0xAFB4 */
  "AFB5",                          /* 0xAFB5 */
  "AFB6",                          /* 0xAFB6 */
  "AFB7",                          /* 0xAFB7 */
  "AFB8",                          /* 0xAFB8 */
  "AFB9",                          /* 0xAFB9 */
  "AFBA",                          /* 0xAFBA */
  "AFBB",                          /* 0xAFBB */
  "AFBC",                          /* 0xAFBC */
  "AFBD",                          /* 0xAFBD */
  "AFBE",                          /* 0xAFBE */
  "AFBF",                          /* 0xAFBF */
  "AFC0",                          /* 0xAFC0 */
  "AFC1",                          /* 0xAFC1 */
  "AFC2",                          /* 0xAFC2 */
  "AFC3",                          /* 0xAFC3 */
  "AFC4",                          /* 0xAFC4 */
  "AFC5",                          /* 0xAFC5 */
  "AFC6",                          /* 0xAFC6 */
  "AFC7",                          /* 0xAFC7 */
  "AFC8",                          /* 0xAFC8 */
  "IconDispatch",                 /* 0xAFC9 */
  "DeviceLoop",              /* 0xAFCA */
  "AFCB",                          /* 0xAFCB */
  "AFCC",                          /* 0xAFCC */
  "AFCD",                          /* 0xAFCD */
  "AFCE",                          /* 0xAFCE */
  "AFCF",                          /* 0xAFCF */
  "AFD0",                          /* 0xAFD0 */
  "AFD1",                          /* 0xAFD1 */
  "AFD2",                          /* 0xAFD2 */
  "AFD3",                          /* 0xAFD3 */
  "AFD4",                          /* 0xAFD4 */
  "AFD5",                          /* 0xAFD5 */
  "AFD6",                          /* 0xAFD6 */
  "AFD7",                          /* 0xAFD7 */
  "AFD8",                          /* 0xAFD8 */
  "AFD9",                          /* 0xAFD9 */
  "AFDA",                          /* 0xAFDA */
  "AFDB",                          /* 0xAFDB */
  "AFDC",                          /* 0xAFDC */
  "AFDD",                          /* 0xAFDD */
  "AFDE",                          /* 0xAFDE */
  "AFDF",                          /* 0xAFDF */
  "AFE0",                          /* 0xAFE0 */
  "AFE1",                          /* 0xAFE1 */
  "AFE2",                          /* 0xAFE2 */
  "AFE3",                          /* 0xAFE3 */
  "AFE4",                          /* 0xAFE4 */
  "AFE5",                          /* 0xAFE5 */
  "AFE6",                          /* 0xAFE6 */
  "AFE7",                          /* 0xAFE7 */
  "AFE8",                          /* 0xAFE8 */
  "AFE9",                          /* 0xAFE9 */
  "AFEA",                          /* 0xAFEA */
  "AFEB",                          /* 0xAFEB */
  "AFEC",                          /* 0xAFEC */
  "DragManagerDispatch",          /* 0xAFED */
  "AFEE",                          /* 0xAFEE */
  "AFEF",                          /* 0xAFEF */
  "AFF0",                          /* 0xAFF0 */
  "AFF1",                          /* 0xAFF1 */
  "AFF2",                          /* 0xAFF2 */
  "AFF3",                          /* 0xAFF3 */
  "AFF4",                          /* 0xAFF4 */
  "AFF5",                          /* 0xAFF5 */
  "AFF6",                          /* 0xAFF6 */
  "AFF7",                          /* 0xAFF7 */
  "StdOpcodeProc",                /* 0xAFF8 */
  "AFF9",                          /* 0xAFF9 */
  "AFFA",                          /* 0xAFFA */
  "AFFB",                          /* 0xAFFB */
  "AFFC",                          /* 0xAFFC */
  "AFFD",                          /* 0xAFFD */
  "AFFE",                          /* 0xAFFE */
  "DebugStr",                     /* 0xAFFF */
};

#endif /* !NDEBUG */
